(* ========================================================================= *)
(* Treatment of type equality                                                *)
(*                                                                           *)
(*       Chantal Keller, Laboratoire d'Informatique de Polytechnique         *)
(*                                                                           *)
(*    (c) Copyright, Laboratoire d'Informatique de Polytechnique 2010        *)
(*              (c) Copyright, Chantal Keller 2010                           *)
(* ========================================================================= *)


Require Import ssreflect ssrbool ssrnat eqtype.
Require Import NArith.
Require Import hol ssrpositive.


(* Generic Cast *)

Inductive cast_result (A: Type) (n m: A) : Type :=
| Cast (k: forall P, P n -> P m)
| NoCast.

Implicit Arguments Cast [A].
Implicit Arguments NoCast [A n m].

Notation idcast := (Cast _ _ (fun _ x => x)).


(* Cast on names *)

Section namesCast.

  Definition T := names_dom.

  Open Local Scope positive_scope.

  Fixpoint names_cast (n m: T) {struct n} : cast_result T n m :=
    match n, m with
      | 1, 1 => idcast
      | p~0, q~0 =>
        match names_cast p q with
          | Cast k => Cast _ _ (fun Q H => k (fun x => Q (x~0)) H)
          | _ => NoCast
        end
      | p~1, q~1 =>
        match names_cast p q with
          | Cast k => Cast _ _ (fun Q H => k (fun x => Q (x~1)) H)
          | _ => NoCast
        end
      | _, _ => NoCast
    end.


  Lemma names_cast_same : forall n, names_cast n n = idcast.
  Proof. by elim => /= [p ->|p ->|]. Qed.


  Lemma names_cast_diff : forall n m, n != m -> names_cast n m = NoCast.
  Proof.
    elim => [n IHn|n IHn|] [m|m|] //= => H; case e: (n == m);
      by [move: H; move /eqP: e ->; move /eqP => H; elim H|
        rewrite (IHn m (negbT e))].
  Qed.

End namesCast.


(* Cast on types *)

Fixpoint cast (a1 a2 : type) {struct a2} : cast_result type a1 a2 :=
  match a1 as d1, a2 as d2 return cast_result type d1 d2 with
    | TVar A, TVar B =>
      match names_cast A B with
        | Cast k => Cast _ _ (fun P => k (fun p => P (TVar p)))
        | _ => NoCast
      end
    | TDef X l1, TDef Y l2 =>
      match names_cast X Y, list_cast l1 l2 with
        | Cast k, Cast l =>
          Cast _ _ (fun P h => k (fun Z => P (TDef Z l2))
            (l (fun l3 => P (TDef X l3)) h))
        | _, _ => NoCast
      end
    | Bool, Bool => idcast
    | b1 --> c1, b2 --> c2 =>
      match cast b1 b2, cast c1 c2 with
        | Cast kb, Cast kc =>
          let ka P :=
            let Pb d := P (d --> c1) in let Pc d := P (b2 --> d) in
              fun x => kc Pc (kb Pb x)
                in Cast _ _ ka
        | _, _ => NoCast
      end
    | _, _ => NoCast
  end

with list_cast (l1 l2: list_type) {struct l2} :
  cast_result list_type l1 l2 :=
  match l1 as m1, l2 as m2 return cast_result list_type m1 m2 with
    | Tnil, Tnil => idcast
    | Tcons A q1, Tcons B q2 =>
      match cast A B, list_cast q1 q2 with
        | Cast k, Cast l => Cast _ _ (fun P h => k
          (fun C => P (Tcons C q2)) (l (fun q3 => P (Tcons A q3)) h))
        | _, _ => NoCast
      end
    | _, _ => NoCast
  end.


Lemma cast_same_aux :
  (forall a, cast a a = idcast) /\ (forall l, list_cast l l = idcast).
Proof.
  apply ltype_ind => /= [X|C l ->||A -> B ->||A -> l ->] //=;
    by rewrite names_cast_same.
Qed.


Lemma cast_same : forall a, cast a a = idcast.
Proof. by move: cast_same_aux => [H _]. Qed.


Definition eqst a1 a2 := if cast a1 a2 is Cast _ then true else false.

Definition list_eqst l1 l2 :=
  if list_cast l1 l2 is Cast _ then true else false.


Lemma eqst_arrow : forall b1 c1 b2 c2,
  eqst (b1 --> c1) (b2 --> c2) = eqst b1 b2 && eqst c1 c2.
Proof. by move=> *; rewrite /eqst /=; do 2 case: (cast _ _). Qed.


Lemma eqst_tvar : forall n m, eqst (TVar n) (TVar m) = (n == m).
Proof.
  rewrite /eqst /= => n m; case e: (_ == _).
  - by move /eqP: e ->; rewrite names_cast_same.
  - by rewrite (names_cast_diff n m (negbT e)).
Qed.


Lemma eqst_tdef : forall X Y l1 l2,
  eqst (TDef X l1) (TDef Y l2) = ((X == Y) && (list_eqst l1 l2)).
Proof.
  rewrite /eqst /list_eqst /= => X Y l1 l2; case e: (_ == _).
  - by move /eqP: e ->; rewrite names_cast_same; case: (list_cast _ _).
  - by rewrite (names_cast_diff X Y (negbT e)).
Qed.


Lemma list_eqst_tcons : forall A B l1 l2,
  list_eqst (Tcons A l1) (Tcons B l2) = ((eqst A B) && (list_eqst l1 l2)).
Proof.
  rewrite /list_eqst /eqst /= => A B l1 l2;
    by case: (cast _ _); case: (list_cast _ _).
Qed.


Lemma eqstP : forall x y, reflect (x = y) (eqst x y).
Proof.
  apply (type_ltype_rec (fun x => forall y, reflect (x = y) (eqst x y))
    (fun l1 => forall l2, reflect (l1 = l2) (list_eqst l1 l2))) => [X [Y|||]|X l IHl [|Y l2||]|[|||]|A IHA B IHB [|||C D]|[|]|A IHA l IHl [|B l2]];
  try by constructor.
  - rewrite eqst_tvar; case e: (_ == _); constructor.
  by move /eqP: e ->.
  by move => [H]; rewrite H in e; rewrite eqxx in e.
  - rewrite eqst_tdef; case e: (_ == _) => /=.
  move: (IHl l2); case: (list_eqst _ _) => H; inversion_clear H; constructor.
  by move /eqP: e ->; rewrite H0.
  by move => [_ H]; rewrite H in H0.
  by constructor => [[H _]]; rewrite H in e; rewrite eqxx in e.
  - rewrite eqst_arrow; move: (IHA C); case: (eqst A C) => /= H; inversion_clear H.
  move: (IHB D); case: (eqst _ _) => /= H; inversion_clear H; constructor.
  by rewrite H0 H1.
  by move => [_ H]; rewrite H in H1.
  by constructor; move => [H _]; rewrite H in H0.
  - rewrite list_eqst_tcons; move: (IHA B); case: (eqst _ _) => /= H; inversion_clear H.
  move: (IHl l2); case: (list_eqst _ _) => /= H; inversion_clear H; constructor.
  by rewrite H0 H1.
  by move => [_ H]; rewrite H in H1.
  by constructor; move => [H _]; rewrite H in H0.
Qed.

Canonical Structure simple_type_domM := EqMixin eqstP.
Canonical Structure simple_type_dom := EqType simple_type_domM.


Lemma neq_cast :
   forall a1 a2, cast a1 a2 = if a1 == a2 then cast a1 a2 else NoCast.
Proof. by rewrite /eq_op /= /eqst => *; case cast. Qed.








(* Terms equality *)

Definition eq_cst (c d: cst) : bool :=
  match c, d with
    | Heq A, Heq B => A == B
    | Heps A, Heps B => A == B
    | Hand, Hand => true
    | Hor, Hor => true
    | Hnot, Hnot => true
    | Himp, Himp => true
    | Htrue, Htrue => true
    | Hfalse, Hfalse => true
    | Hforall A, Hforall B => A == B
    | Hexists A, Hexists B => A == B
    | _, _ => false
  end.


Fixpoint eq_term (t1 t2: term) {struct t1} : bool :=
  match t1, t2 with
    | Dbr n, Dbr m => n == m
    | Var x X, Var y Y => (x == y) && (X == Y)
    | Cst c, Cst d => eq_cst c d
    | Def c C, Def d D => (c == d) && (C == D)
    | App u1 u2, App v1 v2 => (eq_term u1 v1) && (eq_term u2 v2)
    | Abs A t, Abs B u => (A == B) && (eq_term t u)
    | _, _ => false
  end.


Lemma eq_termP : forall t1 t2, reflect (t1 = t2) (eq_term t1 t2).
Proof.
  elim => [n|x X|[C|C|||||||C|C]|c C|u1 IHu1 u2 IHu2|A t IHt]
    [m|y Y|[D|D|||||||D|D]|d D|v1 v2|B u] /=; try by constructor.
  - case e: (_ == _); constructor;
    by [move /eqP: e ->|move => [H]; move: e; rewrite H eq_refl].
  - case e: (_ == _);
    last by (constructor => [[H]]; move: e; rewrite H eq_refl).
  move /eqP: e ->; case e: (_ == _); constructor;
    by [move /eqP: e ->|move => [H]; move: e; rewrite H eq_refl].
  - case e: (_ == _); constructor;
    by [move /eqP: e ->|move => [H]; move: e; rewrite H eq_refl].
  - case e: (_ == _); constructor;
    by [move /eqP: e ->|move => [H]; move: e; rewrite H eq_refl].
  - case e: (_ == _); constructor;
    by [move /eqP: e ->|move => [H]; move: e; rewrite H eq_refl].
  - case e: (_ == _); constructor;
    by [move /eqP: e ->|move => [H]; move: e; rewrite H eq_refl].
  - case e: (_ == _);
    last by (constructor => [[H]]; move: e; rewrite H eq_refl).
  move /eqP: e ->; case e: (_ == _); constructor;
    by [move /eqP: e ->|move => [H]; move: e; rewrite H eq_refl].
  - move: (IHu1 v1) => [H1|H1] /=;
    last by (constructor => [[H2 _]]; apply H1).
  rewrite H1; move: (IHu2 v2) => {H1} [H1|H1] //=;
    constructor; first by rewrite H1.
  move => [H]; by apply H1.
  - case e: (_ == _);
    last by (constructor => [[H]]; move: e; rewrite H eq_refl).
  move /eqP: e ->; move: (IHt u) => [H1|H1]; constructor;
    first by rewrite H1.
  move => [H]; by apply H1.
Qed.

Canonical Structure term_domM := EqMixin eq_termP.
Canonical Structure term_dom := EqType term_domM.
