(* ========================================================================= *)
(* Representing HOL-Light derivations                                        *)
(*                                                                           *)
(*       Chantal Keller, Laboratoire d'Informatique de Polytechnique         *)
(*                                                                           *)
(*    (c) Copyright, Laboratoire d'Informatique de Polytechnique 2010        *)
(*              (c) Copyright, Chantal Keller 2010                           *)
(* ========================================================================= *)


Require Import ssreflect ssrbool eqtype ssrnat.
Require Import List.
Require Import list_ext ssrpositive axioms hol cast typing translation subst_db
  substitution.
Require ssrfsets.


Set Implicit Arguments.


(* Notations *)

(* Equality *)

Notation heq := (fun A a b => App (App (Cst (Heq A)) a) b).


(* Equivalent *)

Notation "p 'hequiv' q" := (App (App (Cst (Heq Bool)) p) q)
  (no associativity, at level 60).


Lemma wt_hequiv_inv1 : forall g p q,
  wt g (p hequiv q) Bool -> wt g p Bool /\ wt g q Bool.
Proof.
  by move => g p q h; move: (wt_eq_inv1 h).
Qed.


Lemma wt_hequiv_inv2 : forall g p q,
  wt g p Bool -> wt g q Bool -> wt g (p hequiv q) Bool.
Proof.
  by rewrite /wt => g p q hp hq /=; rewrite hp hq.
Qed.


(* Epsilon *)

Notation heps := Heps.


(* And *)

Notation "p 'hand' q" := (App (App (Cst Hand) p) q)
  (no associativity, at level 60).


Lemma wt_hand_inv1 : forall g p q,
  wt g (p hand q) Bool -> wt g p Bool /\ wt g q Bool.
Proof.
  by move => g p q h; move: (wt_app_inv1 h) => {h} [A [h1 h2]];
    move: (wt_app_inv1 h1) => {h1} [B [h1 h3]];
      rewrite /wt in h1; simpl in h1;
        move: h1 h2 h3 => [[<- [<-]]] {A B}.
Qed.


Lemma wt_hand_inv2 : forall g p q,
  wt g p Bool -> wt g q Bool -> wt g (p hand q) Bool.
Proof.
  by rewrite /wt => g p q hp hq /=; rewrite hp hq.
Qed.


(* Or *)

Notation "p 'hor' q" := (App (App (Cst Hor) p) q)
  (no associativity, at level 60).


Lemma wt_hor_inv1 : forall g p q,
  wt g (p hor q) Bool -> wt g p Bool /\ wt g q Bool.
Proof.
  by move => g p q h; move: (wt_app_inv1 h) => {h} [A [h1 h2]];
    move: (wt_app_inv1 h1) => {h1} [B [h1 h3]];
      rewrite /wt in h1; simpl in h1;
        move: h1 h2 h3 => [[<- [<-]]] {A B}.
Qed.


Lemma wt_hor_inv2 : forall g p q,
  wt g p Bool -> wt g q Bool -> wt g (p hor q) Bool.
Proof.
  by rewrite /wt => g p q hp hq /=; rewrite hp hq.
Qed.


(* Not *)

Notation "'hnot' p" := (App (Cst Hnot) p) (at level 60).


Lemma wt_hnot_inv1 : forall g p,
  wt g (hnot p) Bool -> wt g p Bool.
Proof.
  by move => g p h; move: (wt_app_inv1 h) => {h} [A [h1 h2]];
    rewrite /wt in h1; simpl in h1; move: h1 h2 => [[<-]] {A}.
Qed.


Lemma wt_hnot_inv2 : forall g p,
  wt g p Bool -> wt g (hnot p) Bool.
Proof.
  by rewrite /wt => g p hp /=; rewrite hp.
Qed.


(* Implies *)

Notation "p 'himp' q" := (App (App (Cst Himp) p) q)
  (no associativity, at level 60).


Lemma wt_himp_inv1 : forall g p q,
  wt g (p himp q) Bool -> wt g p Bool /\ wt g q Bool.
Proof.
  by move => g p q h; move: (wt_app_inv1 h) => {h} [A [h1 h2]];
    move: (wt_app_inv1 h1) => {h1} [B [h1 h3]];
      rewrite /wt in h1; simpl in h1;
        move: h1 h2 h3 => [[<- [<-]]] {A B}.
Qed.


Lemma wt_himp_inv2 : forall g p q,
  wt g p Bool -> wt g q Bool -> wt g (p himp q) Bool.
Proof.
  by rewrite /wt => g p q hp hq /=; rewrite hp hq.
Qed.


(* True *)

Notation htrue := (Cst Htrue).


(* False *)

Notation hfalse := (Cst Hfalse).


(* Forall *)

Notation hforall := (fun A P => App (Cst (Hforall A)) (Abs A P)).


Lemma wt_hforall_inv1 : forall g P A,
  wt g (hforall A P) Bool -> wt (A::g) P Bool.
Proof.
  move => g P A H1; move: (wt_app_inv1 H1) => {H1} [B [H1 H2]];
    move: H1 H2; rewrite {1}/wt /= => [[[<-]]] => H;
      move: (wt_abs_inv1 H) => [C [[H1] H2]]; move /eqP: H1 => [H1];
        by rewrite -H1 in H2.
Qed.


Lemma wt_hforall_inv2 : forall g P A,
  wt (A::g) P Bool -> wt g (hforall A P) Bool.
Proof.
  move => g P A H1; apply: wt_app_inv2; split; first done.
  by apply wt_abs_inv2.
Qed.


(* Exists *)

Notation hexists := (fun A P => App (Cst (Hexists A)) (Abs A P)).


Lemma wt_hexists_inv1 : forall g P A,
  wt g (hexists A P) Bool -> wt (A::g) P Bool.
Proof.
  move => g P A H1; move: (wt_app_inv1 H1) => {H1} [B [H1 H2]];
    move: H1 H2; rewrite {1}/wt /= => [[[<-]]] => H;
      move: (wt_abs_inv1 H) => [C [[H1] H2]]; move /eqP: H1 => [H1];
        by rewrite -H1 in H2.
Qed.


Lemma wt_hexists_inv2 : forall g P A,
  wt (A::g) P Bool -> wt g (hexists A P) Bool.
Proof.
  move => g P A H1; apply: wt_app_inv2; split; first done.
  by apply wt_abs_inv2.
Qed.


(* Sets of hypothesis (contexts of derivations) *)

Definition hyp_set := ssrfsets.t term_dom.
Definition hyp_empty := ssrfsets.empty (E:=term_dom).
Definition hyp_remove := ssrfsets.remove.
Definition hyp_add := ssrfsets.add.
Definition hyp_union := ssrfsets.union.
Definition hyp_singl := ssrfsets.singleton.
Definition hyp_subset := ssrfsets.Subset.
Definition wf := ssrfsets.For_all (fun t => wt nil t Bool).
Definition hyp_is_not_free (x: idV) (X: type) :=
  ssrfsets.for_all (is_not_free x X).
Definition hyp_subst_idt (h: hyp_set) (S: substitution_idt) :=
  ssrfsets.map (fun t => subst_idt t S) h.
Definition hyp_subst_idv (h: hyp_set) (s: substitution_idv) :=
  ssrfsets.map (fun t => subst_idv t s) h.
Definition hyp_frees_idv (h: hyp_set) : list idV :=
  flatten (List.map frees_idv (ssrfsets.elements h)).

Implicit Arguments hyp_singl [E].
Implicit Arguments hyp_subset [E].


(* Utilities about well-formedness *)

Lemma wf_empty : wf hyp_empty.
Proof.
  apply: ssrfsets.For_all_empty.
Qed.


Lemma wf_singl1 : forall t,
  wf (hyp_singl t) -> wt nil t Bool.
Proof.
  move => t h; apply (ssrfsets.For_all_singleton1 h).
Qed.


Lemma wf_singl2 : forall t,
  wt nil t Bool -> wf (hyp_singl t).
Proof.
  move => t h;
    apply (ssrfsets.For_all_singleton2 (fun t => wt nil t Bool) h).
Qed.


Lemma wf_union1 : forall h1 h2,
  wf (hyp_union h1 h2) -> wf h1 /\ wf h2.
Proof. rewrite /wf => h1 h2; apply: ssrfsets.For_all_union1. Qed.


Lemma wf_union2 : forall h1 h2,
  wf h1 -> wf h2 -> wf (hyp_union h1 h2).
Proof. rewrite /wf => h1 h2; apply: ssrfsets.For_all_union2. Qed.


Lemma wf_remove1 : forall p h,
  wt nil p Bool -> wf (hyp_remove p h) -> wf h.
Proof. by rewrite /wf => p h H; apply: ssrfsets.For_all_remove1. Qed.


Lemma wf_remove2 : forall p h,
  wf h -> wf (hyp_remove p h).
Proof. by rewrite /wf => p h H; apply: ssrfsets.For_all_remove2. Qed.


(* Utilities about free variables *)

Lemma free_singl1 : forall x X t,
  hyp_is_not_free x X (hyp_singl t) -> is_not_free x X t.
Proof. move => x X t h; apply (ssrfsets.for_all_singleton1 h). Qed.


Lemma free_add1 : forall x X t h, hyp_is_not_free x X (hyp_add t h) ->
  ((hyp_is_not_free x X h) && (is_not_free x X t)).
Proof.
  rewrite /hyp_is_not_free => x X t h; apply: ssrfsets.for_all_add1.
Qed.


Lemma free_forallb : forall x h,
  forallb (fun y : idV => x != y) (hyp_frees_idv h) ->
  forall X, hyp_is_not_free x X h.
Proof.
  rewrite /hyp_frees_idv /hyp_is_not_free => x h H1 X;
    move: (forallb_flatten1 _ _ H1) => {H1} H1; ssrfsets.for_all_tac => a H;
      apply frees_idv_is_not_free; move: (ssrfsets.elements_1 H);
        move: (forallb_map1 _ _ _ H1) => H2 H3; by apply H2.
Qed.


Lemma free_forallb2 : forall x h,
  forallb (fun y : idV => y != x) (hyp_frees_idv h) ->
  forall X, hyp_is_not_free x X h.
Proof.
  move => x h H1 X; apply free_forallb;
    have H: ((fun y : idV => x != y) = (fun y : idV => y != x));
    last by rewrite H.
  extensionality y; apply not_eq_sym.
Qed.

Implicit Arguments free_forallb2 [].


Lemma hyp_is_not_free_remove : forall y X p h,
  hyp_is_not_free y X (hyp_remove p h) -> is_not_free y X p ->
  hyp_is_not_free y X h.
Proof.
  rewrite /hyp_is_not_free /hyp_remove => y X p h H1 H2;
    apply: ssrfsets.for_all_remove1; by [eexact H2|].
Qed.


(* Renaming *)

Fixpoint renaming (t: term) (x y: idV) (X: type) :=
  match t with
    | Var z Z => if ((x == z) && (X == Z)) then Var y Z else Var z Z
    | App u v => App (renaming u x y X) (renaming v x y X)
    | Abs A u => Abs A (renaming u x y X)
    | u => u
  end.


Lemma wt_renaming : forall x y X t g,
  infer g (renaming t x y X) = infer g t.
Proof.
  move => x y X; elim => [n|z Z|c|c C|t IHt u IHu|A t IHt] g //=.
  - by case: (_ == _); [case: (_ == _)|].
  - by rewrite (IHt g) (IHu g).
  - by rewrite (IHt (A::g)).
Qed.


Lemma wt_renaming_inv2 : forall x y X g t A,
  wt g t A -> wt g (renaming t x y X) A.
Proof.
  by move => x y X g t A; rewrite /wt (wt_renaming x y X).
Qed.


Lemma renaming_lift_term : forall x y X j u i,
  renaming (lift_term u i j) x y X = lift_term (renaming u x y X) i j.
Proof.
  move => x y X j; elim => [n|z Z|c|c C|t IHt u IHu|A t IHt] i //=.
  - by case: (_ <= _).
  - by case: (_ == _); [case: (_ == _)|].
  - by rewrite (IHt i) (IHu i).
  - by rewrite (IHt (i.+1)).
Qed.


Lemma renaming_subst_db : forall x y X t i u,
  renaming (subst_db t i u) x y X =
  subst_db (renaming t x y X) i (renaming u x y X).
Proof.
  move => x y X; elim => [n|z Z|c|c C|t IHt v IHv|A t IHt] i u //=.
  - by case: (_ < _); [|case: (_ == _)].
  - by case: (_ == _); [case: (_ == _)|].
  - by rewrite (IHt i u) (IHv i u).
  - by rewrite (IHt (i.+1) (lift_term u 0 1)) renaming_lift_term.
Qed.


Lemma renaming_open : forall t u x y X, renaming (open t u) x y X =
  open (renaming t x y X) (renaming u x y X).
Proof.
  move => t u x y X; apply renaming_subst_db.
Qed.


Lemma free_renaming : forall x y X t,
  is_not_free x X t -> renaming t x y X = t.
Proof.
  move => x y X; elim => [n|z Z|c|c C|t IHt u IHu|A t IHt] //=.
  - move /orP => [H|H]; move: (negbTE H) ->; [done|by rewrite andb_false_r].
  - move /andP => [H1 H2]; by rewrite (IHt H1) (IHu H2).
  - move => H; by rewrite (IHt H).
Qed.

Implicit Arguments free_renaming [].


(* Injectivity of renaming under freshness conditions *)

Lemma inj_renaming : forall x y X u v, is_not_free y X u ->
  is_not_free y X v -> (u = v <-> renaming u x y X = renaming v x y X).
Proof.
  move => x y X u v H1 H2; split; first by move => ->.
  move: u v H1 H2.
  elim => [n|z Z|c|c C|t IHt u IHu|A t IHt]
    [i|z2 Z2|d|d D|v w|B v] //=; try (move => _ _; case: (_ == _);
      case: (_ == _) => /= H; discriminate H).
  - case e: (x == z) => /=.
  move /eqP: e => -> {x}.
  case e: (X == Z) => /=.
  move /eqP: e => -> {X}.
  case e: (z == z2) => /=.
  move /eqP: e => -> {z}.
  case e: (Z == Z2) => /=.
  by move /eqP: e => -> {Z}.
  rewrite orb_false_r => H _ H1.
  move: H1 H => [-> _]; by rewrite eq_refl.
  move => _ H1 H2.
  move: H2 H1 => [-> ->]. by rewrite !eq_refl.
  case e2: (z == z2) => /=; last done.
  move /eqP: e2 => -> {z}.
  move => _; move /orP => [H|H].
  case: (X == Z2) => /=; last done.
  by move => [_ ->].
  by move: (negbTE H) ->.
  case e2: (x == z2) => /=; last done.
  move /eqP: e2 => <- {z2}.
  move => H1; move /orP => [H2|H2].
  case e2: (X == Z2); last done.
  move /eqP: e2 => <- {Z2}.
  move => H3; move: H3 e H1 H2 => [-> ->] {z Z}.
  by rewrite !eq_refl.
  by move: (negbTE H2) ->.
  - move /andP => [H1 H2]; move /andP => [H3 H4] [H5 H6];
    by rewrite (IHt v H1 H3 H5) (IHu w H2 H4 H6).
  - move => H1 H2 [-> H3]; by rewrite (IHt v H1 H2 H3).
Qed.


(* Renaming and sets of hypotheses *)

Definition hyp_renaming (h: hyp_set) x y X :=
  ssrfsets.map (fun t => renaming t x y X) h.


Lemma hyp_renaming_remove : forall x y X p h, hyp_subset
    (hyp_remove (renaming p x y X) (hyp_renaming h x y X))
    (hyp_renaming (hyp_remove p h) x y X).
Proof.
  move => x y X p h; exact (ssrfsets.map_remove
    (fun t => renaming t x y X) p h).
Qed.


Lemma free_hyp_renaming : forall x y X h,
  hyp_is_not_free x X h -> hyp_subset (hyp_renaming h x y X) h.
Proof.
  rewrite /hyp_is_not_free => x y X h H1; ssrfsets.for_all_tac;
    apply: ssrfsets.map_subset_For_all; first eassumption.
  simpl => t. apply free_renaming.
Qed.

Implicit Arguments free_hyp_renaming [X h].


Lemma hyp_renaming_subset : forall h1 h2 x y X, hyp_subset h1 h2 ->
  hyp_subset (hyp_renaming h1 x y X) (hyp_renaming h2 x y X).
Proof. move => h1 h2 x y X; apply ssrfsets.map_subset. Qed.


Lemma hyp_is_not_free_subset : forall h1 h2 y X,
  hyp_subset h1 h2 -> hyp_is_not_free y X h2 -> hyp_is_not_free y X h1.
Proof. move => h1 h2 y X; apply ssrfsets.for_all_subset. Qed.


Lemma hyp_renaming_singl : forall t x y X,
  hyp_subset (hyp_singl (renaming t x y X))
  (hyp_renaming (hyp_singl t) x y X).
Proof.
  move => t x y X;
    apply (ssrfsets.map_singleton (fun t => renaming t x y X)).
Qed.


Lemma hyp_renaming_add : forall h t x y X,
  hyp_subset (hyp_renaming (hyp_add t h) x y X)
  (hyp_add (renaming t x y X) (hyp_renaming h x y X)).
Proof. move => h t x y X; apply: ssrfsets.map_add. Qed.

Implicit Arguments hyp_renaming_add [].


Lemma hyp_renaming_union : forall h1 h2 x y X,
  hyp_subset (hyp_union (hyp_renaming h1 x y X) (hyp_renaming h2 x y X))
  (hyp_renaming (hyp_union h1 h2) x y X).
Proof. move => h1 h2 x y X; apply ssrfsets.map_union. Qed.


(* Substituting and sets of hypotheses *)

Lemma hyp_subst_idt_remove : forall h p S, hyp_subset
  (hyp_remove (subst_idt p S) (hyp_subst_idt h S))
  (hyp_subst_idt (hyp_remove p h) S).
Proof.
  move => h p S; exact (ssrfsets.map_remove
    (fun t => subst_idt t S) p h).
Qed.


Lemma hyp_subst_idt_subset : forall h1 h2 S, hyp_subset h1 h2 ->
  hyp_subset (hyp_subst_idt h1 S) (hyp_subst_idt h2 S).
Proof. move => h1 h2 S; apply ssrfsets.map_subset. Qed.


Lemma hyp_subst_idt_singl : forall t S,
  hyp_subset (hyp_singl (subst_idt t S)) (hyp_subst_idt (hyp_singl t) S).
Proof.
  move => t S;
    apply (ssrfsets.map_singleton (fun t => subst_idt t S)).
Qed.


Lemma hyp_subst_idt_union : forall h1 h2 S,
  hyp_subset (hyp_union (hyp_subst_idt h1 S) (hyp_subst_idt h2 S))
  (hyp_subst_idt (hyp_union h1 h2) S).
Proof. move => h1 h2 S; apply ssrfsets.map_union. Qed.


Lemma hyp_subst_idt_add : forall h t S,
  hyp_subset (hyp_subst_idt (hyp_add t h) S)
  (hyp_add (subst_idt t S) (hyp_subst_idt h S)).
Proof. move => h t S; apply: ssrfsets.map_add. Qed.


Lemma hyp_subst_idv_remove : forall h p s, hyp_subset
  (hyp_remove (subst_idv p s) (hyp_subst_idv h s))
  (hyp_subst_idv (hyp_remove p h) s).
Proof.
  move => h p s; exact (ssrfsets.map_remove
    (fun t => subst_idv t s) p h).
Qed.


Lemma hyp_subst_idv_subset : forall h1 h2 s, hyp_subset h1 h2 ->
  hyp_subset (hyp_subst_idv h1 s) (hyp_subst_idv h2 s).
Proof. move => h1 h2 s; apply ssrfsets.map_subset. Qed.


Lemma hyp_subst_idv_singl : forall t s,
  hyp_subset (hyp_singl (subst_idv t s)) (hyp_subst_idv (hyp_singl t) s).
Proof.
  move => t s;
    apply (ssrfsets.map_singleton (fun t => subst_idv t s)).
Qed.


Lemma hyp_subst_idv_union : forall h1 h2 s,
  hyp_subset (hyp_union (hyp_subst_idv h1 s) (hyp_subst_idv h2 s))
  (hyp_subst_idv (hyp_union h1 h2) s).
Proof. move => h1 h2 s; apply ssrfsets.map_union. Qed.


Lemma hyp_subst_idv_add : forall h t s,
  hyp_subset (hyp_subst_idv (hyp_add t h) s)
  (hyp_add (subst_idv t s) (hyp_subst_idv h s)).
Proof. move => h t s; apply: ssrfsets.map_add. Qed.


(* Judgments *)

Inductive deriv : hyp_set -> term -> Prop :=

(* Rules without binders *)

| Drefl : forall t A, wt nil t A ->
  deriv hyp_empty (heq A t t)
| Dassume : forall t, wt nil t Bool -> deriv (hyp_singl t) t
| Dbeta1 : forall t u X A, wt (X::nil) t A -> wt nil u X ->
  deriv hyp_empty (heq A (App (Abs X t) u) (open t u))
| Deta : forall f A B, wt nil f (A --> B) ->
  deriv hyp_empty (heq (A --> B) (Abs A (App f (Dbr 0))) f)
| Deps : forall A P x, wt nil P (A --> Bool) -> wt nil x A ->
  deriv hyp_empty ((App P x) himp (App P (App (Cst (Heps A)) P)))
| Dtrans : forall h1 h2 t1 t2 t3 A,
  deriv h1 (heq A t1 t2) -> deriv h2 (heq A t2 t3) ->
  deriv (hyp_union h1 h2) (heq A t1 t3)
| Dsym : forall h t1 t2 A,
  deriv h (heq A t2 t1) -> deriv h (heq A t1 t2)
| Deqmp : forall h1 h2 a b,
  deriv h1 (a hequiv b) -> deriv h2 a -> deriv (hyp_union h1 h2) b
| Dmp : forall h1 h2 a b,
  deriv h1 (a himp b) -> deriv h2 a -> deriv (hyp_union h1 h2) b
| Dimpantisym : forall h1 h2 a b,
  deriv h1 (a himp b) -> deriv h2 (b himp a) ->
  deriv (hyp_union h1 h2) (b hequiv a)
| Dconj : forall h1 h2 a b,
  deriv h1 a -> deriv h2 b -> deriv (hyp_union h1 h2) (a hand b)
| Dconj1 : forall h a b,
  deriv h (a hand b) -> deriv h a
| Dconj2 : forall h a b,
  deriv h (a hand b) -> deriv h b
| Dnoti : forall h a,
  deriv h (a himp hfalse) -> deriv h (hnot a)
| Dnote : forall h a,
  deriv h (hnot a) -> deriv h (a himp hfalse)
| Ddisj1 : forall h a b, wt nil b Bool ->
  deriv h a -> deriv h (a hor b)
| Ddisj2 : forall h a b, wt nil a Bool ->
  deriv h b -> deriv h (a hor b)
| Ddisjcases : forall h1 h2 h3 a b c,
  deriv h1 (a hor b) -> deriv h2 c -> deriv h3 c ->
  deriv (hyp_union h1 (hyp_union (hyp_remove a h2) (hyp_remove b h3))) c
| Dcontr : forall h a, wt nil a Bool ->
  deriv h hfalse -> deriv (hyp_remove (hnot a) h) a
| Dtrue : deriv hyp_empty htrue
| Ddisch : forall h p q, wt nil p Bool -> deriv h q ->
  deriv (hyp_remove p h) (p himp q)

(* Higher-order rules *)

| Dmkcomb : forall h1 h2 f g u v A B,
  deriv h1 (heq (A --> B) f g) -> deriv h2 (heq A u v) ->
  deriv (hyp_union h1 h2) (heq B (App f u) (App g v))
| Dabs1 : forall h u v (L: list idV) A X,
  (forall x, (forallb (fun y => y != x) L) ->
    deriv h (heq A (open u (Var x X)) (open v (Var x X)))) ->
  deriv h (heq (X --> A) (Abs X u) (Abs X v))
| Dspec : forall h p t A, wt nil t A ->
  deriv h (hforall A p) -> deriv h (open p t)
| Dgen1 : forall h t (L: list idV) X,
  (forall x, (forallb (fun y => y != x) L) ->
    deriv h (open t (Var x X))) ->
  deriv h (hforall X t)
| Dexists : forall h a t A, wt nil t A ->
  deriv h (open a t) -> deriv h (hexists A a)
| Dchoose1 : forall h1 h2 (L: list idV) a c A, deriv h1 (hexists A a) ->
  (forall x, (forallb (fun y => y != x) L) ->
    deriv (hyp_add (open a (Var x A)) h2) c) -> deriv (hyp_union h1 h2) c

(* Weakening rule *)

| Dweak : forall h1 h2 t, hyp_subset h1 h2 -> deriv h1 t -> deriv h2 t.

Implicit Arguments Dchoose1 [h1 h2 a c A].
Implicit Arguments Dabs1 [h A].
Implicit Arguments Dgen1 [h].


(* Derivability of the renaming rule *)

Lemma drenaming : forall h t, deriv h t -> forall x y X,
  deriv (hyp_renaming h x y X) (renaming t x y X).
Proof.
  move => h0 t0; elim => {h0 t0} /=.
  - move => t A h x y X; apply: Dweak; first apply ssrfsets.empty_subset.
  apply Drefl; by rewrite /wt wt_renaming.
  - move => t H1 x y X; apply: Dweak; first apply hyp_renaming_singl.
  apply Dassume; by apply wt_renaming_inv2.
  - move => t u X A H1 H2 x y B; apply: Dweak;
    first apply ssrfsets.empty_subset.
  rewrite renaming_open; apply Dbeta1; by apply wt_renaming_inv2.
  - move => f A B H1 x y X; apply: Dweak; first apply ssrfsets.empty_subset.
  apply Deta; by apply wt_renaming_inv2.
  - move => A P x H1 H2 y z Y; apply: Dweak;
    first apply ssrfsets.empty_subset.
  apply Deps; by apply wt_renaming_inv2.
  - move => h1 h2 t1 t2 t3 A _ IH1 _ IH2 x y X;
    apply: Dweak; first apply hyp_renaming_union.
  by apply: Dtrans.
  - move => h t1 t2 A _ IH x y X; by apply Dsym.
  - move => h1 h2 a b _ IH1 _ IH2 x y X;
    apply: Dweak; first apply hyp_renaming_union.
  by apply: Deqmp.
  - move => h1 h2 a b _ IH1 _ IH2 x y X;
    apply: Dweak; first apply hyp_renaming_union.
  by apply: Dmp.
  - move => h1 h2 a b _ IH1 _ IH2 x y X;
    apply: Dweak; first apply hyp_renaming_union.
  by apply Dimpantisym.
  - move => h1 h2 a b _ IH1 _ IH2 x y X;
    apply: Dweak; first apply hyp_renaming_union.
  by apply Dconj.
  - move => h a b _ IH x y X; by apply: Dconj1.
  - move => h a b _ IH x y X; by apply: Dconj2.
  - move => h a _ IH x y X; by apply Dnoti.
  - move => h a _ IH x y X; by apply Dnote.
  - move => h a b H1 _ IH x y X; apply Ddisj1; last done.
  by apply wt_renaming_inv2.
  - move => h a b H1 _ IH x y X; apply Ddisj2; last done.
  by apply wt_renaming_inv2.
  - move => h1 h2 h3 a b c _ IH1 _ IH2 _ IH3 x y X;
    apply: Dweak; first apply hyp_renaming_union.
  apply: Dweak;
    first (apply ssrfsets.union_subset; apply hyp_renaming_union).
  apply: Dweak; first (apply ssrfsets.union_subset;
    apply ssrfsets.union_subset_all; apply hyp_renaming_remove).
  by apply Ddisjcases.
  - move => h a H1 _ IH x y X;
    apply: Dweak; first apply hyp_renaming_remove.
  apply Dcontr; last done.
  by apply wt_renaming_inv2.
  - move => x y X; apply: Dweak; first apply ssrfsets.empty_subset.
  apply Dtrue.
  - move => h p q H1 H2 H3 x y X;
    apply: Dweak; first apply hyp_renaming_remove.
  apply Ddisch; first by apply wt_renaming_inv2.
  apply H3.
  - move => h1 h2 f g u v A B _ IH1 _ IH2 x y X;
    apply: Dweak; first apply hyp_renaming_union.
  by apply: Dmkcomb.
  - move => h u v L A X _ IH z y Y;
    apply (Dabs1 _ _ (z::L)) => x /=; move /andP => [H1 H2];
      replace (Var x X) with (renaming (Var x X) z y Y);
        last by (simpl; rewrite (negbTE H1)).
  rewrite -!renaming_open; by apply IH.
  - move => h p t A H1 _ IH x y X; rewrite renaming_open;
    by apply: Dspec; first (apply wt_renaming_inv2; eassumption).
  - move => h t L X _ IH z y Y;
    apply (Dgen1 _ (z::L)) => x /=; move /andP => [H1 H2];
      replace (Var x X) with (renaming (Var x X) z y Y);
        last by (simpl; rewrite (negbTE H1)).
  rewrite -!renaming_open; by apply IH.
  - move => h a t A H1 _ IH x y X; move: (IH x y X);
    rewrite renaming_open => H; apply: Dexists; last eassumption.
  by apply wt_renaming_inv2.
  - move => h1 h2 L a c A _ IH1 _ IH2 x y X; apply: Dweak;
    first apply hyp_renaming_union.
  apply: (Dchoose1 (x::L)); first apply IH1.
  move => z /=; move /andP => [H1 H2];
    replace (Var z A) with (renaming (Var z A) x y X);
      last by rewrite /= (negbTE H1).
  rewrite -renaming_open; apply: Dweak; first apply hyp_renaming_add.
  simpl; by apply IH2.
  - move => h1 h2 t H1 H2 H3 x y X; apply: Dweak;
    first (apply hyp_renaming_subset; eexact H1).
  apply H3.
Qed.


(* Derivability of the real Dbeta rule *)

Lemma Dbeta : forall x t X A, wt nil t A ->
  deriv hyp_empty (heq A (App (Abs X (close t x X)) (Var x X)) t).
Proof.
  move => x t X A H1; rewrite -{2}(open_close x X t);
    apply Dbeta1; last done.
  by apply wt_close_inv2.
Qed.


(* Derivability of the real Dabs rule *)

Lemma Dabs : forall h x u v A X,
  hyp_is_not_free x X h -> deriv h (heq A u v) ->
  deriv h (heq (X --> A) (Abs X (close u x X)) (Abs X (close v x X))).
Proof.
  move => h x u v A X H1;
    rewrite -{1}(open_close x X u) -{1}(open_close x X v) => H2;
      apply (Dabs1 (h:=h) (close u x X) (close v x X) nil
        X) => y _; apply: Dweak;
      first (apply (free_hyp_renaming x y); eassumption).
  move: (drenaming H2 x y X);
    rewrite /= !renaming_open /= !eq_refl /= !free_renaming;
      [done|apply close_not_free|apply close_not_free].
Qed.


(* Derivability of the real Dgen rule *)

Lemma Dgen : forall h c x A, hyp_is_not_free x A h ->
  deriv h c -> deriv h (hforall A (close c x A)).
Proof.
  move => /= h c x A H1; rewrite -{1}(open_close x A c) => H2;
    apply (Dgen1 (h:=h) (close c x A) nil A) => y _;
      apply: Dweak; first (apply (free_hyp_renaming x y); eassumption).
  move: (drenaming H2 x y A);
    rewrite renaming_open /= !eq_refl /= free_renaming; first done.
  apply close_not_free.
Qed.


(* Derivability of the real Dchoose rule *)

Lemma Dchoose2 : forall h1 h2 v P Q A,
  hyp_is_not_free v A h2 -> is_not_free v A Q -> is_not_free v A P ->
  deriv h1 (hexists A P) -> deriv (hyp_add (open P (Var v A)) h2) Q ->
  deriv (hyp_union h1 h2) Q.
Proof.
  move => h1 h2 v P Q A H1 H2 H3 /= H4 H5;
    apply: (Dchoose1 nil); first eexact H4.
  move => x _; move: (drenaming H5 v x A);
    rewrite (free_renaming _ x _ _ H2) => H6;
      move: (Dweak (hyp_renaming_add h2 (open P (Var v A)) v x A) H6);
        rewrite renaming_open /= !eq_refl /= (free_renaming _ x _ _ H3)
          => H7; apply: Dweak; last eexact H7.
  apply ssrfsets.add_subset; by apply free_hyp_renaming.
Qed.


Lemma Dchoose : forall h1 h2 v P Q A,
  hyp_is_not_free v A (hyp_remove (open P (Var v A)) h2) ->
  is_not_free v A Q -> is_not_free v A P ->
  deriv h1 (hexists A P) -> deriv h2 Q ->
  deriv (hyp_union h1 (hyp_remove (open P (Var v A)) h2)) Q.
Proof.
  move => h1 h2 v P Q A H1 H2 H5 /= H3 H4;
    apply: Dchoose2; [eexact H1|exact H2|eexact H5|eexact H3|].
  apply: Dweak; last eexact H4.
  apply ssrfsets.add_remove_subset.
Qed.


(* Derivability of the type substitution rule *)

Lemma Dinsttype : forall h c S, deriv h c ->
  deriv (hyp_subst_idt h S) (subst_idt c S).
Proof.
  move => h0 c0 S; elim => {h0 c0} /=.
  - move => t A h1; apply: Dweak; first apply ssrfsets.empty_subset.
  apply Drefl; apply (wt_subst_idt_inv2 S h1).
  - move => t H1; apply: Dweak; first apply hyp_subst_idt_singl.
  apply Dassume; apply (wt_subst_idt_inv2 S H1).
  - move => t u X A H1 H2; apply: Dweak; first apply ssrfsets.empty_subset.
  rewrite subst_idt_open; apply Dbeta1;
    by [apply (wt_subst_idt_inv2 S (g:=X::nil) (t:=t) (A:=A))|
      apply (wt_subst_idt_inv2 S (g:=nil) (t:=u) (A:=X))].
  - move => f A B H1; apply: Dweak; first apply ssrfsets.empty_subset.
  apply Deta; by apply (wt_subst_idt_inv2 S H1).
  - move => A P x H1 H2; apply: Dweak; first apply ssrfsets.empty_subset.
  apply Deps;
    by [apply (wt_subst_idt_inv2 S H1)|apply (wt_subst_idt_inv2 S H2)].
  - move => h1 h2 t1 t2 t3 A _ IH1 _ IH2;
    apply: Dweak; first apply hyp_subst_idt_union.
  apply: Dtrans; eassumption.
  - move => h t1 t2 A _ IH; by apply Dsym.
  - move => h1 h2 a b _ IH1 _ IH2;
    apply: Dweak; first apply hyp_subst_idt_union.
  apply: Deqmp; eassumption.
  - move => h1 h2 a b _ IH1 _ IH2;
    apply: Dweak; first apply hyp_subst_idt_union.
  apply: Dmp; eassumption.
  - move => h1 h2 a b _ IH1 _ IH2;
    apply: Dweak; first apply hyp_subst_idt_union.
  by apply Dimpantisym.
  - move => h1 h2 a b _ IH1 _ IH2;
    apply: Dweak; first apply hyp_subst_idt_union.
  by apply Dconj.
  - move => h a b _ IH; apply: Dconj1; eexact IH.
  - move => h a b _ IH; apply: Dconj2; eexact IH.
  - move => h a _ IH; by apply Dnoti.
  - move => h a _ IH; by apply Dnote.
  - move => h a b H1 _ IH; apply Ddisj1; last done.
  apply (wt_subst_idt_inv2 S H1).
  - move => h a b H1 _ IH; apply Ddisj2; last done.
  apply (wt_subst_idt_inv2 S H1).
  - move => h1 h2 h3 a b c _ IH1 _ IH2 _ IH3;
    apply: Dweak; first apply hyp_subst_idt_union.
  apply: Dweak;
    first (apply ssrfsets.union_subset; apply hyp_subst_idt_union).
  apply: Dweak; first (apply ssrfsets.union_subset;
    apply ssrfsets.union_subset_all; apply hyp_subst_idt_remove).
  by apply Ddisjcases.
  - move => h a H1 _ IH;
    apply: Dweak; first apply hyp_subst_idt_remove.
  apply Dcontr; last done.
  apply (wt_subst_idt_inv2 S H1).
  - apply: Dweak; first apply ssrfsets.empty_subset.
  apply Dtrue.
  - move => h p q H1 H2 H3; apply: Dweak; first apply hyp_subst_idt_remove.
  apply Ddisch; last done.
  exact (wt_subst_idt_inv2 S H1).
  - move => h1 h2 f g u v A B _ IH1 _ IH2;
    apply: Dweak; first apply hyp_subst_idt_union.
  apply: Dmkcomb; eassumption.
  - move => h u v L A X H IH; apply (Dabs1 _ _ L) => x H1;
    move: (IH x H1); by rewrite !subst_idt_open /=.
  - move => h p t A H1 _ IH; rewrite subst_idt_open; apply: Dspec;
    last eassumption.
  by apply (wt_subst_idt_inv2 S (g:=nil)).
  - move => h t L X _ IH; apply (Dgen1 _ L) => x H1;
    move: (IH x H1); by rewrite !subst_idt_open /=.
  - move => h a t A H1 _; rewrite subst_idt_open => IH;
    apply: Dexists; last eassumption.
  by apply (wt_subst_idt_inv2 S (g:=nil)).
  - move => h1 h2 L a c A _ IH1 _ IH2;
    apply: Dweak; first apply hyp_subst_idt_union.
  apply: (Dchoose1 L); first eexact IH1.
  move => x Hx;
    replace (open (subst_idt a S) (Var x (subst_idt_type A S))) with
      (subst_idt (open a (Var x A)) S); last apply subst_idt_open.
  apply: Dweak; first apply hyp_subst_idt_add.
  simpl; by apply IH2.
  - move => h1 h2 t H1 _ H2; apply: Dweak; last eexact H2.
  by apply hyp_subst_idt_subset.
Qed.


Lemma Dinst : forall h c s, wf_substitution_idv s -> deriv h c ->
  deriv (hyp_subst_idv h s) (subst_idv c s).
Proof.
  move => h0 c0 s hs; elim => {h0 c0}.
  - move => t A h /=; apply: Dweak; first apply ssrfsets.empty_subset.
  apply Drefl; apply (wt_subst_idv_inv2 s hs h).
  - move => t H1; apply: Dweak; first apply hyp_subst_idv_singl.
  - apply Dassume; apply (wt_subst_idv_inv2 s hs H1).
  - move => t u X A H1 H2; apply: Dweak; first apply ssrfsets.empty_subset.
  simpl; rewrite (subst_idv_open s hs); apply Dbeta1;
    by [apply (wt_subst_idv_inv2 s hs (g:=X::nil) (t:=t) (A:=A))|
      apply (wt_subst_idv_inv2 s hs (g:=nil) (t:=u) (A:=X))].
  - move => f A B H1 /=; apply: Dweak; first apply ssrfsets.empty_subset.
  apply Deta; by apply (wt_subst_idv_inv2 s hs H1).
  - move => A P x H1 H2 /=; apply: Dweak; first apply ssrfsets.empty_subset.
  apply Deps; by [apply (wt_subst_idv_inv2 s hs
    H1)|apply (wt_subst_idv_inv2 s hs H2)].
  - move => h1 h2 t1 t2 t3 A _ IH1 _ IH2;
    apply: Dweak; first apply hyp_subst_idv_union.
  apply: Dtrans; eassumption.
  - move => h t1 t2 A _ IH /=; by apply Dsym.
  - move => /= h1 h2 a b _ IH1 _ IH2;
    apply: Dweak; first apply hyp_subst_idv_union.
  apply: Deqmp; eassumption.
  - move => /= h1 h2 a b _ IH1 _ IH2;
    apply: Dweak; first apply hyp_subst_idv_union.
  apply: Dmp; eassumption.
  - move => /= h1 h2 a b _ IH1 _ IH2;
    apply: Dweak; first apply hyp_subst_idv_union.
  by apply Dimpantisym.
  - move => /= h1 h2 a b _ IH1 _ IH2;
    apply: Dweak; first apply hyp_subst_idv_union.
  by apply Dconj.
  - move => h a b _ IH; apply: Dconj1; eexact IH.
  - move => /= h a b _ IH; apply: Dconj2; eexact IH.
  - move => h a _ IH; by apply Dnoti.
  - move => /= h a _ IH; by apply Dnote.
  - move => /= h a b H1 _ IH; apply Ddisj1; last done.
  apply (wt_subst_idv_inv2 s hs H1).
  - move => /= h a b H1 _ IH; apply Ddisj2; last done.
  apply (wt_subst_idv_inv2 s hs H1).
  - move => h1 h2 h3 a b c _ IH1 _ IH2 _ IH3;
    apply: Dweak; first apply hyp_subst_idv_union.
  apply: Dweak;
    first (apply ssrfsets.union_subset; apply hyp_subst_idv_union).
  apply: Dweak; first (apply ssrfsets.union_subset;
    apply ssrfsets.union_subset_all; apply hyp_subst_idv_remove).
  by apply Ddisjcases.
  - move => h a H1 _ IH;
    apply: Dweak; first apply hyp_subst_idv_remove.
  apply Dcontr; last done.
  apply (wt_subst_idv_inv2 s hs H1).
  - simpl; apply: Dweak; first apply ssrfsets.empty_subset.
  apply Dtrue.
  - move => h p q H1 H2 H3; apply: Dweak; first apply hyp_subst_idv_remove.
  apply Ddisch; last done.
  exact (wt_subst_idv_inv2 s hs H1).
  - move => /= h1 h2 f g u v A B _ IH1 _ IH2;
    apply: Dweak; first apply hyp_subst_idv_union.
  apply: Dmkcomb; eassumption.
  - move => /= h u v L A X H IH;
    apply (Dabs1 _ _ ((dom_substitution_idv s)++L)) => x H1;
      move /andP: (forallb_app1 _ _ _ H1) => {H1} [H1 H2];
        move: (IH x H2); rewrite !(subst_idv_open _ hs);
          by rewrite (dom_substitution_idv_id _ _ _ H1).
  - move => /= h p t A H1 _ IH; rewrite (subst_idv_open _ hs); apply: Dspec;
    last eassumption.
  by apply (wt_subst_idv_inv2 s hs (g:=nil)).
  - move => /= h t L X _ IH;
    apply (Dgen1 _ ((dom_substitution_idv s)++L)) => x H1;
      move /andP: (forallb_app1 _ _ _ H1) => {H1} [H1 H2];
        move: (IH x H2); rewrite !(subst_idv_open _ hs);
          by rewrite (dom_substitution_idv_id _ _ _ H1).
  - move => /= h a t A H1 _; rewrite (subst_idv_open _ hs) => IH;
    apply: Dexists; last eassumption.
  by apply (wt_subst_idv_inv2 s hs (g:=nil)).
  - move => h1 h2 L a c A _ /= IH1 _ IH2;
    apply: Dweak; first apply hyp_subst_idv_union.
  apply: (Dchoose1 ((dom_substitution_idv s)++L)); first eexact IH1.
  move => x Hx; move /andP: (forallb_app1 _ _ _ Hx) => {Hx} [H1 H2];
    replace (open (subst_idv a s) (Var x A)) with
      (subst_idv (open a (Var x A)) s);
      last by (move: (subst_idv_open s hs a (Var x A));
        move: (dom_substitution_idv_id s x A H1) => /= ->).
  apply: Dweak; first apply hyp_subst_idv_add.
  simpl; by apply IH2.
  - move => h1 h2 t H1 _ H2; apply: Dweak; last eexact H2.
  by apply hyp_subst_idv_subset.
Qed.


(* Fresh variables generation *)

Definition fresh_var := fresh_names.


(* Well-typed judgments *)

Lemma wt_deriv : forall (h: hyp_set) (t: term),
  deriv h t -> wt nil t Bool.
Proof.
  move => h0 t0; elim => {h0 t0} //=.
  - move => t A H; by apply wt_eq_inv2.
  - move => t u X A H1 H2; apply wt_eq_inv2.
  apply: wt_app_inv2; split; last eexact H2.
  by apply wt_abs_inv2.
  apply: wt_open_inv2; eassumption.
  - move => f A B H1; apply wt_eq_inv2; last done.
  apply wt_abs_inv2; apply: wt_app_inv2; split; last done.
  exact (weakening (A::nil) H1).
  - move => A P x H1 H2 /=; apply: wt_app_inv2; split.
  apply: wt_app_inv2; split; first done.
  apply: wt_app_inv2; split; [eassumption|assumption].
  apply: wt_app_inv2; split; first eassumption.
  by apply: wt_app_inv2; split.
  - move => h1 h2 t1 t2 t3 A _ IH1 _ IH2; move: (wt_eq_inv1 IH1) => [H1 _];
    move: (wt_eq_inv1 IH2) => [_ H3]; by apply wt_eq_inv2.
  - move => h t1 t2 A _ IH.
  - move: (wt_eq_inv1 IH) => [H1 H2]; by apply wt_eq_inv2.
  - move => h1 h2 a b _ IH _ _; by move: (wt_hequiv_inv1 IH) => [_ H].
  - move => h1 h2 a b _ IH _ _; by move: (wt_himp_inv1 IH) => [_ H].
  - move => h1 h2 a b _ IH1 _ IH2; by apply wt_hand_inv2.
  - move => h a b _ IH; by move: (wt_hand_inv1 IH) => [H1 H2].
  - move => h a b _ IH; by move: (wt_hand_inv1 IH) => [H1 H2].
  - move => h a _ IH; apply wt_hnot_inv2;
    by move: (wt_himp_inv1 IH) => [H _].
  - move => /= h a _ IH; apply wt_himp_inv2;
    by [move: (wt_hnot_inv1 IH)|].
  - move => /= h a b H1 _ IH; by apply wt_hor_inv2.
  - move => /= h a b H1 _ IH; by apply wt_hor_inv2.
  - move => h p q H1 H2 IH; by apply: wt_himp_inv2.
  - move => h1 h2 f g u v A B _ IH1 _ IH2;
    move: (wt_eq_inv1 IH1) => [H1 H2];
      move: (wt_eq_inv1 IH2) => [H3 H4];
        apply: wt_eq_inv2; by apply (wt_app_inv2 (B:=A)).
  - move => h u v L A X H IH; move: (fresh_var L) => [x Hx];
    move: (IH x Hx) => H1; move: (wt_eq_inv1 H1) => {H1} [H1 H2];
      move: (wt_open_inv1 (g:=nil) (t:=Var x X) u (A:=A) (B:=X)
        (refl_equal (Some X)) H1) => {H1} H1;
      move: (wt_open_inv1 (g:=nil) (t:=Var x X) v (A:=A) (B:=X)
        (refl_equal (Some X)) H2) => {H2} H2;
      by apply wt_eq_inv2; apply wt_abs_inv2.
  - move => h p t A H1 _ IH; move: (wt_hforall_inv1 IH) => H;
    apply: wt_open_inv2; eassumption.
  - move => h t L X _ IH; move: (fresh_var L) => [x Hx];
    move: (IH x Hx) => {IH Hx} IH; move: (wt_open_inv1 (g:=nil)
      (t:=Var x X) _ (B:=X) (refl_equal (Some X)) IH) => {IH} IH;
    by apply wt_hforall_inv2.
  - move => h a t A H1 _ IH; move: (wt_open_inv1 _ H1 IH) => H;
    by apply wt_hexists_inv2.
  - move => h1 h2 L a c A _ _ _ IH2; move: (fresh_var L) => [x Hx];
    by apply (IH2 x).
Qed.


(* Semantics *)

Section Semantics.

  Variable tc: type_context.
  Variable tdt: type_definition_tr.
  Variable se: term_context tc tdt.
  Variable sdt: term_definition_tr tc tdt.


  Definition has_sem (t: term) : Prop :=
    match sem_term nil t with
      | Some (existT Bool evT) => evT tc tdt se sdt (tr_nil tc tdt)
      | _ => False
    end.


  (* Well-formed sets of hypothesis *)

  Definition sem_hyp (h: hyp_set) := ssrfsets.For_all has_sem h.


  Lemma sem_singl1 : forall t,
    sem_hyp (hyp_singl t) -> has_sem t.
  Proof. move => t h; apply (ssrfsets.For_all_singleton1 h). Qed.


  Lemma sem_singl2 : forall t,
    has_sem t -> sem_hyp (hyp_singl t).
  Proof. move => t h; apply (ssrfsets.For_all_singleton2 has_sem h). Qed.


  Lemma sem_empty : sem_hyp hyp_empty.
  Proof. apply ssrfsets.For_all_empty. Qed.


  Lemma sem_union1 : forall h1 h2,
    sem_hyp (hyp_union h1 h2) -> sem_hyp h1 /\ sem_hyp h2.
  Proof. apply ssrfsets.For_all_union1. Qed.


  Lemma sem_remove1 : forall p h,
    sem_hyp (hyp_remove p h) -> has_sem p -> sem_hyp h.
  Proof.
    move => p h h1 h2; apply: ssrfsets.For_all_remove1;
      [eexact h2|eexact h1].
  Qed.


  Lemma sem_add2 : forall a h,
    sem_hyp h -> has_sem a -> sem_hyp (hyp_add a h).
  Proof. move => a h H1 H2; by apply ssrfsets.For_all_add2. Qed.


  Lemma sem_subset : forall h1 h2,
    hyp_subset h1 h2 -> sem_hyp h2 -> sem_hyp h1.
  Proof. move => h1 h2; apply ssrfsets.For_all_subset. Qed.

End Semantics.

Implicit Arguments sem_union1 [tc tdt se sdt h1 h2].
Implicit Arguments sem_remove1 [tc tdt se sdt p h].


Lemma sem_change_se_term : forall tc tdt (se: term_context tc tdt) sdt x X
  t, is_not_free x X t -> has_sem se sdt t -> forall a,
  has_sem (change_se se x X a) sdt t.
Proof.
  rewrite /has_sem => tc tdt se sdt x X t H1 H2 a;
    move: (change_se_not_free x X t nil H1) H2;
      case: (sem_term nil t) => [[[|||] evT]|] //= => H2;
        by rewrite H2.
Qed.


Lemma sem_change_se : forall tc tdt (se: term_context tc tdt) sdt x X h,
  hyp_is_not_free x X h -> sem_hyp se sdt h ->
  forall a, sem_hyp (change_se se x X a) sdt h.
Proof.
  rewrite /hyp_is_not_free /sem_hyp => tc tdt se sdt x X h H1 H2 a;
    apply (ssrfsets.For_all_imp
      (P:=fun t => is_not_free x X t /\ has_sem se sdt t)).
  ssrfsets.for_all_tac; by apply ssrfsets.For_all_and.
  move => t [H3 H4]; by apply sem_change_se_term.
Qed.

Implicit Arguments sem_change_se [tc tdt se sdt h x0].


(* Semantics of HOL derivations *)

Theorem sem_deriv : forall (h: hyp_set) (t: term),
  deriv h t -> forall tc tdt (se: term_context tc tdt)
    (sdt : term_definition_tr tc tdt), sem_hyp se sdt h -> has_sem se sdt t.
Proof.
  move => h0 t0; elim => {h0 t0} //=.

  - rewrite /has_sem => t A H tc tdt se sdt _ /=;
    by move: (wt_sem_inv2 H) => {H} [evT ->]; rewrite !cast_same.

  - move => t H1 tc tdt se sdt H2; by apply sem_singl1.

  - move => t u X A H1 H2 tc tdt se sdt _; rewrite /has_sem /=;
    move: (sem_open nil u t); move: (wt_sem_inv2 H2) => [evU ->];
      move: (wt_sem_inv2 H1) => [evT ->];
        case: (sem_term nil (open t u)) => //= [[B evB]].
  rewrite neq_cast; case e: (_ == _); last done.
  by move /eqP: e evB => -> {B} evB; rewrite !cast_same => ->.

  - move => f A B H1 tc tdt se sdt _;
    move: (wt_sem_inv2 H1) => [evF H2];
      move: (weakening (A::nil) H1) => {H1} /= H1;
        move: (wt_sem_inv2 H1) => {H1} [evE H1];
          move: (sem_weak (A::nil) f nil);
            rewrite /has_sem /= H2 H1 !cast_same => H3;
              extensionality a;
              by rewrite -H3 tr_nil_weak /=.

  -move => A P x H1 H2 tc tdt se sdt _;
    move: (wt_sem_inv2 H1) => {H1} [evP H1];
      move: (wt_sem_inv2 H2) => {H2} [evX H2];
        rewrite /has_sem /= H1 H2 !cast_same /eps => IH;
          apply epsilon_spec;
            by exists (evX tc tdt se sdt (tr_nil tc tdt)).

  - move => h1 h2 t1 t2 t3 A H3 IH1 H5 IH2 tc tdt se sdt H;
    move: (sem_union1 H) => {H} [H1 H2].
      move: (IH1 tc tdt se sdt H1) (IH2 tc tdt se sdt H2);
        move: (wt_eq_inv1 (wt_deriv H3)) => {H3} [H3 H4];
          move: (wt_eq_inv1 (wt_deriv H5)) => {H5} [_ H5];
            move: (wt_sem_inv2 H3) => {H3} [ev1 H3];
              move: (wt_sem_inv2 H4) => {H4} [ev2 H4];
                move: (wt_sem_inv2 H5) => {H5} [ev3 H5];
                  by rewrite /has_sem /= H3 H4 H5 !cast_same => -> ->.

  - move => h t1 t2 A H IH tc tdt se sdt H1; move: (wt_deriv H) => {H} H;
    move: (wt_eq_inv1 H) => {H} [H2 H3];
      move: (wt_sem_inv2 H2) => {H2} [ev2 H2];
        move: (wt_sem_inv2 H3) => {H3} [ev1 H3];
          move: (IH tc tdt se sdt H1);
            by rewrite /has_sem /= H2 H3 !cast_same.

  - move => h1 h2 a b H3 IH1 _ IH2 tc tdt se sdt H;
    move: (sem_union1 H) => {H} [H1 H2]; move: (wt_deriv H3) => {H3} H3;
      move: (wt_hequiv_inv1 H3) => {H3} [H3 H4];
        move: (wt_sem_inv2 H3) => {H3} [evA H3];
          move: (wt_sem_inv2 H4) => {H4} [evB H4];
            move: (IH1 tc tdt se sdt H1) (IH2 tc tdt se sdt H2);
              by rewrite /has_sem /= H3 H4 !cast_same => ->.

  - move => h1 h2 a b H3 IH1 _ IH2 tc tdt se sdt H;
    move: (sem_union1 H) => {H} [H1 H2]; move: (wt_deriv H3) => {H3} H3;
      move: (wt_himp_inv1 H3) => {H3} [H3 H4];
        move: (wt_sem_inv2 H3) => {H3} [evA H3];
          move: (wt_sem_inv2 H4) => {H4} [evB H4];
            move: (IH1 tc tdt se sdt H1) (IH2 tc tdt se sdt H2);
              by rewrite /has_sem /= H3 H4 !cast_same.

  - move => h1 h2 a b H3 IH1 _ IH2 tc tdt se sdt H;
    move: (sem_union1 H) => {H} [H1 H2]; move: (wt_deriv H3) => {H3} H3;
      move: (wt_himp_inv1 H3) => {H3} [H3 H4];
        move: (wt_sem_inv2 H3) => {H3} [evA H3];
          move: (wt_sem_inv2 H4) => {H4} [evB H4];
            move: (IH1 tc tdt se sdt H1) (IH2 tc tdt se sdt H2);
              rewrite /has_sem /= H3 H4 !cast_same => H5 H6.
                by apply prop_extensionality.


  - move => h1 h2 a b H1 IH1 H2 IH2 tc tdt se sdt Hh;
    move: (wt_sem_inv2 (wt_deriv H1)) => {H1} [evA H1];
      move: (wt_sem_inv2 (wt_deriv H2)) => {H2} [evB H2];
        move: (sem_union1 Hh) => {Hh} [Hh1 Hh2];
          move: (IH1 tc tdt se sdt Hh1) (IH2 tc tdt se sdt Hh2);
            by rewrite /has_sem /= H1 H2 !cast_same.

  - move => h a b H IH tc tdt se sdt Hh;
    move: (wt_hand_inv1 (wt_deriv H)) => {H} [H1 H2];
      move: (wt_sem_inv2 H1) => {H1} [evA H1];
        move: (wt_sem_inv2 H2) => {H2} [evB H2];
          move: (IH tc tdt se sdt Hh);
            by rewrite /has_sem /= H1 H2 !cast_same => [[H _]].

  - move => h a b H IH tc tdt se sdt Hh;
    move: (wt_hand_inv1 (wt_deriv H)) => {H} [H1 H2];
      move: (wt_sem_inv2 H1) => {H1} [evA H1];
        move: (wt_sem_inv2 H2) => {H2} [evB H2];
          move: (IH tc tdt se sdt Hh);
            by rewrite /has_sem /= H1 H2 !cast_same => [[_ H]].

  - move => h a H IH tc tdt se sdt Hh;
    move: (wt_himp_inv1 (wt_deriv H)) => {H} [H _];
      move: (wt_sem_inv2 H) => {H} [evA H];
        move: (IH tc tdt se sdt Hh);
          by rewrite /has_sem /= H !cast_same.

  - move => h a H IH tc tdt se sdt Hh;
    move: (wt_sem_inv2 (wt_hnot_inv1 (wt_deriv H))) => {H} [evA H];
      move: (IH tc tdt se sdt Hh);
        by rewrite /has_sem /= H !cast_same.

  - move => h a b H1 H IH tc tdt se sdt Hh;
    move: (wt_sem_inv2 (wt_deriv H)) => {H} [evA H];
      move: (wt_sem_inv2 H1) => {H1} [evB H1];
        move: (IH tc tdt se sdt Hh);
          rewrite /has_sem /= H H1 !cast_same => H2; by left.

  - move => h a b H1 H IH tc tdt se sdt Hh;
    move: (wt_sem_inv2 (wt_deriv H)) => {H} [evA H];
      move: (wt_sem_inv2 H1) => {H1} [evB H1];
        move: (IH tc tdt se sdt Hh);
          rewrite /has_sem /= H H1 !cast_same => H2; by right.

  - move => h1 h2 h3 a b c H1 IH1 _ IH2 H3 IH3 tc tdt se sdt Hh;
    move: (wt_hor_inv1 (wt_deriv H1)) => {H1} [H1 H2];
      move: (wt_sem_inv2 H1) => {H1} [evA H1];
        move: (wt_sem_inv2 H2) => {H2} [evB H2];
          move: (wt_sem_inv2 (wt_deriv H3)) => {H3} [evC H3];
            move: (sem_union1 Hh) => {Hh} [Hh1 Hh2];
              move: (sem_union1 Hh2) => {Hh2} [Hh2 Hh3];
                move: (IH1 tc tdt se sdt Hh1);
                  rewrite {1}/has_sem /= H1 H2 !cast_same => [[H|H]].
  apply IH2; apply: sem_remove1; by [eexact Hh2|rewrite /has_sem /= H1].
  apply IH3; apply: sem_remove1; by [eexact Hh3|rewrite /has_sem /= H2].

  - move => h a H1 _ IH tc tdt se sdt H2;
    move: (wt_sem_inv2 H1) => {H1} [evA H1]; apply NNPP => H4;
      have H3: sem_hyp se sdt h.
  apply: sem_remove1; first eexact H2.
  move: H4; by rewrite /has_sem /= H1.
  move: (IH tc tdt se sdt H3); by rewrite /has_sem /=.

  - move => h p q H1 H IH tc tdt se sdt H2;
    move: (wt_sem_inv2 (wt_deriv H)) => {H} [evQ H];
      move: (wt_sem_inv2 H1) => {H1} [evP H1]; move: (sem_remove1 H2);
        rewrite /has_sem /= H H1 !cast_same => H3 H4;
          move: (IH tc tdt se sdt (H3 H4)); by rewrite /has_sem /= H.

  - move => h1 h2 f g u v A B H3 IH1 H4 IH2 tc tdt se sdt H;
    move: (sem_union1 H) => [H1 H2];
      move: (wt_eq_inv1 (wt_deriv H3)) => {H3} [H3 H5];
        move: (wt_eq_inv1 (wt_deriv H4)) => {H4} [H4 H6];
          move: (wt_sem_inv2 H3) => {H3} [evF H3];
            move: (wt_sem_inv2 H4) => {H4} [evU H4];
              move: (wt_sem_inv2 H5) => {H5} [evG H5];
                move: (wt_sem_inv2 H6) => {H6} [evV H6];
                  move: (IH1 tc tdt se sdt H1) (IH2 tc tdt se sdt H2);
                    by rewrite /has_sem /= H3 H4 H5 H6 !cast_same => -> ->.

  - move => h u v L A X H IH tc tdt se sdt Hh;
    move: (fresh_var ((frees_idv u)++(frees_idv v)++L++(hyp_frees_idv h)))
      => [x Hx];
      move /andP: (forallb_app1 _ _ _ Hx) => {Hx} [H1 H2];
        move /andP: (forallb_app1 _ _ _ H2) => {H2} [H2 H3];
          move /andP: (forallb_app1 _ _ _ H3) => {H3} [H3 H6];
            move: (frees_idv_is_not_free2 _ _ H1 X) => {H1} H1;
              move: (frees_idv_is_not_free2 _ _ H2 X) => {H2} H2;
                move: (free_forallb2 _ _ H6 X) => {H6} H6;
                  move: (wt_deriv (H x H3)) => {H} H;
                    move: (wt_eq_inv1 H) => {H} [H4 H5];
                      move: (wt_sem_inv2 H4) => [evB H7];
                        move: (wt_sem_inv2 H5) => [evC H8];
                          move: (wt_open_inv1 (g:=nil) (t:=Var x X) _ (B:=X)
                            (refl_equal (Some X)) H4) => {H4} H4;
                          move: (wt_open_inv1 (g:=nil) (t:=Var x X) _ (B:=X)
                            (refl_equal (Some X)) H5) => {H5} H5;
                          move: (wt_sem_inv2 H4) => {H4} [evU H4];
                            move: (wt_sem_inv2 H5) => {H5} [evV H5];
                              rewrite /has_sem /= H4 H5 !cast_same;
                                extensionality a; move: (IH x H3 tc tdt
                                  (change_se se x X a) sdt (sem_change_se
                                    x X H6 Hh a)) (sem_open nil (Var x X) u)
                                (sem_open nil (Var x X) v); rewrite
                                  /has_sem /= H4 H5 H7 H8 !cast_same =>
                                  {H7 H8} H7 H8 H9;
    move: H7; rewrite H8 H9 change_se_id => H7;
      move: (change_se_not_free _ _ _ (X::nil) H1)
        (change_se_not_free _ _ _ (X::nil) H2);
        rewrite H4 H5 => H10 H11;
          move: H7; by rewrite H10 H11.

  - move => h p t A H1 H IH tc tdt se sdt H2;
    move: (wt_sem_inv2 (wt_hforall_inv1 (wt_deriv H))) => {H} [evP H];
      move: (wt_sem_inv2 H1) => {H1} [evT H1];
        move: (IH tc tdt se sdt H2) (sem_open nil t p);
          rewrite /has_sem /= H H1 !cast_same H => H3;
            case: (sem_term nil (open p t)) => [[[|||] evB]|] //= => H4.
  rewrite H4; apply H3.

  - move => h t L X H IH tc tdt se sdt Hh;
    move: (fresh_var ((frees_idv t)++L++(hyp_frees_idv h))) => [x Hx];
      move /andP: (forallb_app1 _ _ _ Hx) => [H1 H2];
        move /andP: (forallb_app1 _ _ _ H2) => {H2} [H2 H5];
          move: (frees_idv_is_not_free2 _ _ H1 X) => {H1} H1;
           move: (free_forallb2 _ _ H5 X) => {H5} H5;
             move: (wt_deriv (H x H2)) => H3;
               move: (wt_sem_inv2 H3) => [evB H8];
                 move: (wt_open_inv1 (g:=nil) (t:=Var x X) _ (B:=X)
                   (refl_equal (Some X)) H3) => H4;
                 move: (wt_sem_inv2 H4) => {Hx H3 H4} [evT H3];
                   rewrite /has_sem /= H3 cast_same => a;
                     move: (IH x H2 tc tdt (change_se se x X a) sdt
                       (sem_change_se x X H5 Hh a))
                     (sem_open nil (Var x X) t);
                   rewrite /has_sem /= H3 H8 !cast_same => H4 H6;
                     move: H4; rewrite H6 change_se_id => H4;
                       move: (change_se_not_free _ _ _ (X::nil) H1);
                         rewrite H3 => H7; by rewrite H7 in H4.

  - move => h a t A H1 H IH tc tdt se sdt Hh; move: (wt_deriv H) => {H} H;
    move: (wt_open_inv1 _ H1 H) => H3;
      move: (wt_sem_inv2 H3) => {H3} [evA H2];
        move: (wt_sem_inv2 H1) => {H1} [evT H1];
          move: (wt_sem_inv2 H) => {H} [evB H];
            move: (IH tc tdt se sdt Hh) (sem_open nil t a);
              rewrite /has_sem /= H H1 H2 !cast_same => H4 H5;
                move: H4; rewrite H5 => H4;
                  by exists (evT tc tdt se sdt (tr_nil tc tdt)).
  - move => h1 h2 L a c A H1 IH1 H2 IH2 tc tdt se sdt Hh;
    move: (wt_deriv H1) => {H1} H1;
      move: (wt_sem_inv2 (wt_hexists_inv1 H1)) => {H1} [evA H1];
        move: (fresh_var ((frees_idv c)++(hyp_frees_idv h2)++L++
          (frees_idv a))) => [x Hx];
        move /andP: (forallb_app1 _ _ _ Hx) => {Hx} [H3 H4];
          move /andP: (forallb_app1 _ _ _ H4) => {H4} [H4 H5];
            move /andP: (forallb_app1 _ _ _ H5) => {H5} [H5 H8];
              move: (frees_idv_is_not_free2 _ _ H3 A) => {H3} H3;
                move: (frees_idv_is_not_free2 _ _ H8 A) => {H8} H8;
                  move: (free_forallb2 _ _ H4 A) => {H4} H4;
                    move: (wt_sem_inv2 (wt_deriv (H2 x H5))) => {H2} [evC
                      H2]; move: (sem_union1 Hh) => {Hh} [Hh1 Hh2];
                    move: (sem_open nil (Var x A) a);
                      rewrite /has_sem /= H1 H2; case e: (sem_term nil
                        (open a (Var x A))) => [[[|||] evD]|] //= => H6.
  move: (IH1 tc tdt se sdt Hh1);
    rewrite /has_sem /= H1 !cast_same => [[b H7]];
      move: (change_se_not_free _ _ _ nil H3); rewrite H2 => H;
        rewrite -(H _ _ _ _ _ b) => {H};
          move: (change_se_not_free _ _ _ (A::nil) H8); rewrite H1 => H;
            replace b with ((change_se se x A b) x A) in H7;
              last by rewrite /change_se /= cast_same eq_refl.
  rewrite -(H _ _ _ _ _ b) in H7 => {H}; rewrite -H6 in H7;
    have H9: (sem_hyp (change_se se x A b) sdt (hyp_add (open a (Var x A))
      h2)).
  apply sem_add2; first by apply sem_change_se.
  by rewrite /has_sem /= e.
  move: (IH2 _ H5 _ _ _ _ H9); by rewrite /has_sem /= H2.

  - move => h1 h2 t H1 H2 IH tc tdt se sdt H3; apply IH;
    apply: sem_subset; by [eexact H1|].
Qed.

Unset Implicit Arguments.
