(* ========================================================================= *)
(* Representing HOL-Light types and terms                                    *)
(*                                                                           *)
(*       Chantal Keller, Laboratoire d'Informatique de Polytechnique         *)
(*                                                                           *)
(*    (c) Copyright, Laboratoire d'Informatique de Polytechnique 2010        *)
(*              (c) Copyright, Chantal Keller 2010                           *)
(* ========================================================================= *)


Require Import ssreflect ssrbool eqtype ssrnat.
Require Import Arith List.
Require Import ssrpositive.

Set Implicit Arguments.


(* Identifiers for type and term variables *)

Definition idT : eqType := names_dom.
Definition defT : eqType := names_dom.

Definition idV : eqType := names_dom.
Definition defV : eqType := names_dom.


(* HOL types *)

Inductive type : Type :=
| TVar : idT -> type
| TDef : defT -> list_type -> type
| Bool : type
| Arrow : type -> type -> type

with list_type : Type :=
| Tnil : list_type
| Tcons : type -> list_type -> list_type.


Scheme type_ltype_ind := Induction for type Sort Prop
with ltype_type_ind := Induction for list_type Sort Prop.

Combined Scheme ltype_ind from type_ltype_ind, ltype_type_ind.

Scheme type_ltype_rec := Induction for type Sort Type
with ltype_type_rec := Induction for list_type Sort Type.


Notation "t1 '-->' t2" := (Arrow t1 t2) (right associativity, at level 55).


(* HOL constants *)

Inductive cst : Type :=
| Heq : type -> cst
| Heps : type -> cst
| Hand : cst
| Hor : cst
| Hnot : cst
| Himp : cst
| Htrue : cst
| Hfalse : cst
| Hforall : type -> cst
| Hexists : type -> cst.

(* HOL terms *)

Inductive term : Type :=
| Dbr : nat -> term
| Var : idV -> type -> term
| Cst : cst -> term
| Def : defV -> type -> term
| App : term -> term -> term
| Abs : type -> term -> term.


(* Type equality is decidable *)

Lemma type_eq_dec : forall (A B: type), {A = B} + {A <> B}.
Proof.
  apply (type_ltype_rec (fun A => forall B, {A = B} + {A <> B})
    (fun l1 => forall l2, {l1 = l2} + {l1 <> l2})) => [X [Y|||]|X l IHl [|Y l2||]|[|||]|A IHA B IHB [|||C D]|[|]|A IHA l IHl [|B l2]];
  try (right; discriminate); try by left.
  - case e: (X == Y).
  by left; move /eqP: e => ->.
  by right => h; move: h e => [->] e; move: (eq_refl Y); rewrite e.
  - case e: (X == Y).
  move /eqP: e => <- {Y}; move: (IHl l2) => {IHl} [->|e]; first by left.
  by right => h; apply e; move: h => [->].
  by right => h; move: h e (eq_refl Y) => [-> _] ->.
  - move: (IHA C) => {IHA} [->|hA].
  move: (IHB D) => {IHB} [->|hB]; first by left.
  by right => h; apply hB; move: h => [->].
  by right => h; apply hA; move: h => [->].
  - move: (IHA B) => {IHA} [->|e].
  move: (IHl l2) => {IHl} [->|el]; first by left.
  by right => h; apply el; move: h => [->].
  by right => h; apply e; move: h => [->].
Defined.


(* Constants equality is decidable *)

Lemma cst_eq_dec : forall (c d: cst), {c = d} + {c <> d}.
Proof.
  move => [A|A|||||||A|A] [B|B|||||||B|B];
    try (right => h; discriminate h); try (left; reflexivity).
  move: (type_eq_dec A B) => [->|h]; first by left.
  by right => h1; apply h; move: h1 => [->].
  move: (type_eq_dec A B) => [->|h]; first by left.
  by right => h1; apply h; move: h1 => [->].
  move: (type_eq_dec A B) => [->|h]; first by left.
  by right => h1; apply h; move: h1 => [->].
  move: (type_eq_dec A B) => [->|h]; first by left.
  by right => h1; apply h; move: h1 => [->].
Defined.


(* Term equality is decidable *)

Lemma term_eq_dec : forall (u v: term), {u = v} + {u <> v}.
Proof.
  elim => [n|x A|c|x A|t1 IHt1 t2 IHt2|A t IHt] [m|y B|d|y B|u1 u2|B u];
    try (right => h; discriminate h).
  - move: (eq_nat_dec n m) => [->|h]; first by left.
  by right => h1; apply h; move: h1 => [->].
  - case e: (x == y); last by right => h; move: h e => [-> _] => e;
    move: (eq_refl y); rewrite e.
  move: (type_eq_dec A B) => [->|h]; first by move /eqP: e => ->; left.
  by right => h1; apply h; move: h1 => [_ ->].
  - move: (cst_eq_dec c d) => [->|h]; first by left.
  by right => h1; apply h; move: h1 => [->].
  - case e: (x == y); last by right => h; move: h e => [-> _] => e;
    move: (eq_refl y); rewrite e.
  move: (type_eq_dec A B) => [->|h]; first by move /eqP: e => ->; left.
  by right => h1; apply h; move: h1 => [_ ->].
  - move: (IHt1 u1) => {IHt1} [->|IHt1].
  move: (IHt2 u2) => {IHt2} [->|IHt2]; first by left.
  by right => h; apply IHt2; move: h => [->].
  by right => h; apply IHt1; move: h => [->].
  - move: (IHt u) => {IHt} [->|IHt].
  move: (type_eq_dec A B) => [->|h]; first by left.
  by right => h1; apply h; move: h1 => [->].
  by right => h; apply IHt; move: h => [_ ->].
Defined.


Unset Implicit Arguments.
