(* ========================================================================= *)
(* Supplementary definitions and properties about Coq's lists                *)
(*                                                                           *)
(*       Chantal Keller, Laboratoire d'Informatique de Polytechnique         *)
(*                                                                           *)
(*    (c) Copyright, Laboratoire d'Informatique de Polytechnique 2010        *)
(*              (c) Copyright, Chantal Keller 2010                           *)
(* ========================================================================= *)


Require Import ssreflect ssrbool eqtype.
Require Import List SetoidList.

Set Implicit Arguments.


(* Propositional for all *)

Fixpoint forallp A P (l: list A) : Prop :=
  match l with
    | nil => True
    | t::q => P t /\ forallp P q
  end.


(* Properties of append *)

Lemma forallb_app1 : forall A P (m l: list A),
  forallb P (l ++ m) -> ((forallb P l) && (forallb P m)).
Proof.
  move => A P m; elim => [|t q IHq] //=.
  - move /andP => [H1 H2]; rewrite -andb_assoc; apply andb_true_intro;
    split; by [|apply IHq].
Qed.


Lemma forallb_app2 : forall A P (m l: list A),
  ((forallb P l) && (forallb P m)) -> forallb P (l ++ m).
Proof.
  move => A P m; elim => [|t q IHq] //=.
  - rewrite -andb_assoc; move /andP => [H1 H2];
    apply andb_true_intro; split; by [|apply IHq].
Qed.


(* Flatten *)

Fixpoint flatten (A: Type) (l: list (list A)) : list A :=
  match l with
    | nil => nil
    | t::q => t++(flatten q)
  end.


Lemma forallb_flatten1 : forall A P (l: list (list A)),
  forallb P (flatten l) -> forallb (forallb P) l.
Proof.
  move => A P; elim => [|t q IHq] //=.
  - move => H; move /andP: (forallb_app1 _ _ _ H) => [H1 H2];
    apply andb_true_intro; split; by [|apply IHq].
Qed.


Lemma forallb_flatten2 : forall A P (l: list (list A)),
  forallb (forallb P) l -> forallb P (flatten l).
Proof.
  move => A P; elim => [|t q IHq] //=.
  - move /andP => [H1 H2]; apply forallb_app2;
    apply andb_true_intro; split; by [|apply IHq].
Qed.


(* Properties of map *)

Lemma forallb_map1 : forall (A: eqType) B P (f: A -> B) l,
  forallb P (map f l) -> forall a, InA (fun x y => x == y) a l -> P (f a).
Proof.
  move => A B P f; elim => [|t q IHq] /=.
  - move => _ a h; move: (InA_nil (fun x y => x == y) a) => [H1 _]; by elim H1.
  - move /andP => [H1 H2] a H; move: (InA_cons (fun x y => x == y) a t q) => [H3 _];
    move: (H3 H) => {H3 H} [H4|H]; first by move /eqP: H4 ->.
  by apply IHq.
Qed.

Lemma forallb_map2 : forall (A: eqType) B (P: B -> bool) (f: A -> B) l,
  (forall a, InA (fun x y => x == y) a l -> P (f a)) -> forallb P (map f l).
Proof.
  move => A B P f; elim => [|t q IHq] //=.
  - move => H; apply andb_true_intro; split.
  apply H; by constructor.
  apply IHq => a H1; apply H; by apply InA_cons_tl.
Qed.


(* Implication *)

Lemma forallb_imp : forall A (P Q: A -> bool), (forall x, P x -> Q x) ->
  forall L, forallb P L -> forallb Q L.
Proof.
  move => A P Q H; elim => [|t q IHq] //=.
  move /andP => [H1 H2]; apply andb_true_intro; split;
    by [apply H|apply IHq].
Qed.

Unset Implicit Arguments.
