(* ========================================================================= *)
(* Representing HOL-Light's proofs skeletons                                 *)
(*                                                                           *)
(*       Chantal Keller, Laboratoire d'Informatique de Polytechnique         *)
(*                                                                           *)
(*    (c) Copyright, Laboratoire d'Informatique de Polytechnique 2010        *)
(*              (c) Copyright, Chantal Keller 2010                           *)
(* ========================================================================= *)


Require Import ssreflect ssrbool eqtype ssrnat.
Require Import list_ext hol cast typing translation subst_db substitution
  deriv.
Require Import List NArith.

Open Local Scope positive_scope.

Set Implicit Arguments.


(* Dedicated sigma type *)

Inductive my_sig : Type :=
  my_exist : forall h t, deriv h t -> my_sig.


(* Proofs *)
(* For a T substitution, the first T is the redex and the second one is
   the residue *)

Inductive proof : Type :=
| Prefl : term -> proof
| Pbeta : idV -> type -> term -> proof
| Pbeta1 : term -> term -> proof
| Peta : term -> proof
| Peps : term -> term -> proof
| Pinstt : proof -> list (idT * type) -> proof
| Pabs : proof -> idV -> type -> proof
| Pdisch : proof -> term -> proof
| Phyp : term -> proof
| Pspec : proof -> term -> proof
| Pinst : proof -> list (idV * type * term) -> proof
| Pgen : proof -> idV -> type -> proof
| Psym : proof -> proof
| Ptrans : proof -> proof -> proof
| Pcomb : proof -> proof -> proof
| Peqmp : proof -> proof -> proof
| Pmp : proof -> proof -> proof
| Pexists : proof -> term -> term -> proof
| Pchoose : idV -> type -> proof -> proof -> proof
| Pconj : proof -> proof -> proof
| Pconjunct1 : proof -> proof
| Pconjunct2 : proof -> proof
| Pdisj1 : proof -> term -> proof
| Pdisj2 : proof -> term -> proof
| Pdisjcases : proof -> proof -> proof -> proof
| Pnoti : proof -> proof
| Pnote : proof -> proof
| Pcontr : proof -> term -> proof
| Ptrue : proof
| Pimpas : proof -> proof -> proof
| Poracle : my_sig -> proof.


Fixpoint proof2deriv (p: proof) : option (hyp_set * term) :=
  match p with

    | Prefl t =>
      match infer nil t with
        | Some A => Some (hyp_empty, heq A t t)
        | None => None
      end

    | Pbeta x X t =>
      match infer nil t with
        | Some A => Some (hyp_empty,
          heq A (App (Abs X (close t x X)) (Var x X)) t)
        | None => None
      end

    | Pbeta1 t u =>
      match infer nil u with
        | Some X => match infer (X::nil) t with
                      | Some A => Some (hyp_empty, heq A (App (Abs X t) u)
                        (open t u))
                      | None => None
                    end
        | None => None
      end

    | Peta f =>
      match infer nil f with
        | Some (Arrow A B) =>
          Some (hyp_empty, heq (A --> B) (Abs A (App f (Dbr 0))) f)
        | _ => None
      end

    | Peps P x =>
      match infer nil P, infer nil x with
        | Some (Arrow A Bool), Some B =>
          if (A == B) then
            Some (hyp_empty, (App P x) himp (App P (App (Cst (Heps A)) P)))
            else None
        | _, _ => None
      end

    | Pinstt q l =>
      match proof2deriv q with
        | Some (h,v) => Some (hyp_subst_idt h l, subst_idt v l)
        | None => None
      end

    | Pabs q x X =>
      match proof2deriv q with
        | Some (h, t) =>
          match t with
            | App (App (Cst (Heq A)) t1) t2 =>
              if hyp_is_not_free x X h then Some (h, heq (X --> A)
                (Abs X (close t1 x X)) (Abs X (close t2 x X))) else None
            | _ => None
          end
        | None => None
      end

    | Pdisch q t =>
      match proof2deriv q, infer nil t with
        | Some (h, u), Some Bool => Some (hyp_remove t h, t himp u)
        | _, _ => None
      end

    | Phyp t =>
      match infer nil t with
        | Some Bool => Some (hyp_singl t, t)
        | _ => None
      end

    | Pspec q t =>
      match proof2deriv q, infer nil t with
        | Some (h, u), Some A =>
          match u with
            | App (Cst (Hforall B)) (Abs C v) =>
              if ((A == B) && (B == C)) then Some (h, open v t) else None
            | _ => None
          end
        | _, _ => None
      end

    | Pinst q l =>
      match proof2deriv q, wf_substitution_idv l with
        | Some (h, v), true => Some (hyp_subst_idv h l, subst_idv v l)
        | _, _ => None
      end

    | Pgen q x X =>
      match proof2deriv q with
        | Some (h, t) => if hyp_is_not_free x X h
          then Some (h, hforall X (close t x X)) else None
        | None => None
      end

    | Psym q =>
      match proof2deriv q with
        | Some (h, t) =>
          match t with
            | App (App (Cst (Heq A)) u) v => Some (h, heq A v u)
            | _ => None
          end
        | None => None
      end

    | Ptrans q1 q2 =>
      match proof2deriv q1, proof2deriv q2 with
        | Some (h1, t1), Some (h2, t2) =>
          match t1, t2 with
            | App (App (Cst (Heq A)) u1) u2,
              App (App (Cst (Heq B)) v2) v3 =>
              if (A == B) && (u2 == v2)
                then Some (hyp_union h1 h2, heq A u1 v3) else None
            | _, _ => None
          end
        | _, _ => None
      end

    | Pcomb q1 q2 =>
      match proof2deriv q1, proof2deriv q2 with
        | Some (h1, t1), Some (h2, t2) =>
          match t1, t2 with
            | App (App (Cst (Heq (Arrow A B))) f) g,
              App (App (Cst (Heq C)) u) v =>
              if A == C
                then Some (hyp_union h1 h2, heq B (App f u) (App g v))
                else None
            | _, _ => None
          end
        | _, _ => None
      end

    | Peqmp q1 q2 =>
      match proof2deriv q1, proof2deriv q2 with
        | Some (h1, t1), Some (h2, t2) =>
          match t1 with
            | App (App (Cst (Heq Bool)) a) b =>
              if a == t2 then Some (hyp_union h1 h2, b) else None
            | _ => None
          end
        | _, _ => None
      end

    | Pmp q1 q2 =>
      match proof2deriv q1, proof2deriv q2 with
        | Some (h1, t1), Some (h2, t2) =>
          match t1 with
            | App (App (Cst Himp) a) b =>
              if a == t2 then Some (hyp_union h1 h2, b) else None
            | _ => None
          end
        | _, _ => None
      end

    | Pexists q b t =>
      match proof2deriv q, b, infer nil t with
        | Some (h, u), Abs B a, Some A =>
          if ((A == B) && (open a t == u))
            then Some (h, hexists A a) else None
        | _, _, _ => None
      end

    | Pchoose v A q1 q2 =>
      match proof2deriv q1, proof2deriv q2 with
        | Some (h1, t), Some (h2, c) =>
          match t with
            | App (Cst (Hexists B)) (Abs C a) =>
              let s := (hyp_remove (open a (Var v A)) h2) in
                if ((A == B) && (B == C) && (hyp_is_not_free v A s) && (is_not_free v A c)
                  && (is_not_free v A a))
                  then Some (hyp_union h1 s, c) else None
            | _ => None
          end
        | _, _ => None
      end

    | Pconj q1 q2 =>
      match proof2deriv q1, proof2deriv q2 with
        | Some (h1, a), Some (h2, b) =>
          Some (hyp_union h1 h2, a hand b)
        | _, _ => None
      end

    | Pconjunct1 q =>
      match proof2deriv q with
        | Some (h, v) =>
          match v with
            | App (App (Cst Hand) t) u =>
              Some (h, t)
            | _ => None
          end
        | _ => None
      end

    | Pconjunct2 q =>
      match proof2deriv q with
        | Some (h, v) =>
          match v with
            | App (App (Cst Hand) t) u =>
              Some (h, u)
            | _ => None
          end
        | _ => None
      end

    | Pdisj1 q b =>
      match proof2deriv q, infer nil b with
        | Some (h, a), Some Bool =>
          Some (h, a hor b)
        | _, _ => None
      end

    | Pdisj2 q a =>
      match proof2deriv q, infer nil a with
        | Some (h, b), Some Bool =>
          Some (h, a hor b)
        | _, _ => None
      end

    | Pdisjcases q1 q2 q3 =>
      match proof2deriv q1, proof2deriv q2, proof2deriv q3 with
        | Some (h1, t), Some (h2, c1), Some (h3, c2) =>
          match t with
            | App (App (Cst Hor) a) b =>
              if c1 == c2
                then Some (hyp_union h1 (hyp_union (hyp_remove a h2)
                  (hyp_remove b h3)), c1) else None
            | _ => None
          end
        | _, _, _ => None
      end

    | Pnoti q =>
      match proof2deriv q with
        | Some (h, t) =>
          match t with
            | App (App (Cst Himp) a) (Cst Hfalse) => Some (h, hnot a)
            | _ => None
          end
        | _ => None
      end

    | Pnote q =>
      match proof2deriv q with
        | Some (h, t) =>
          match t with
            | App (Cst Hnot) a => Some (h, a himp hfalse)
            | _ => None
          end
        | _ => None
      end

    | Pcontr q a =>
      match proof2deriv q, infer nil a with
        | Some (h, t), Some Bool =>
          match t with
            | Cst Hfalse =>
              Some (hyp_remove (hnot a) h, a)
            | _ => None
          end
        | _, _ => None
      end

    | Ptrue => Some (hyp_empty, htrue)

    | Pimpas q1 q2 =>
      match proof2deriv q1, proof2deriv q2 with
        | Some (h1, t), Some (h2, u) =>
          match t, u with
            | App (App (Cst Himp) a1) b1,
              App (App (Cst Himp) b2) a2 =>
              if ((a1 == a2) && (b1 == b2))
                then Some (hyp_union h1 h2, b1 hequiv a1) else None
            | _, _ => None
          end
        | _, _ => None
      end

    | Poracle (my_exist h t _) => Some (h, t)
(*     | Poracle (my_exist h t _) => Some (h(TermSet.this), t) *)

  end.


Lemma proof2deriv_correct : forall p,
  match proof2deriv p with
    | Some (h,t) => deriv h t
    | None => True
  end.
Proof.
  elim => //=.
  - move => t; case e: (infer _ _) => [A|]; last done.
  apply: Dweak; last by apply Drefl. apply ssrfsets.empty_subset.
  - move => s A t; case e: (infer _ _) => [B|]; last done.
  apply: Dweak; last by apply Dbeta. apply ssrfsets.empty_subset.
  - move => t u; case e: (infer _ _) => [X|]; last done.
  case f: (infer _ _) => [A|]; last done.
  apply: Dweak; last by apply Dbeta1. apply ssrfsets.empty_subset.
  - move => f; case e: (infer _ _) => [[|||A B]|] //=;
    apply: Dweak; last by apply Deta. apply ssrfsets.empty_subset.
  - move => P x; case e: (infer _ _) => [[|||A B]|] //=;
    move: e; case: B => //= e; case f: (infer _ _) => [B|] //=;
      case g: (_ == _); last done; move: e; move /eqP: g -> => e;
        apply: Dweak; last by apply Deps. apply ssrfsets.empty_subset.
  - move => p; case: (proof2deriv _) => [[h t]|]; last done.
  move => H1 l; apply: Dweak; last (apply Dinsttype; eauto).
  apply hyp_subst_idt_subset; apply ssrfsets.refl_subset.
  - move => p; case: (proof2deriv _) =>
    [[h [||||[||||[||[A|||||||||]|||] v|] u|]]|] //=.
  move => H1 s t; case e: (hyp_is_not_free _ _ _); last done.
  by apply Dabs.
  - move => p; case: (proof2deriv _) => [[h t]|]; last done.
  move => H1 u; case e: (infer _ _) => [[|||]|] //=. by apply Ddisch.
  - move => t; case e: (infer _ _) => [[|||]|] //=. by apply Dassume.
  - move => p; case: (proof2deriv _) => [[h t]|]; last done.
  move => H1 u; case e1: (infer _ _) => [A|]; last done.
  move: t H1 => [||||t v|] //=. move: t => [||[||||||||B|]|||] //=.
  move: v => [|||||C v] //= H1. case e: (_ == _); last done.
  case f: (_ == _); last done. simpl; apply: Dspec; first eexact e1.
  by move: H1; move /eqP: e ->; move /eqP: f ->.
  - move => p; case: (proof2deriv _) => [[h t]|]; last done.
  move => H1 l; case e: (wf_substitution_idv _); last done. by apply Dinst.
  - move => p; case: (proof2deriv _) => [[h t]|]; last done.
  move => H1 s u; case e: (hyp_is_not_free _ _ _); last done. by apply Dgen.
  - move => p; case: (proof2deriv _) =>
    [[h [||||[||||[||[A|||||||||]|||] u|] v|]]|] //=.
  move => H1; by apply Dsym.
  - move => p; case: (proof2deriv _) => [[h t]|]; last done.
  move => H1 q; case: (proof2deriv _) => [[h2 u]|]; last done.
  move: t H1 => [||||[||||[||[A|||||||||]|||] u1|] u2|] //=.
  move: u => [||||[||||[||[B|||||||||]|||] v2|] v3|] //= => H1 H2.
  case e: (_ == _); last done. case f: (_ == _); last done.
  simpl; move : H1 H2; move /eqP: e ->; move /eqP: f -> => H1 H2;
    apply: Dtrans; by [eexact H1|].
  - move => p; case: (proof2deriv _) => [[h t]|] //=.
  move => H1 q; case: (proof2deriv _) => [[h2 u]|] //=.
  move: t H1 => [||||[||||[||[[|||A B]|||||||||]|||] f|] g|] //=.
  move: u => [||||[||||[||[C|||||||||]|||] u|] v|] //=.
  move => H1 H2; case e: (_ == _); last done.
  apply: Dmkcomb; first eexact H1. by move /eqP: e ->.
  - move => p; case: (proof2deriv _) => [[h t]|]; last done.
  move => H1 q; case: (proof2deriv _) => [[h2 u]|]; last done.
  move: t H1 => [||||[||||[||[[|||]|||||||||]|||] t|] v|] //=.
  move => H1 H2; case e: (_ == _); last done. apply: Deqmp; first eexact H1.
  by move /eqP: e ->.
  - move => p; case: (proof2deriv _) => [[h t]|]; last done.
  move => H1 q; case: (proof2deriv _) => [[h2 u]|]; last done.
  move: t H1 => [||||[||||[||[[|||]|||||||||]|||] t|] v|] //=.
  move => H1 H2; case e: (_ == _); last done. apply: Dmp; first eexact H1.
  by move /eqP: e ->.
  - move => p; case: (proof2deriv _) => [[h t]|] //=.
  move => H1 [|||||B u] v //=.
  case e: (infer _ _) => [A|] //=.
  case f1: (_ == _); last done. case f2: (_ == _); last done.
  simpl. apply: Dexists; first eexact e. by move /eqP: f2 ->.
  - move => s A p; case: (proof2deriv _) => [[h t]|]; last done.
  move => H1 q; case: (proof2deriv _) => [[h2 u]|]; last done.
  move: t H1 => [||||[||[|||||||||B]|||] t|] //=.
  move: t => [|||||C a] //=. move => H1 H2; case e: (_ == _); last done.
  case f: (_ == _); last done. case e1: (hyp_is_not_free _ _ _); last done.
  case e2: (is_not_free _ _ _); last done.
  case e3: (is_not_free _ _ _); last done.
  simpl; apply Dchoose => //=. by move:H1; move /eqP: e ->; move /eqP: f ->.
  - move => p; case: (proof2deriv _) => [[h t]|]; last done.
  move => H1 q; case: (proof2deriv _) => [[h2 u]|]; last done.
  move => H2; by apply Dconj.
  - move => p; case: (proof2deriv _) =>
    [[h [||||[||||[||[|||||||||]|||] u|] t|]]|] //=.
  move => H1; apply: Dconj1; eexact H1.
  - move => p; case: (proof2deriv _) =>
    [[h [||||[||||[||[|||||||||]|||] u|] t|]]|] //=.
  move => H1; apply: Dconj2; eexact H1.
  - move => p; case: (proof2deriv _) => [[h t]|]; last done.
  move => H1 u; case e: (infer _ _) => [[|||]|] //=. by apply Ddisj1.
  - move => p; case: (proof2deriv _) => [[h t]|]; last done.
  move => H1 u; case e: (infer _ _) => [[|||]|] //=. by apply Ddisj2.
  - move => p; case: (proof2deriv _) => [[h1 t]|]; last done.
  move => H1 q; case: (proof2deriv _) => [[h2 u]|]; last done.
  move => H2 r; case: (proof2deriv _) => [[h3 v]|]; last done.
  move: t H1 => [||||[||||[||[|||||||||]|||] b|] a|] //=.
  move => H1 H3; case e: (_ == _); last done.
  apply Ddisjcases => //=. by move /eqP: e ->.
  - move => p; case: (proof2deriv _) =>
    [[h [||||[||||[||[|||||||||]|||] a|] t|]]|] //=.
  move: t => [||[|||||||||]|||] //=. move => H1; by apply Dnoti.
  - move => p; case: (proof2deriv _) =>
    [[h [||||[||[|||||||||]|||] a|]]|] //=. move => H1; by apply Dnote.
  - move => p; case: (proof2deriv _) => [[h t]|]; last done.
  move => H1 u; case e: (infer _ u) => [[|||]|] //=.
  move: t H1 => [||[|||||||||]|||] //=. move => H1; by apply Dcontr.
  - apply Dtrue.
  - move => p; case: (proof2deriv _) => [[h t]|]; last done.
  move => H1 q; case: (proof2deriv _) => [[h2 u]|]; last done.
  move: t H1 => [||||[||||[||[|||||||||]|||] a1|] b1|] //=.
  move: u => [||||[||||[||[|||||||||]|||] a2|] b2|] //=.
  move => H1 H2; case e: (_ == _); last done. case f: (_ == _); last done.
  simpl; apply Dimpantisym; first done. by move /eqP: e ->; move /eqP: f ->.
  - by move => [h t H].
Qed.


(* Logical connectives *)

Definition x_var := 1.
Definition y_var := 2.
Definition z_var := 3.


Definition DEF_T := Pimpas (Pdisch Ptrue (heq (Bool --> Bool) (Abs Bool
  (Dbr 0)) (Abs Bool (Dbr 0)))) (Pdisch (Prefl (Abs Bool (Dbr 0)))
  htrue).


Definition DEF_AND := (Ptrans (Ptrans (Psym (Ptrans (Pabs (Peta ((App
  (Cst Hand) (Var x_var Bool)))) x_var Bool) (Peta (Cst Hand)))) (Pabs
  (Pabs (Pimpas (Pdisch (Pconj (Pconjunct2 (Phyp ((Var x_var Bool) hand
  (Var y_var Bool)))) (Pconjunct1 (Phyp ((Var x_var Bool) hand (Var
  y_var Bool))))) ((Var x_var Bool) hand (Var y_var Bool))) (Pdisch
  (Pconj (Pconjunct2 (Phyp ((Var y_var Bool) hand (Var x_var Bool))))
  (Pconjunct1 (Phyp ((Var y_var Bool) hand (Var x_var Bool))))) ((Var
  y_var Bool) hand (Var x_var Bool)))) x_var Bool) y_var Bool)) (Pabs
  (Pabs (Pimpas (Pdisch (Pconj (Peqmp (Ptrans (Psym (Ptrans (Pcomb (Phyp
  (heq ((Bool --> Bool --> Bool) --> Bool) (Abs (Bool --> Bool --> Bool)
  (App (App (Dbr 0) (Var y_var Bool)) (Var x_var Bool))) (Abs (Bool -->
  Bool --> Bool) (App (App (Dbr 0) htrue) htrue)))) (Prefl (Abs Bool
  (Abs Bool (Dbr 0))))) (Ptrans (Ptrans (Pbeta1 (App (App (Dbr 0) htrue)
  htrue) (Abs Bool (Abs Bool (Dbr 0)))) (Pcomb (Pbeta1 (Abs Bool (Dbr
  0)) htrue) (Prefl htrue))) (Pbeta1 (Dbr 0) htrue)))) (Ptrans (Ptrans
  (Pbeta1 (App (App (Dbr 0) (Var y_var Bool)) (Var x_var Bool)) (Abs
  Bool (Abs Bool (Dbr 0)))) (Pcomb (Pbeta1 (Abs Bool (Dbr 0)) (Var y_var
  Bool)) (Prefl (Var x_var Bool)))) (Pbeta1 (Dbr 0) (Var x_var Bool))))
  Ptrue) (Peqmp (Ptrans (Psym (Ptrans (Pcomb (Phyp (heq ((Bool --> Bool
  --> Bool) --> Bool) (Abs (Bool --> Bool --> Bool) (App (App (Dbr 0)
  (Var y_var Bool)) (Var x_var Bool))) (Abs (Bool --> Bool --> Bool)
  (App (App (Dbr 0) htrue) htrue)))) (Prefl (Abs Bool (Abs Bool (Dbr
  1))))) (Ptrans (Ptrans (Pbeta1 (App (App (Dbr 0) htrue) htrue) (Abs
  Bool (Abs Bool (Dbr 1)))) (Pcomb (Pbeta1 (Abs Bool (Dbr 1)) htrue)
  (Prefl htrue))) (Pbeta1 htrue htrue)))) (Ptrans (Ptrans (Pbeta1 (App
  (App (Dbr 0) (Var y_var Bool)) (Var x_var Bool)) (Abs Bool (Abs Bool
  (Dbr 1)))) (Pcomb (Pbeta1 (Abs Bool (Dbr 1)) (Var y_var Bool)) (Prefl
  (Var x_var Bool)))) (Pbeta1 (Var y_var Bool) (Var x_var Bool))))
  Ptrue)) (App (App (Cst (Heq ((Bool --> Bool --> Bool) --> Bool))) (Abs
  (Bool --> Bool --> Bool) (App (App (Dbr 0) (Var y_var Bool)) (Var
  x_var Bool)))) (Abs (Bool --> Bool --> Bool) (App (App (Dbr 0) htrue)
  htrue)))) (Pdisch (Pabs (Pcomb (Pcomb (Prefl (Var z_var (Bool --> Bool
  --> Bool))) (Pimpas (Pdisch (Pconjunct2 (Phyp ((Var x_var Bool) hand
  (Var y_var Bool)))) htrue) (Pdisch Ptrue (Var y_var Bool)))) (Pimpas
  (Pdisch (Pconjunct1 (Phyp ((Var x_var Bool) hand (Var y_var Bool))))
  htrue) (Pdisch Ptrue (Var x_var Bool)))) z_var (Bool --> Bool -->
  Bool)) ((Var x_var Bool) hand (Var y_var Bool)))) x_var Bool) y_var
  Bool)).


Definition DEF_IMP := (Psym (Ptrans (Pabs (Ptrans (Pabs (Pimpas (Pdisch
  (Pimpas (Pdisch (Pconj (Phyp (Var x_var Bool)) (Pmp (Phyp ((Var x_var
  Bool) himp (Var y_var Bool))) (Phyp (Var x_var Bool)))) (Var x_var
  Bool)) (Pdisch (Pconjunct1 (Phyp ((Var x_var Bool) hand (Var y_var
  Bool)))) ((Var x_var Bool) hand (Var y_var Bool)))) ((Var x_var Bool)
  himp (Var y_var Bool))) (Pdisch (Pdisch (Pconjunct2 (Peqmp (Psym (Phyp
  (((Var x_var Bool) hand (Var y_var Bool)) hequiv (Var x_var Bool))))
  (Phyp (Var x_var Bool)))) (Var x_var Bool)) (((Var x_var Bool) hand
  (Var y_var Bool)) hequiv (Var x_var Bool)))) y_var Bool) (Peta (App
  (Cst Himp) (Var x_var Bool)))) x_var Bool) (Peta (Cst Himp)))).


Definition DEF_FORALL := (Psym (Ptrans (Pabs (Pimpas (Pdisch (Psym
  (Ptrans (Pabs (Pimpas (Pdisch Ptrue (App (Var x_var (TVar 1 -->
  Bool)) (Var y_var (TVar 1)))) (Pdisch (Pspec (Peqmp (Psym (Pcomb
  (Prefl (Cst (Hforall (TVar 1)))) (Peta (Var x_var (TVar 1 -->
  Bool))))) (Phyp (App (Cst (Hforall (TVar 1))) (Var x_var (TVar
  1 --> Bool))))) (Var y_var (TVar 1))) htrue)) y_var (TVar
  1)) (Peta (Var x_var (TVar 1 --> Bool))))) (App (Cst (Hforall
  (TVar 1))) (Var x_var (TVar 1 --> Bool)))) (Pdisch (Peqmp
  (Pcomb (Prefl (Cst (Hforall (TVar 1)))) (Peta (Var x_var (TVar
  1 --> Bool)))) (Pgen (Peqmp (Psym (Ptrans (Pcomb (Phyp (heq (TVar
  1 --> Bool) (Var x_var (TVar 1 --> Bool)) (Abs (TVar 1)
  htrue))) (Prefl (Var y_var (TVar 1)))) (Pbeta1 htrue (Var y_var
  (TVar 1))))) Ptrue) y_var (TVar 1))) (heq (TVar 1 -->
  Bool) (Var x_var (TVar 1 --> Bool)) (Abs (TVar 1) htrue))))
  x_var (TVar 1 --> Bool)) (Peta (Cst (Hforall (TVar 1)))))).


Definition DEF_EXISTS := (Ptrans (Psym (Peta (Cst (Hexists (TVar
  1))))) (Pabs (Pimpas (Pdisch (Peqmp (Pimpas (Pdisch (Pgen (Pdisch
  (Phyp (App (Cst (Hexists (TVar 1))) (Var x_var (TVar 1 -->
  Bool)))) (App (Var x_var (TVar 1 --> Bool)) (Var y_var (TVar
  1)))) y_var (TVar 1)) (App (Cst (Hexists (TVar 1))) (Var
  x_var (TVar 1 --> Bool)))) (Pspec (Phyp (hforall Bool ((hforall
  (TVar 1) ((App (Var x_var (TVar 1 --> Bool)) (Dbr 0)) himp
  (Dbr 1))) himp (Dbr 0)))) (App (Cst (Hexists (TVar 1))) (Var x_var
  (TVar 1 --> Bool))))) (Pgen (Pdisch (Peqmp (Pcomb (Prefl (Cst
  (Hexists (TVar 1)))) (Peta (Var x_var (TVar 1 --> Bool))))
  (Pexists (Phyp (App (Var x_var (TVar 1 --> Bool)) (Var y_var (TVar
  1)))) (Abs (TVar 1) (App (Var x_var (TVar 1 --> Bool))
  (Dbr 0))) (Var y_var (TVar 1)))) (App (Var x_var (TVar 1 -->
  Bool)) (Var y_var (TVar 1)))) y_var (TVar 1))) (hforall Bool
  ((hforall (TVar 1) ((App (Var x_var (TVar 1 --> Bool)) (Dbr
  0)) himp (Dbr 1))) himp (Dbr 0)))) (Pdisch (Pgen (Pdisch (Pchoose
  z_var (TVar 1) (Peqmp (Pcomb (Prefl (Cst (Hexists (TVar 1))))
  (Psym (Peta (Var x_var (TVar 1 --> Bool))))) (Phyp (App (Cst
  (Hexists (TVar 1))) (Var x_var (TVar 1 --> Bool))))) (Pmp
  (Pspec (Phyp (hforall (TVar 1) ((App (Var x_var (TVar 1 -->
  Bool)) (Dbr 0)) himp (Var y_var Bool)))) (Var z_var (TVar 1)))
  (Phyp (App (Var x_var (TVar 1 --> Bool)) (Var z_var (TVar
  1)))))) (hforall (TVar 1) ((App (Var x_var (TVar 1 -->
  Bool)) (Dbr 0)) himp (Var y_var Bool)))) y_var Bool) (App (Cst
  (Hexists (TVar 1))) (Var x_var (TVar 1 --> Bool))))) x_var
  (TVar 1 --> Bool))).


Definition DEF_OR := (Ptrans (Ptrans (Psym (Ptrans (Pabs (Peta (App (Cst
  Hor) (Var x_var Bool))) x_var Bool) (Peta (Cst Hor)))) (Pabs (Pabs
  (Pimpas (Pdisch (Pdisjcases (Phyp ((Var x_var Bool) hor (Var y_var
  Bool))) (Pdisj2 (Phyp (Var x_var Bool)) (Var y_var Bool)) (Pdisj1
  (Phyp (Var y_var Bool)) (Var x_var Bool))) ((Var x_var Bool) hor (Var
  y_var Bool))) (Pdisch (Pdisjcases (Phyp ((Var y_var Bool) hor (Var
  x_var Bool))) (Pdisj2 (Phyp (Var y_var Bool)) (Var x_var Bool))
  (Pdisj1 (Phyp (Var x_var Bool)) (Var y_var Bool))) ((Var y_var Bool)
  hor (Var x_var Bool)))) x_var Bool) y_var Bool)) (Pabs (Pabs (Pimpas
  (Pdisch (Pmp (Pmp (Pspec (Phyp (hforall Bool (((Var y_var Bool) himp
  (Dbr 0)) himp (((Var x_var Bool) himp (Dbr 0)) himp (Dbr 0))))) ((Var
  x_var Bool) hor (Var y_var Bool))) (Pdisch (Pdisj2 (Phyp (Var y_var
  Bool)) (Var x_var Bool)) (Var y_var Bool))) (Pdisch (Pdisj1 (Phyp (Var
  x_var Bool)) (Var y_var Bool)) (Var x_var Bool))) (hforall Bool (((Var
  y_var Bool) himp (Dbr 0)) himp (((Var x_var Bool) himp (Dbr 0)) himp
  (Dbr 0))))) (Pdisch (Pgen (Pdisch (Pdisch (Pdisjcases (Phyp ((Var
  x_var Bool) hor (Var y_var Bool))) (Pmp (Phyp ((Var x_var Bool) himp
  (Var z_var Bool))) (Phyp (Var x_var Bool))) (Pmp (Phyp ((Var y_var
  Bool) himp (Var z_var Bool))) (Phyp (Var y_var Bool)))) ((Var x_var
  Bool) himp (Var z_var Bool))) ((Var y_var Bool) himp (Var z_var
  Bool))) z_var Bool) ((Var x_var Bool) hor (Var y_var Bool)))) x_var
  Bool) y_var Bool)).


Definition DEF_F := (Pimpas (Pdisch (Pspec (Phyp (hforall Bool (Dbr 0)))
  hfalse) (hforall Bool (Dbr 0))) (Pdisch (Pgen (Pcontr (Phyp hfalse)
  (Var x_var Bool)) x_var Bool) hfalse)).


Definition DEF_NOT := (Psym (Ptrans (Pabs (Pimpas (Pdisch (Pnote (Phyp
  (hnot (Var x_var Bool)))) (hnot (Var x_var Bool))) (Pdisch (Pnoti
  (Phyp ((Var x_var Bool) himp hfalse))) ((Var x_var Bool) himp
  hfalse))) x_var Bool) (Peta (Cst Hnot)))).


Definition ETA_AX := (Pgen (Peta (Var x_var ((TVar 1) --> (TVar
  2)))) x_var ((TVar 1) --> (TVar 2))).


Definition SELECT_AX := (Pgen (Pgen (Peps (Var x_var ((TVar 1) -->
  Bool)) (Var y_var (TVar 1))) y_var (TVar 1)) x_var ((TVar
  1) --> Bool)).

Unset Implicit Arguments.
