(* ========================================================================= *)
(* An effective finite sets library, basing on the ssreflect approach        *)
(*                                                                           *)
(*       Chantal Keller, Laboratoire d'Informatique de Polytechnique         *)
(*                                                                           *)
(*    (c) Copyright, Laboratoire d'Informatique de Polytechnique 2010        *)
(*              (c) Copyright, Chantal Keller 2010                           *)
(* ========================================================================= *)


Require Import ssreflect ssrbool eqtype ssrnat.
Require Export List SetoidList.

Set Implicit Arguments.
Unset Strict Implicit.


(* Finite sets over an eqType *)

Section Ssrfsets.

  Variable E: eqType.
  Definition elt : Type := E.
  Definition t := list elt.

  Notation In := (InA (fun x y => x == y)).

  Definition Subset s s' := forall a : elt, In a s -> In a s'.
  Definition Empty s := forall a : elt, ~ In a s.
  Definition For_all (P : elt -> Prop) s := forall x, In x s -> P x.

  Notation "s  [<=]  t" := (Subset s t) (at level 70, no associativity).


  Definition empty : t := nil.


  Fixpoint add (e: elt) (l: t) : t :=
    match l with
      | nil => e::nil
      | t::q => if (e == t) then t::q else t::(add e q)
    end.


  Definition singleton (e: elt) : t := e::nil.


  Fixpoint remove (e: elt) (l: t) {struct l} : t :=
    match l with
      | nil => nil
      | t::q => if (e == t) then remove e q else t::(remove e q)
    end.


  Definition union (l m: t) : t := fold_left (fun n e => add e n) l m.


  Definition map (f: elt -> elt) (l: t) : t :=
    fold_left (fun m e => add (f e) m) l empty.


  Fixpoint for_all (P: elt -> bool) (l: t) {struct l} : bool :=
    match l with
      | nil => true
      | t::q => P t && for_all P q
    end.


  Definition elements (l: t) : list elt := l.


  Definition is_empty (l: t) : bool :=
    match l with
      | nil => true
      | _ => false
    end.


  (** Specification of [add] *)

  Lemma add_1 : forall x s, In x (add x s).
  Proof.
    move => x; elim => [|z q IHq] /=; intuition.
    case e: (x == z); intuition.
  Qed.

  Lemma add_2 : forall x y s, In y s -> In y (add x s).
  Proof.
    move => x y; elim => [|z q IHq] /=; intuition.
    case e: (x == z); intuition.
    inversion_clear H; intuition.
  Qed.

  Lemma add_3 : forall x y s, x != y -> In y (add x s) -> In y s.
  Proof.
    move => x y s H; elim s => [|z q IHq] /=.
    - move => H1; inversion_clear H1; intuition.
    by move: H; move /eqP: H0 ->; rewrite eqxx.
    - case e: (x == z); first intuition.
    move => H1; inversion_clear H1; intuition.
  Qed.

  (** Specification of [remove] *)

  Lemma remove_1 : forall x s, ~ In x (remove x s).
  Proof.
    move => x; elim => [|z s1 IHs] /=.
    - move => H; inversion H.
    - case e: (x == z); first intuition.
    move => H; inversion_clear H; last intuition.
    by rewrite H0 in e.
  Qed.

  Lemma remove_2 : forall x y s, x != y -> In y s -> In y (remove x s).
  Proof.
    move => x y; elim => [|z s1 IHs] //=.
    move => H H1; inversion_clear H1.
    - have H1: ((x == z) = false).
    by apply negbTE; move /eqP: H0 <-.
    rewrite H1; move /eqP: H0 ->; intuition.
    - case e: (x == z); intuition.
  Qed.

  Lemma remove_3 : forall x y s, In y (remove x s) -> In y s.
  Proof.
    move => x y; elim => [|z s1 IHs] //=.
    case e: (x == z); intuition.
    inversion_clear H; intuition.
  Qed.

  (** Specification of [singleton] *)

  Lemma singleton_1 : forall x y, In y (singleton x) -> x == y.
  Proof.
    by move => x y H; inversion_clear H; [rewrite eq_sym|inversion H0].
  Qed.

  Lemma singleton_2 : forall x, In x (singleton x).
  Proof. by constructor. Qed.

  (** Specification of [union] *)

  Lemma union_1 : forall x s s', In x (union s s') -> In x s \/ In x s'.
  Proof.
    move => x; elim => [|z s1 IHs] /=; intuition.
    move: (IHs _ H) => [H1|H1]; intuition.
    case e: (z == x).
    - move /eqP: e ->; intuition.
    - right; apply: add_3; eauto. by apply negbT.
  Qed.

  Lemma union_aux_2 : forall x s s',
    (In x s \/ In x s') -> In x (union s s').
  Proof.
    move => x; elim => [|z s1 IHs] /=.
    - by move => s' [H|H]; [inversion H|].
    - move => s' [H|H].
    inversion_clear H; last intuition.
    apply IHs; right; move /eqP: H0 ->; apply add_1.
    by apply IHs; right; apply add_2.
  Qed.

  Lemma union_2 : forall x s s', In x s -> In x (union s s').
  Proof. by move => x s s' H; apply union_aux_2; left. Qed.

  Lemma union_3 : forall x s s', In x s' -> In x (union s s').
  Proof. by move => x s s' H; apply union_aux_2; right. Qed.

  (** Specification of [map] *)

  Lemma map_aux_1 : forall (f: elt -> elt) x s acc,
    In x (fold_left (fun m e => add (f e) m) s acc) ->
    (exists y, In y s /\ x == f y) \/ (In x acc).
  Proof.
    move => f x; elim => [|z s1 IHs] /=; intuition.
    move: (IHs _ H) => [[a [H1 H2]]|H1].
    - left; exists a; intuition.
    - case e: (f z == x).
    by left; exists z; split; [intuition|rewrite eq_sym].
    right; apply: add_3; eauto. by apply negbT.
  Qed.

  Lemma map_1 : forall (f: elt -> elt) x s,
    In x (map f s) -> exists y, In y s /\ x == f y.
  Proof.
    move => f x s H; move: (map_aux_1 H) => [[a [H1 H2]]|H1].
    - exists a; intuition.
    - inversion H1.
  Qed.

  Lemma map_aux_2 : forall (f: elt -> elt) x s acc,
    ((exists y, In y s /\ x == f y) \/ (In x acc)) ->
    In x (fold_left (fun m e => add (f e) m) s acc).
  Proof.
    move => f x; elim => [|z s1 IHs] /=.
    - by move => acc [[a [H _]]|H]; [inversion H|].
    - move => acc H; apply IHs; move: H => [[a [H1 H2]]|H].
    inversion_clear H1.
    right; move /eqP: H2 ->; move /eqP: H ->; apply add_1.
    left; exists a; intuition.
    by right; apply add_2.
  Qed.

  Lemma map_2 : forall (f: elt -> elt) x s,
    (exists y, In y s /\ x == f y) -> In x (map f s).
  Proof. by move => f x s H; apply map_aux_2; left. Qed.

  (** Specification of [for_all] *)

  Lemma for_all_1 : forall P s,
    For_all (fun x => P x = true) s -> for_all P s.
  Proof.
    move => P; elim => [|z s1 IHs] //=.
    move => H; apply andb_true_intro; intuition.
    apply IHs => a Ha; apply H; intuition.
  Qed.

  Lemma for_all_2 : forall P s,
    for_all P s -> For_all (fun x => P x = true) s.
  Proof.
    move => P; elim => [|z s1 IHs] /=.
    - move => _ a Ha; inversion Ha.
    - move => H a Ha; move /andP: H => [H1 H2];
      inversion_clear Ha; intuition. by move /eqP: H ->.
  Qed.

  (** Specification of [elements] *)

  Lemma elements_1 : forall x s,
    In x s -> InA (fun x y => x == y) x (elements s).
  Proof. done. Qed.

  Lemma elements_2 : forall x s,
    InA (fun x y => x == y) x (elements s) -> In x s.
  Proof. done. Qed.


  Set Strict Implicit.


  Section Properties.

    (* Properties about For_all *)

    Lemma For_all_empty : forall P, For_all P empty.
    Proof. move => P x Hx; inversion Hx. Qed.

    Lemma For_all_singleton1 : forall P a, For_all P (singleton a) -> P a.
    Proof. move => P a H; apply H; by apply singleton_2. Qed.

    Lemma For_all_singleton2 : forall (P: elt -> Prop) a,
      P a -> For_all P (singleton a).
    Proof. move => P a H x H1; by move /eqP: (singleton_1 H1) <-. Qed.

    Lemma For_all_union1 : forall (P: elt -> Prop) s1 s2,
      For_all P (union s1 s2) -> For_all P s1 /\ For_all P s2.
    Proof.
      move => P s1 s2 H; split => a Ha; apply H;
        [by apply union_2|by apply union_3].
    Qed.

    Lemma For_all_union2 : forall (P: elt -> Prop) s1 s2,
      For_all P s1 -> For_all P s2 -> For_all P (union s1 s2).
    Proof.
      move => P s1 s2 H1 H2 x H;
        move: (union_1 H) => {H} [H|H]; [by apply H1|by apply H2].
    Qed.

    Lemma For_all_remove1 : forall (P: elt -> Prop) a s,
      P a -> For_all P (remove a s) -> For_all P s.
    Proof.
      move => P a q H1 H2 x H; case e: (a == x); first by move /eqP: e <-.
      apply H2; apply remove_2; by [apply negbT|].
    Qed.

    Lemma For_all_remove2 : forall (P: elt -> Prop) a s,
      For_all P s -> For_all P (remove a s).
    Proof. move => P a s H1 x H; apply H1; apply (remove_3 H). Qed.

    Lemma For_all_add1 : forall P a s,
      For_all P (add a s) -> For_all P s /\ P a.
    Proof.
      move => P a s H1; split.
      - move => x H; apply H1; by apply add_2.
      - apply H1; by apply add_1.
    Qed.

    Lemma For_all_add2 : forall P a s,
      For_all P s -> P a -> For_all P (add a s).
    Proof.
      move => P a s H1 H2 x Hx; case e: (a == x); first by move /eqP: e <-.
      apply H1; apply: add_3; last by eexact Hx. by apply negbT.
    Qed.

    Lemma For_all_imp : forall (P Q: elt -> Prop) s,
      For_all P s -> (forall x, P x -> Q x) -> For_all Q s.
    Proof. move => P Q s H1 H2 x Hx; apply H2; by apply H1. Qed.

    Lemma For_all_subset : forall P s1 s2,
      Subset s1 s2 -> For_all P s2 -> For_all P s1.
    Proof. firstorder. Qed.

    Lemma For_all_and : forall P Q s,
      For_all P s -> For_all Q s -> For_all (fun x => P x /\ Q x) s.
    Proof. firstorder. Qed.

    (* Properties about for_all *)

    Ltac for_all1 :=
      rewrite /ssrbool.is_true;
        match goal with
          | |- (for_all ?P _ = true) => apply for_all_1
        end.


    Ltac for_all2 :=
      unfold ssrbool.is_true in * |-;
        match goal with
          | H : for_all ?P _ = true |- _ => move: (for_all_2 H) => {H} H
        end.


    Ltac for_all_tac := try for_all1; try for_all2.


    Lemma for_all_singleton1 : forall P a, for_all P (singleton a) -> P a.
    Proof.
      move => P a H1; for_all_tac. by apply (For_all_singleton1 (P:=P)).
    Qed.

    Lemma for_all_remove1 : forall (P: elt -> bool) a s,
      P a -> for_all P (remove a s) -> for_all P s.
    Proof.
      move => P a s H1 H2; for_all_tac; apply: For_all_remove1;
        by [eexact H1|].
    Qed.

    Lemma for_all_add1 : forall P a s,
      for_all P (add a s) -> ((for_all P s) && (P a)).
    Proof.
      move => P a s H1; move: (For_all_add1 (P:=fun x : elt => P x = true)
        a s) => H; apply andb_true_intro; split; for_all_tac;
      by move: (H H1) => {H H1} [H1 H2].
    Qed.

    Lemma for_all_subset : forall P s1 s2,
      Subset s1 s2 -> for_all P s2 -> for_all P s1.
    Proof.
      move => P s1 s2 H1 H2; for_all_tac; apply: For_all_subset;
        by [eexact H1|].
    Qed.

    (* Miscellanous properties *)

    Lemma add_order : forall a x y s,
      In a (add x (add y s)) -> In a (add y (add x s)).
    Proof.
      move => a x y s H; case H1: (x == a).
      - move /eqP: H1 ->; apply add_2; by apply add_1.
      - move: (add_3 (negbT H1) H) => {H1 H} H; case H1: (y == a);
        first by (move /eqP: H1 ->; apply add_1).
      move: (add_3 (negbT H1) H) => {H1 H} H; by do 2 apply add_2.
    Qed.

    Lemma refl_subset : forall s, Subset s s.
    Proof. by move => s x Hx. Qed.

    Lemma empty_subset : forall s, Subset empty s.
    Proof. move => s x Hx; inversion Hx. Qed.

    Lemma union_subset : forall a b c,
      Subset b c -> Subset (union a b) (union a c).
    Proof.
      move => a b c H x Hx; move: (union_1 Hx) => {Hx} [Hx|Hx].
      by apply union_2.
      by apply union_3; apply H.
    Qed.

    Lemma union_subset_all : forall a b c d,
      Subset a d -> Subset b c -> Subset (union a b) (union d c).
    Proof.
      move => a b c d H1 H2 x Hx; move: (union_1 Hx) => {Hx} [Hx|Hx].
      by apply union_2; apply H1.
      by apply union_3; apply H2.
    Qed.

    Lemma add_subset : forall a s1 s2,
      Subset s1 s2 -> Subset (add a s1) (add a s2).
    Proof.
      move => a s1 s2 H1 x Hx; case H: (a == x);
        first by (move /eqP: H ->; apply add_1).
      apply add_2; apply H1; apply: add_3;
        by [apply negbT; eassumption|exact Hx].
    Qed.

    Lemma add_remove_subset : forall a s, Subset s (add a (remove a s)).
    Proof.
      move => a s x Hx; case H: (a == x);
        first by (move /eqP: H ->; apply add_1).
      apply add_2; apply remove_2; by [apply negbT|].
    Qed.

    (* Properties about map *)

    Lemma map_remove : forall g a s,
      Subset (remove (g a) (map g s)) (map g (remove a s)).
    Proof.
      move => g a s x Hx; apply map_2; move: (remove_3 Hx) => H;
        move: (map_1 H) => [y [H1 H2]]; exists y; intuition.
      apply remove_2; last done.
      apply negbT; case e: (a == y); last done.
      by elim (remove_1 (x:=x) (s:=map g s)); move /eqP: H2 => H2;
        rewrite {2}H2; move /eqP: e <-.
    Qed.

    Lemma map_subset_For_all : forall g P s, For_all P s ->
      (forall x, P x -> g x = x) -> Subset (map g s) s.
    Proof.
      move => g P s H1 H2 x Hx; move: (map_1 Hx) => [y [H3 H4]];
        move /eqP: H4 ->; rewrite H2; intuition.
    Qed.

    Lemma map_subset : forall g s1 s2, Subset s1 s2 ->
      Subset (map g s1) (map g s2).
    Proof.
      move => g s1 s2 H x Hx; apply map_2; move: (map_1 Hx) => [y [H1 H2]];
        exists y; intuition.
    Qed.

    Lemma map_singleton : forall g x,
      Subset (singleton (g x)) (map g (singleton x)).
    Proof.
      move => g x y Hy; apply map_2; exists x; split.
      - apply singleton_2.
      - by rewrite eq_sym; apply singleton_1.
    Qed.

    Lemma map_union : forall g s1 s2,
      Subset (union (map g s1) (map g s2)) (map g (union s1 s2)).
    Proof.
      move => g s1 s2 x Hx; apply map_2; move: (union_1 Hx) => {Hx} [Hx|Hx];
        move: (map_1 Hx) => [y [H1 H2]]; exists y; intuition;
          by [apply union_2|apply union_3].
    Qed.

    Lemma map_add : forall g a s,
      Subset (map g (add a s)) (add (g a) (map g s)).
    Proof.
      move => g a s x Hx; move: (map_1 Hx) => [y [H1 H2]]; case e: (a == y).
      - move /eqP: H2 ->; move /eqP: e <-; apply add_1.
      - apply add_2; apply map_2; exists y; intuition. apply: add_3; eauto.
      by apply negbT.
    Qed.

    Lemma map_inv : forall g1 g2 (P: elt -> Prop) s,
      (forall x, P x -> g2 (g1 x) = x) -> For_all P s ->
      Subset (map g2 (map g1 s)) s.
    Proof.
      by move => g1 g2 P s H1 H2 x Hx; move: (map_1 Hx) => {Hx} [y [H3 H4]];
        move: (map_1 H3) => {H3} [z [H3 H5]]; move /eqP: H4 ->;
          move /eqP: H5 ->; rewrite H1; [|apply H2].
    Qed.

  End Properties.

End Ssrfsets.


Ltac for_all1 :=
  rewrite /ssrbool.is_true;
    match goal with
      | |- (for_all ?P _ = true) => apply for_all_1
    end.


Ltac for_all2 :=
  unfold ssrbool.is_true in * |-;
    match goal with
      | H : for_all ?P _ = true |- _ => move: (for_all_2 H) => {H} H
    end.


Ltac for_all_tac := try for_all1; try for_all2.


Implicit Arguments empty [E].
Implicit Arguments For_all_singleton2 [E a].
Implicit Arguments for_all_singleton1 [E P a].

Unset Implicit Arguments.
