(* ========================================================================= *)
(* Positives as eqTypes                                                      *)
(*                                                                           *)
(*       Chantal Keller, Laboratoire d'Informatique de Polytechnique         *)
(*                                                                           *)
(*    (c) Copyright, Laboratoire d'Informatique de Polytechnique 2010        *)
(*              (c) Copyright, Chantal Keller 2010                           *)
(* ========================================================================= *)


Require Import ssreflect eqtype ssrbool ssrnat.
Require Import NArith List.
Require Import list_ext.

Open Local Scope positive_scope.

Set Implicit Arguments.


Section Eq_positive.

  Definition eq_positive (n m: positive) : bool :=
    match Pcompare n m Eq with
      | Eq => true
      | _ => false
    end.


  Lemma eq_positiveP : forall n m, reflect (n = m) (eq_positive n m).
  Proof.
    rewrite /eq_positive => n m; case e: ((n ?= m) Eq); constructor.
    - apply (Pcompare_Eq_eq _ _ e).
    - by move => H; rewrite H in e; rewrite Pcompare_refl in e.
    - by move => H; rewrite H in e; rewrite Pcompare_refl in e.
  Qed.


  Canonical Structure positive_domM := EqMixin eq_positiveP.
  Canonical Structure positive_dom := EqType positive_domM.

End Eq_positive.


Definition names := positive.
Definition names_dom := positive_dom.


(* New definitions of Plt, Pgt, Ple and Pge *)

Definition Plt2 (x y: positive) :=
  match Pcompare x y Eq with
    | Lt => true
    | _  => false
  end.

Definition Pgt2 (x y: positive) :=
  match Pcompare x y Eq with
    | Gt => true
    | _  => false
  end.

Definition Ple2 (x y: positive) :=
  match Pcompare x y Eq with
    | Gt => false
    | _  => true
  end.

Definition Pge2 (x y: positive) :=
  match Pcompare x y Eq with
    | Lt => false
    | _  => true
  end.

Lemma PltP : forall x y, reflect (x < y) (Plt2 x y).
Proof. by rewrite /Plt /Plt2 => x y; case: ((x ?= y) Eq); constructor. Qed.

Lemma PgtP : forall x y, reflect (x > y) (Pgt2 x y).
Proof. by rewrite /Pgt /Pgt2 => x y; case: ((x ?= y) Eq); constructor. Qed.

Lemma PleP : forall x y, reflect (x <= y) (Ple2 x y).
Proof. by rewrite /Ple /Ple2 => x y; case: ((x ?= y) Eq); constructor. Qed.

Lemma PgeP : forall x y, reflect (x >= y) (Pge2 x y).
Proof. by rewrite /Pge /Pge2 => x y; case: ((x ?= y) Eq); constructor. Qed.


Lemma Plt_intro : forall x y, x < y -> Plt2 x y.
Proof. by move => x y H; move /PltP: H. Qed.

Lemma Pgt_intro : forall x y, x > y -> Pgt2 x y.
Proof. by move => x y H; move /PgtP: H. Qed.

Lemma Ple_intro : forall x y, x <= y -> Ple2 x y.
Proof. by move => x y H; move /PleP: H. Qed.

Lemma Pge_intro : forall x y, x >= y -> Pge2 x y.
Proof. by move => x y H; move /PgeP: H. Qed.


(* Properties of Ple *)

Lemma Ple_refl : forall n, n <= n.
Proof. by rewrite /Ple => n; rewrite Pcompare_refl. Qed.


Lemma Ple_trans : forall n m p, n <= m -> m <= p -> n <= p.
Proof.
  rewrite /Ple => n m p.
  case e: ((n ?= m) Eq); last done; move => _.
  - by rewrite (Pcompare_Eq_eq _ _ e).
  - case f: ((m ?= p) Eq); last done; move => _.
  by rewrite -(Pcompare_Eq_eq _ _ f) e.
  by rewrite (Plt_trans _ _ _ e f).
Qed.


(* Properties mixing Ple and Plt *)

Lemma Plt_Ple : forall n m, n < m -> n <= m.
Proof. by rewrite /Plt /Ple => n m ->. Qed.


Lemma Plt_Ple_Plt : forall n m t, t < n -> n <= m -> t < m.
Proof.
  rewrite /Plt /Ple => n m t; case e: ((t ?= n) Eq); try done; move => _.
  case f: ((n ?= m) Eq); try done; move => _.
  - by rewrite -(Pcompare_Eq_eq _ _ f) e.
  - by rewrite (Plt_trans _ _ _ e f).
Qed.


Lemma Ple_Plt_Plt : forall n m t, n <= m -> m < t -> n < t.
Proof.
  rewrite /Ple /Plt => n m t; case e: ((n ?= m) Eq); last done; move => _.
  - by rewrite (Pcompare_Eq_eq _ _ e).
  - move => f; apply (Plt_trans _ _ _ e f).
Qed.


(* Properties of Pmax *)

Lemma Pmax_1 : forall n m, n <= Pmax n m.
Proof.
  rewrite /Pmax => n m; case e: ((n ?= m) Eq).
  - rewrite (Pcompare_Eq_eq _ _ e); apply Ple_refl.
  - by apply Plt_Ple.
  - apply Ple_refl.
Qed.


Lemma Pmax_2 : forall n m, m <= Pmax n m.
Proof.
  rewrite /Pmax => n m; case e: ((n ?= m) Eq); try apply Ple_refl.
  by apply Plt_Ple; apply ZC1.
Qed.


Lemma Pmax_3 : forall n m t, n <= m -> Pmax n t <= Pmax m t.
Proof.
  rewrite /Pmax => n m t H; case e: ((n ?= t) Eq).
  - rewrite -(Pcompare_Eq_eq _ _ e);
    case: ((m ?= n) Eq); by try apply Ple_refl.
  - case f: ((m ?= t) Eq); try apply Ple_refl.
  by apply Plt_Ple; apply ZC1.
  - by rewrite (ZC2 _ _ (Plt_Ple_Plt (ZC1 _ _ e) H)).
Qed.


Lemma Pmax_4 : forall n m t, n <= m -> Pmax t n <= Pmax t m.
Proof.
  rewrite /Pmax => n m t H; case e: ((t ?= n) Eq).
  - rewrite (Pcompare_Eq_eq _ _ e);
    case: ((n ?= m) Eq); by try apply Ple_refl.
  - by rewrite (Plt_Ple_Plt e H).
  - case f: ((t ?= m) Eq).
  rewrite (Pcompare_Eq_eq _ _ f); apply Ple_refl.
  by apply Plt_Ple.
  apply Ple_refl.
Qed.


(* Fresh names generation *)

Fixpoint max (L: list names) (m: positive) {struct L} : positive :=
  match L with
    | nil => m
    | t::q => max q (Pmax t m)
  end.


Lemma max_growing : forall q n m, n <= m -> max q n <= max q m.
Proof.
  elim => [|t q IHq] n m H1 //=.
  apply IHq; move: H1; apply Pmax_4.
Qed.


Lemma max_is_max : forall L m,
  (forallb (fun (y: names) => Ple2 y (max L m)) L) /\ (m <= (max L m)).
Proof.
  elim => [|t q IHq] m /=.
  - split; first done. apply Ple_refl.
  - split; first (apply andb_true_intro; split).
  apply Ple_intro; move: (IHq t) => [_ H]; eapply Ple_trans; first eexact H.
  apply max_growing; apply Pmax_1.
  move: (IHq (Pmax t m)) => [H _]; apply H.
  move: (IHq m) => [_ H]; eapply Ple_trans; first eexact H.
  apply max_growing; apply Pmax_2.
Qed.


Lemma fresh_names1 : forall L, List.forallb
  (fun (y: names) => y != (Psucc (max L 1))) L.
Proof.
  move => L; move: (max_is_max L 1) => [H _]; apply: forallb_imp;
    last eexact H.
  move => /= x; move /PleP => H1; have H2: (x < Psucc (max L 1)).
  eapply Ple_Plt_Plt; first eexact H1. apply Pcompare_p_Sp.
  have H3: (x <> Psucc (max L 1)).
  move => H3; rewrite -H3 in H2; apply: Plt_irrefl; eexact H2.
  by move /eqP: H3.
Qed.


Definition fresh_names (L: list names) :
  {x : names | List.forallb (fun (y: names) => y != x) L} :=
  exist _ (Psucc (max L 1)) (fresh_names1 L).


Unset Implicit Arguments.
