(* ========================================================================= *)
(* Substitutions for De Bruijn indices                                       *)
(*                                                                           *)
(*       Chantal Keller, Laboratoire d'Informatique de Polytechnique         *)
(*                                                                           *)
(*    (c) Copyright, Laboratoire d'Informatique de Polytechnique 2010        *)
(*              (c) Copyright, Chantal Keller 2010                           *)
(* ========================================================================= *)


Require Import ssreflect ssrbool eqtype ssrnat.
Require Import hol cast typing.
Require Import List.

Set Implicit Arguments.


(* Definitions *)

Fixpoint lift_term (u: term) (i: nat) (j: nat) {struct u} : term :=
  match u with
    | Dbr n => if n >= i then Dbr (j + n) else Dbr n
    | App u1 u2 => App (lift_term u1 i j) (lift_term u2 i j)
    | Abs A t => Abs A (lift_term t (S i) j)
    | u => u
  end.


Fixpoint subst_db (t: term) (n: nat) (u: term) {struct t} : term :=
  match t with
    | Dbr i => if i < n then Dbr i else if i == n then u else Dbr (i-1)
    | App t1 t2 => App (subst_db t1 n u) (subst_db t2 n u)
    | Abs A t => Abs A (subst_db t (S n) (lift_term u 0 1))
    | u => u
  end.


Definition open (t u: term) := subst_db t O u.


(* Lifting lemmas *)

Lemma lift_id : forall (u: term) (n: nat),
   lift_term u n 0 = u.
Proof.
elim => //=.
- by move => n m; case: (_ <= _).
- by move => t1 IHt1 t2 IHt2 n; move: (IHt1 n) => {IHt1} ->;
   move: (IHt2 n) => {IHt2} ->.
- by move => A t IHt n; move: (IHt (n.+1)) => {IHt} ->.
Qed.


Lemma lift_S : forall (p: nat) (u: term) (n: nat),
   lift_term u n (S p) = lift_term (lift_term u n p) n 1.
Proof.
move => p; elim => //=.
- move => n m;
  case e: (m <= n) => /=.
  assert (m <= p+n); first by apply: (leq_trans (n:=n)); last apply: leq_addl.
  by move: H; case: (m <= p+n).
  by move: e; case: (m <= n).
- by move => t1 IHt1 t2 IHt2 n;
  move: (IHt1 n) => {IHt1} ->;
  move: (IHt2 n) => {IHt2} ->.
- by move => A t IHt n;
  move: (IHt (n.+1)) => ->.
Qed.


(* Interaction between lifting and well-typedness *)

Lemma wt_lift_1 : forall (g: context) (A: type) (t: term)
  (d: context),
  infer (d ++ A::g) (lift_term t (length d) 1) = infer (d ++ g) t.
Proof.
  move => g A; elim => [n|//=|//=|//=|t1 IHt1 t2 IHt2|B t IHt].
  - rewrite /infer => /= d; case e: (_ <= _).
  rewrite add1n opt_nth_gt; last by apply leqW.
  rewrite opt_nth_gt; last done.
  by rewrite leq_subS.
  by do 2 (rewrite opt_nth_le; last by rewrite ltnNge; apply (negbT e)).
  - by move => d /=; move: (IHt1 d) => {IHt1} ->;
    move: (IHt2 d) => {IHt2} ->.
  - by move => d; move: (IHt (B::d)) => {IHt} /= ->.
Qed.


Lemma wt_lift_1_inv1 : forall (g: context) (A: type) (t: term)
  (d: context) (T: type),
  wt (d ++ A::g) (lift_term t (length d) 1) T -> wt (d ++ g) t T.
Proof.
  rewrite /wt => g A t d T <-; symmetry; apply wt_lift_1.
Qed.


Lemma wt_lift_1_inv2 : forall (g: context) (A: type) (t: term)
  (d: context) (T: type),
   wt (d ++ g) t T -> wt (d ++ A::g) (lift_term t (length d) 1) T.
Proof.
  rewrite /wt => g A t d T <-; apply wt_lift_1.
Qed.


Lemma wt_lift : forall (u: term) (g d: context),
  infer g u = infer (d++g) (lift_term u 0 (length d)).
Proof.
  move => u g; elim => [|B d IHd] /=; first by rewrite lift_id.
  by rewrite IHd lift_S
    (wt_lift_1 (d++g) B (lift_term u 0 (length d)) nil).
Qed.


Lemma wt_lift_inv1 : forall (u: term) (A: type) (g d: context),
  wt (d++g) (lift_term u 0 (length d)) A -> wt g u A.
Proof.
  by move => u A g d; rewrite /wt (wt_lift u g d).
Qed.


Lemma wt_lift_inv2 : forall (u: term) (A: type) (g d: context),
  wt g u A -> wt (d++g) (lift_term u 0 (length d)) A.
Proof.
  by move => u A g d; rewrite /wt (wt_lift u g d).
Qed.


(* Substitution lemma for wt *)

Lemma wt_subst_db : forall (g: context) (A: type) (u: term),
  wt g u A -> forall (t: term) (d: context),
    infer (d ++ (A::g)) t =
    infer (d ++ g) (subst_db t (length d) (lift_term u 0 (length d))).
Proof.
  move => g A u h; elim => [n|//=|//=|//=|t1 IHt1 t2 IHt2|B t IHt] /=.
  - move => d; case e: (n < length d) => /=;
    first by do 2 (rewrite opt_nth_le; last done).
  case e2: (_ == _) => /=; first (move /eqP: e2 ->;
    rewrite opt_nth_app -h; apply wt_lift).
  rewrite opt_nth_gt; last by rewrite leqNgt; apply negbT.
  rewrite -(prednK (n := n - length d)) /=; last by
    rewrite -(subnn (length d)); apply ltn_sub2r;
      (rewrite ltn_neqAle leqNgt negbT;
        [by rewrite negbT|by rewrite eq_sym]).
  rewrite opt_nth_gt; last by rewrite -(succnK (length d)) -subn1;
    apply leq_sub2r; rewrite ltn_neqAle leqNgt negbT;
      [by rewrite negbT|by rewrite eq_sym].
  by rewrite -subn1 !subn_sub addnC.
  - by move => d; rewrite (IHt1 d) (IHt2 d).
  - by move => d; rewrite (IHt (B::d)) /= lift_S.
Qed.


Lemma wt_subst_db_inv1 : forall (g: context) (A: type) (u: term),
  wt g u A -> forall (t: term) (d: context) (B: type),
  wt (d ++ g) (subst_db t (length d) (lift_term u 0 (length d))) B ->
  wt (d ++ (A::g)) t B.
Proof.
  by rewrite /wt => g A u h t d B; rewrite (wt_subst_db h).
Qed.


Lemma wt_subst_db_inv2 : forall (g: context) (A: type) (u: term),
  wt g u A -> forall (t: term) (d: context) (B: type),
  wt (d ++ (A::g)) t B ->
  wt (d ++ g) (subst_db t (length d) (lift_term u 0 (length d))) B.
Proof.
  by rewrite /wt => g A u h t d B; rewrite (wt_subst_db h).
Qed.


(* Well-typedness of open *)

Lemma wt_open : forall (g: context) (t u: term) (B: type),
  wt g t B -> infer g (open u t) = infer (B::g) u.
Proof.
  by rewrite /open => g t u B h; rewrite (wt_subst_db h u nil) /= lift_id.
Qed.


Lemma wt_open_inv1 : forall g t u A B,
  wt g t B -> wt g (open u t) A -> wt (B::g) u A.
Proof.
  by move => g t u A B h1; rewrite /wt (wt_open u h1).
Qed.


Lemma wt_open_inv2 : forall (g: context) (t u: term) (A B: type),
  wt g t B -> wt (B::g) u A -> wt g (open u t) A.
Proof.
  by move => g t u A B h1; rewrite /wt (wt_open u h1).
Qed.


(* lift_term is the identity for locally closed terms *)

Lemma lift_term_id_aux : forall t g A, wt g t A ->
  forall i j, i >= length g -> lift_term t i j = t.
Proof.
  elim => [n|x X|c|c C|t IHt v IHv|B t IHt] g A H1 i j H2 //=.
  - case e: (i <= n); last done.
  move: H1; rewrite /wt /=.
  have h: (length g <= n).
  by apply: leq_trans; first eexact H2.
  by move: (opt_nth_gt g nil n h); rewrite opt_nth_nil -app_nil_end => ->.
  - move: (wt_app_inv1 H1) => {H1} [B [H1 H3]];
    by rewrite (IHt g (B --> A) H1 i j H2) (IHv g B H3 i j H2).
  - move: (wt_abs_inv1 H1) => {H1} [C [H1 H3]];
    rewrite (IHt (B::g) C H3); first done.
  by rewrite /= ltnS.
Qed.


Lemma lift_term_id : forall t A, wt nil t A ->
  forall i j, lift_term t i j = t.
Proof.
  move => t A H n u; apply: lift_term_id_aux; [eexact H|done].
Qed.



(* subst_db is the identity for locally closed terms *)

Lemma subst_db_id_aux : forall t g A, wt g t A ->
  forall n u, n >= length g -> subst_db t n u = t.
Proof.
  elim => [i|x X|c|c C|t IHt v IHv|B t IHt] g A H1 n u H2 //=.
  - case e: (i < n); first done.
  move: H1; rewrite /wt /=.
  have h: (length g <= i).
  apply: leq_trans; first eexact H2.
  by rewrite leqNgt; apply negbT.
  by move: (opt_nth_gt g nil i h); rewrite opt_nth_nil -app_nil_end => ->.
  - move: (wt_app_inv1 H1) => {H1} [B [H1 H3]];
    by rewrite (IHt g (B --> A) H1 n u H2) (IHv g B H3 n u H2).
  - move: (wt_abs_inv1 H1) => {H1} [C [H1 H3]];
    rewrite (IHt (B::g) C H3); first done.
  by rewrite /= ltnS.
Qed.


Lemma subst_db_id : forall t A, wt nil t A ->
  forall n u, subst_db t n u = t.
Proof.
  move => t A H n u; apply: subst_db_id_aux; [eexact H|done].
Qed.


(* Link between close and open *)

Lemma open_close_aux : forall x X t i,
  subst_db (close_aux t x X i) i (Var x X) = t.
Proof.
  move => x X; elim => [n|y Y|c|c C|t IHt u IHu|A t IHt] i //=.
  - case e: (n < i); first by rewrite e.
  have h: (n.+1 < i = false).
  apply negbTE; rewrite -leqNgt; apply: leq_trans; last eexact (leqnSn n).
  rewrite leqNgt; by apply negbT.
  rewrite h; have h2: (n.+1 == i = false).
  rewrite eq_sym; apply negbTE; have: (n.+1 > i).
  apply: leq_ltn_trans; last eexact (ltnSn n).
  rewrite leqNgt; by apply negbT.
  rewrite (ltn_neqAle  i (n.+1)); by move /andP => [H _].
  by rewrite h2 subn1.
  - case e: (x == y) => /=; last done.
  case f: (X == Y) => /=; last done.
  rewrite ltnn eq_refl; by move /eqP: e ->; move /eqP: f ->.
  - by rewrite IHt IHu.
  - by rewrite IHt.
Qed.


Lemma open_close : forall x X t, open (close t x X) (Var x X) = t.
Proof. move => x X t; apply open_close_aux. Qed.


Lemma ltn_leq_trans : forall n m p : nat, m < n -> n <= p -> m < p.
Proof.
  move => n m p h1 h2; apply (leq_ltn_trans (n:=n.-1));
    last by rewrite (ltn_predK h1).
  by rewrite -(leq_add2r 1 m (n.-1)) !addn1 (ltn_predK h1).
Qed.


Lemma close_aux_open : forall x X t i, is_not_free x X t ->
  close_aux (subst_db t i (Var x X)) x X i = t.
Proof.
  move => x X; elim => [n|y Y|c|c C|t IHt u IHu|A t IHt] //= i H.
  - case e: (n < i) => /=; first by rewrite e.
  case f: (n == i) => /=; first by rewrite !eq_refl /=; move /eqP: f ->.
  have h1: (n > i).
  rewrite ltn_neqAle leqNgt e /=; rewrite eq_sym in f;
    by move: (negbT f) ->.
  have h: (n - 1 < i = false).
  apply negbTE; rewrite -leqNgt subn1 -(leq_add2r 1 i (n.-1)) !addn1;
    apply: ltn_leq_trans; last eexact (leqSpred n); first exact h1.
  by rewrite h subn1 (ltn_predK h1).
  - move /orP: H => {H} [H|H]; first by move: (negbTE H) => -> /=.
  by move: (negbTE H) => ->; rewrite andb_false_r.
  - by move /andP: H => [H1 H2]; rewrite (IHt i H1) (IHu i H2).
  - by rewrite (IHt (i.+1) H).
Qed.


Lemma close_open : forall x X t, is_not_free x X t ->
  close (open t (Var x X)) x X = t.
Proof. move => x X t H; apply (close_aux_open x X t O H). Qed.


(* Link between free variables and open *)

Lemma is_not_free_subst_db : forall x y X Y u i,
  is_not_free y Y u -> ((y != x) || (Y != X)) ->
  is_not_free y Y (subst_db u i (Var x X)).
Proof.
  move => x y X Y; elim => [n|z Z|c|c C|t IHt u IHu|A t IHt] i //=.
  - by move => _ H; case: (_ < _); [|case: (_ == _)].
  - move /andP => [H1 H2] H3; apply andb_true_intro; split;
    [by apply IHt|by apply IHu].
  - apply IHt.
Qed.


Lemma is_not_free_open : forall x y X Y u,
  is_not_free y Y u -> ((y != x) || (Y != X)) ->
  is_not_free y Y (open u (Var x X)).
Proof. move => x y X Y u H1 H2; by apply is_not_free_subst_db. Qed.


(* Interaction with semantics *)

Require Import translation.
Require Import axioms.


(* Interaction between lifting and semantics *)

Lemma sem_lift_1 : forall (g: context) (A: type) (t: term)
  (d: context),
  match sem_term (d++g) t,
    sem_term (d++A::g) (lift_term t (length d) 1) with
    | Some (existT T evT), Some (existT U evU) =>
      match cast T U with
        | Cast k =>
          forall (tc: type_context) (tdt: type_definition_tr)
            (se: term_context tc tdt) (sdt: term_definition_tr tc tdt)
            (r: tr_context tc tdt (d++g)) (a: tr_type tc tdt A),
            k (tr_type tc tdt) (evT tc tdt se sdt r) = evU tc tdt se sdt
            (tr_middle A g d r a)
        | NoCast => False
      end
    | None, None => True
    | _, _ => False
  end.
Proof.
  move => g A;
    elim => [n /=|x X /=|c /=|c B /=|t1 IHt1 t2 IHt2|B t IHt] d.

  - case e: (length d <= n).
  rewrite /sem_term add1n; move: (sem_var_sem2 g A n d e);
    case: (sem_var (d++g) n) => [[N evN]|];
      case: (sem_var (d ++ A :: g) n.+1) => [[M evM]|] //=;
  case: (cast N M) => [k|] //=.
  have h: (n < length d); first by rewrite ltnNge; apply negbT.
  move: (sem_var_sem1 g A n d h);
    case: (sem_var (d++g) n) => [[N evN]|] //=;
      case: (sem_var (d ++ A :: g) n) => [[M evM]|] //=;
        case: (cast N M) => //=.

  - by rewrite cast_same.

  - by rewrite cast_same.

  - by rewrite cast_same.

  - move: {IHt2} (IHt2 d); move: {IHt1} (IHt1 d);
    rewrite [tr_middle]lock => /=;
      case: (sem_term (d++g) t1) => [[[|||B C] evB]|];
        case: (sem_term (d++g) t2) => [[T2 evT2]|];
          case: (sem_term (d ++ A :: g) (lift_term t1 (length d) 1)) =>
            [[[|||D E] evD]|];
            case: (sem_term (d ++ A :: g) (lift_term t2 (length d) 1)) =>
              [[F evF]|] //=; rewrite -lock.
  rewrite neq_cast; case eBD: (_ == _); last done.
  move /eqP: eBD evB => -> evB; rewrite cast_same.
  rewrite neq_cast; case eCE: (_ == _); last done.
  move /eqP: eCE evB => -> evB; rewrite cast_same => IHt1.
  rewrite neq_cast; case eT2F: (_ == _); last done.
  move /eqP: eT2F evT2 => ->; rewrite cast_same.
  rewrite neq_cast; case eFD: (_ == _); last done.
  by move /eqP: eFD evF => -> evF evT2 IHt2;
    rewrite !cast_same => tc tdt se sdt r a;
      rewrite (IHt1 tc tdt se sdt r a) (IHt2 tc tdt se sdt r a).

  - move: IHt; rewrite [tr_middle]lock /= => IHt; move: {IHt} (IHt (B::d));
    case: (sem_term ((B :: d) ++ g) t) => [[T evT]|];
      case: (sem_term ((B :: d) ++ A :: g)
        (lift_term t (length (B :: d)) 1)) => [[U evU]|] //=.
  rewrite cast_same neq_cast -lock; case eTU: (_ == _); last done.
  by move /eqP: eTU evT => -> evT;
    rewrite cast_same => IHt tc tdt se sdt r a; extensionality b;
      rewrite (IHt tc tdt se sdt (tr_cs B r b) a) tr_cs_middle.
Qed.


Lemma sem_lift : forall (u: term) (g d: context),
  match sem_term g u, sem_term (d++g) (lift_term u 0 (length d)) with
    | Some (existT U evU), Some (existT V evV) =>
      match cast V U with
        | Cast k =>
          forall (tc: type_context) (tdt: type_definition_tr)
            (se: term_context tc tdt) (sdt: term_definition_tr tc tdt)
            (r: tr_context tc tdt (d++g)),
            evU tc tdt se sdt (tr_app d g r) = k (tr_type tc tdt) (evV tc tdt se sdt r)
        | NoCast => False
      end
    | None, None => True
    | _, _ => False
  end.
Proof.
  move => u g; elim => [|B d] /=.
  - by rewrite (lift_id u 0); case: (sem_term g u) => [[U evU]|//=];
    rewrite cast_same.
  - rewrite lift_S; case: (sem_term g u) => [[U evU]|];
    last by move: (sem_lift_1 (d++g) B
      (lift_term u 0 (length d)) nil) => /=; case: (sem_term _ _).
  move: (sem_lift_1 (d++g) B (lift_term u 0 (length d)) nil) => /=;
    case: (sem_term (d ++ g) (lift_term u 0 (length d))) => [[T evT]|//=].
  case: (sem_term (B :: d ++ g) (lift_term (lift_term u 0 (length d)) 0 1))
    => [[V evV]|//=].
  rewrite neq_cast; case eTV: (_ == _); last done.
  move /eqP: eTV evT => ->; rewrite cast_same neq_cast;
    case eVU: (_ == _); last done.
  by move /eqP: eVU evV => ->;
    rewrite cast_same => evT evV h1 h2 tc tdt se sdt r;
      rewrite (h2 tc tdt se sdt (tr_tl r))
        (h1 tc tdt se sdt (tr_tl r) (r 0)) tr_cs_tl_ext.
Qed.


(* Substitution lemma for sem_term *)

Lemma sem_subst_db : forall (g: context) (u t: term) (d: context),
  match sem_term g u with
    | Some (existT U evU) =>
      match sem_term (d++g)
        (subst_db t (length d) (lift_term u 0 (length d))),
        sem_term (d++U::g) t with
        | Some (existT T1 evT1), Some (existT T2 evT2) =>
          match cast T1 T2 with
            | Cast k =>
              forall (tc: type_context) (tdt: type_definition_tr)
                (se: term_context tc tdt) (sdt: term_definition_tr tc tdt)
                (r: tr_context tc tdt (d++g)),
                k (tr_type tc tdt) (evT1 tc tdt se sdt r) =
                evT2 tc tdt se sdt (tr_middle U g d r
                  (evU tc tdt se sdt (tr_app d g r)))
            | NoCast => False
          end
        | None, None => True
        | _, _ => False
      end
    | None => True
  end.
Proof.
  move => g u; elim => [n|x X /=|c|c C /=|t1 IHt1 t2 IHt2|A t IHt] d.

  - case eU: (sem_term _ _) => //= [[U evU]].
  case e: (n < length d) => /=.
  move: (sem_var_sem1 g U n d e); case: (sem_var _ _) => [[N evN]|//=].
  case: (sem_var _ _) => [[M evM]|//=].
  case: (cast _ _) => [k|//=].
  move => sv tc tdt se sdt r; apply: sv.
  case e2: (n == length d).
  move: {e} e2; move /eqP ->; move: (sem_lift u g d); rewrite eU;
    case: (sem_term _ _) => [[T1 evT1]|//=].
  case cast_T1_U: (cast _ _) => [k|//=].
  by move => h; rewrite (sem_var_sem3 d g U) cast_T1_U => tc tdt se sdt r;
    rewrite (h tc tdt se sdt r) tr_middle_app.
  rewrite subn1; have h: (length d < n).
  have H: (~~(n < length d)) by apply: negbT.
  rewrite <- leqNgt in H;
    have H0: (length d != n) by apply: negbT; rewrite eq_sym.
  by rewrite ltn_neqAle; apply: andb_true_intro.
  have h2: (length d <= n.-1).
  by rewrite -(succnK (length d)) -!subn1; apply leq_sub2r.
  move: (sem_var_sem2 g U (n.-1) d h2);
    replace (sem_var (d++U::g) n.-1.+1) with (sem_var (d++U::g) n);
      last by rewrite (ltn_predK h).
  rewrite /sem_term;
    case: (sem_var _ _) => [[N evN]|//=].
  case: (sem_var _ _) => [[M evM]|//=].
  case: (cast _ _) => [k|//=].
  move => IHn tc tdt se sdt r; apply: (IHn tc tdt r).
  by case: (sem_var _ _).

  - by case: (sem_term _ _) => //= [[U evU]]; rewrite cast_same.

  - by case: (sem_term _ _) => //= [[U evU]]; rewrite cast_same.

  - by case: (sem_term _ _) => //= [[U evU]]; rewrite cast_same.

  - move: {IHt2} (IHt2 d); move: {IHt1} (IHt1 d);
    case: (sem_term g u) => //= [[U evU]].
  case: (sem_term (d ++ g)
    (subst_db t1 (length d) (lift_term u 0 (length d)))) =>
  [[[|||A B] evT1]|];
  case: (sem_term (d ++ U :: g) t1) => [[[|||C D] ev1]|] //=.
  rewrite neq_cast; case eAC: (A == C); last done.
  move /eqP: eAC ev1 => <- ev1 {C}; rewrite cast_same;
    rewrite neq_cast; case eBD: (B == D); last done.
  move /eqP: eBD ev1 => <- ev1 {D}; rewrite cast_same => IHt;
    case: (sem_term (d ++ g) (subst_db t2 (length d) (lift_term u 0
      (length d)))) => [[T2 evT2]|];
    case: (sem_term (d ++ U :: g) t2) => [[T3 evT3]|] //=.
  rewrite neq_cast; case e23: (T2 == T3); last done.
  move /eqP: e23 evT3 => <- evT3 {T3}; rewrite cast_same => IHu;
    rewrite neq_cast; case e2A: (T2 == A); last done.
  move /eqP: e2A evT3 evT2 IHu => -> evT3 evT2 IHu {T2}; rewrite !cast_same;
    move => tc tdt se sdt r; by rewrite IHt IHu.
  - move: {IHt} (IHt (A::d)); rewrite [tr_middle]lock /=;
    case: (sem_term _ _) => //= [[U evU]].
  rewrite -lift_S; case: (sem_term _ _) => [[T evT]|];
    case: (sem_term _ _) => [[T2 evT2]|] //=.
  rewrite cast_same neq_cast; case eTT2: (_ == _); last done.
  by move /eqP: eTT2 evT => -> evT;
    rewrite cast_same -lock => IHt tc tdt se sdt r; extensionality a;
      rewrite (IHt tc tdt se sdt (tr_cs A r a)) tr_cs_middle tr_tl_cs_ext.
Qed.


(* Semantics and open *)

Lemma sem_open : forall (g: context) (t u: term),
  match sem_term g t with
    | Some (existT B evB) =>
      match sem_term g (open u t), sem_term (B::g) u with
        | Some (existT U1 ev1), Some (existT U2 ev2) =>
          match cast U1 U2 with
            | Cast k => forall tc tdt se sdt r,
              k (tr_type tc tdt) (ev1 tc tdt se sdt r) = ev2 tc tdt se sdt
              (tr_cs B r (evB tc tdt se sdt r))
            | NoCast => False
          end
        | None, None => True
        | _, _ => False
      end
    | None => True
  end.
Proof.
  move => g t u; move: (sem_subst_db g t u nil) => /=;
    case: (sem_term g t) => [[T evT]|] //=.
  rewrite lift_id /open;
    by case: (sem_term g (subst_db u 0 t)) => [[U evU]|].
Qed.

Unset Implicit Arguments.
