(* ========================================================================= *)
(* Substitutions for named variables                                         *)
(*                                                                           *)
(*       Chantal Keller, Laboratoire d'Informatique de Polytechnique         *)
(*                                                                           *)
(*    (c) Copyright, Laboratoire d'Informatique de Polytechnique 2010        *)
(*              (c) Copyright, Chantal Keller 2010                           *)
(* ========================================================================= *)


Require Import ssreflect ssrbool eqtype ssrnat.
Require Import axioms hol cast typing translation subst_db.
Require Import List.

Set Implicit Arguments.


(* Type substitutions *)

Definition substitution_idt := list (idT * type).


Fixpoint subst_idt_type_aux X s :=
  match s with
    | nil => TVar X
    | (Y,A)::q => if X == Y then A else subst_idt_type_aux X q
  end.


Definition subst_idt_type_aux_inhabited tc tdt X S :
  tr_type tc tdt (subst_idt_type_aux X S) :=
  (tr_type tc tdt (subst_idt_type_aux X S)).(tinhab).


Fixpoint subst_idt_type (T: type) (s: substitution_idt) {struct T} : type :=
  match T with
    | TVar X => subst_idt_type_aux X s
    | TDef a l => TDef a (subst_idt_list_type l s)
    | A --> B => (subst_idt_type A s) --> (subst_idt_type B s)
    | u => u
  end

with subst_idt_list_type (l: list_type) (s: substitution_idt) {struct l} :
  list_type :=
  match l with
    | Tnil => Tnil
    | Tcons A m => Tcons (subst_idt_type A s) (subst_idt_list_type m s)
  end.


Fixpoint subst_idt (t: term) (s: substitution_idt) {struct t} : term :=
  match t with
    | Var x X         => Var x (subst_idt_type X s)
    | Cst (Heq A)     => Cst (Heq (subst_idt_type A s))
    | Cst (Heps A)    => Cst (Heps (subst_idt_type A s))
    | Cst (Hforall A) => Cst (Hforall (subst_idt_type A s))
    | Cst (Hexists A) => Cst (Hexists (subst_idt_type A s))
    | Def c C         => Def c (subst_idt_type C s)
    | App t1 t2       => App (subst_idt t1 s) (subst_idt t2 s)
    | Abs A t         => Abs (subst_idt_type A s) (subst_idt t s)
    | u               => u
  end.


Definition subst_idt_context (g: context) (s: substitution_idt) : context :=
  map (fun A => subst_idt_type A s) g.


(* Properties *)

Section Type_substitutions.

  Variable S: substitution_idt.


  (* Substitution lemma for wt *)

  Lemma wt_subst_idt : forall (t: term) (g: context),
    match infer g t, infer (subst_idt_context g S) (subst_idt t S) with
      | Some A, Some B => (subst_idt_type A S) == B
      | None, _ => true
      | _, _ => false
    end.
  Proof.
    elim => [|x X|[C|C|||||||C|C]|c C|t IHt u IHu|A t IHt] //=.
    - elim => [|n IHn] [|] //=.
    - move => g; move: (IHt g) (IHu g); case: (infer _ _) => [[|||A B]|];
      case: (infer _ _) => [[|||C D]|] //= => H; move /eqP: H => [<- <-];
        case: (infer _ _) => //= E ; case: (infer _ _) => //= F => H;
          move /eqP: H <-; case e: (_ == _) => //=; move /eqP: e ->;
            by rewrite eq_refl.
    - move => g; move: (IHt (A::g)); case: (infer _ _) => //= B;
      case: (infer _ _) => //= C => H; by move /eqP: H <-.
  Qed.


  Lemma wt_subst_idt_inv2 : forall g t A, wt g t A ->
    wt (subst_idt_context g S) (subst_idt t S) (subst_idt_type A S).
  Proof.
    rewrite /wt => g t A H; move: (wt_subst_idt t g);
      rewrite H; case: (infer _ _); last done;
        by move => B; move /eqP <-.
  Qed.


  (* Type substitutions commute with opening *)

  Lemma subst_idt_lift_term : forall j u i,
    subst_idt (lift_term u i j) S = lift_term (subst_idt u S) i j.
  Proof.
    move => j; elim => [n|x X|[C|C|||||||C|C]|c C|t IHt u IHu|A t IHt] //=
      i.
    - by case: (_ <= _).
    - by rewrite IHt IHu.
    - by rewrite IHt.
  Qed.


  Lemma subst_idt_subst_db : forall t i u, subst_idt (subst_db t i u) S =
    subst_db (subst_idt t S) i (subst_idt u S).
  Proof.
    elim => [n|x X|[C|C|||||||C|C]|c C|t IHt v IHv|A t IHt] //=
      i u.
    - by case: (_ < _); [|case: (_ == _)].
    - by rewrite IHt IHv.
    - by rewrite IHt subst_idt_lift_term.
  Qed.


  Lemma subst_idt_open : forall u t,
    subst_idt (open u t) S = open (subst_idt u S) (subst_idt t S).
  Proof.
    by move => u t; apply subst_idt_subst_db.
  Qed.


  (* Substitution lemma for sem_term *)

  Definition subst_idt_type_context (tc: type_context)
    (tdt: type_definition_tr) : type_context :=
    fun X => tr_type tc tdt (subst_idt_type_aux X S).


  (* It commutes with tr_type *)

  Lemma tr_type_subst_idt : forall (tc: type_context)
    (tdt: type_definition_tr) (A: type),
    tr_type (subst_idt_type_context tc tdt) tdt A =
    tr_type tc tdt (subst_idt_type A S).
  Proof.
    move => tc tdt; fix IH 1; move => [X|d l||A B] //=.
    - replace (tr_list_type (subst_idt_type_context tc tdt) tdt l) with
      (tr_list_type tc tdt (subst_idt_list_type l S));
      first done.
    elim: l => //= C l IHl.
    by rewrite (IH C) IHl.
    - by rewrite (IH A) (IH B).
  Defined.


  (* So we have a conversion "function" *)

  Definition conv (tc: type_context) (tdt: type_definition_tr) (A: type) :
    tr_type tc tdt (subst_idt_type A S) ->
    tr_type (subst_idt_type_context tc tdt) tdt A.
  Proof. move => tc tdt A; by rewrite tr_type_subst_idt. Defined.


  (* conv is surjective *)

  Lemma surj_conv : forall tc tdt A
    (a: tr_type (subst_idt_type_context tc tdt) tdt A),
    exists b, a = conv tc tdt A b.
  Proof.
    rewrite /conv => tc tdt A;
      move: (tr_type_subst_idt tc tdt A);
        rewrite tr_type_subst_idt => h a;
          exists a; by replace h with (refl_equal (tr_type tc tdt
            (subst_idt_type A S))); last (symmetry; apply axiom_K).
  Qed.


  Definition subst_idt_term_context (tc: type_context)
    (tdt: type_definition_tr) (se: term_context tc tdt) :
    term_context (subst_idt_type_context tc tdt) tdt :=
    fun x X => conv tc tdt X (se x (subst_idt_type X S)).


  Definition subst_idt_term_definition_tr (tc: type_context)
    (tdt: type_definition_tr) (sdt: term_definition_tr tc tdt) :
    term_definition_tr (subst_idt_type_context tc tdt) tdt :=
    fun x X => conv tc tdt X (sdt x (subst_idt_type X S)).


  Definition subst_idt_tr_context (g: context) (tc: type_context)
    (tdt: type_definition_tr) : tr_context tc tdt (subst_idt_context g S) ->
    tr_context (subst_idt_type_context tc tdt) tdt g.
  Proof.
    move => g tc tdt r n; move: (r n);
      rewrite /subst_idt_context -opt_nth_map;
        case: (opt_nth n g) => [A|]; last exact (fun x => x).
    exact (conv tc tdt A).
  Defined.


  (* Commutation lemmas *)

  Lemma lem1 : forall g A tc tdt
    (r: tr_context tc tdt (subst_idt_context (A::g) S)),
    subst_idt_tr_context (A :: g) r 0 = conv tc tdt A (r 0).
  Proof.
    done.
  Qed.


  Lemma lem2 : forall g A tc tdt (r: tr_context tc tdt (subst_idt_context
    (A::g) S)), tr_tl (subst_idt_tr_context (A :: g) r) =
  subst_idt_tr_context g (tr_tl r).
  Proof.
    done.
  Qed.


  Lemma lem3 : forall tc tdt A B ab a,
    conv tc tdt (A --> B) ab (conv tc tdt A a) = conv tc tdt B (ab a).
  Proof.
    rewrite /conv => tc tdt A B ab a;
      move: (tr_type_subst_idt tc tdt (A -->B))
        (tr_type_subst_idt tc tdt B) (tr_type_subst_idt tc tdt A) => /=;
        change ((fun pv => (forall (tr_type_subst_idt0 : @eq
  type_translation (@mkTT (ttrans (tr_type (subst_idt_type_context tc
  tdt) tdt A) -> ttrans pv) (fun _ : ttrans (tr_type
  (subst_idt_type_context tc tdt) tdt A) => tinhab pv)) (@mkTT (ttrans
  (tr_type tc tdt (subst_idt_type A S)) -> ttrans (tr_type tc tdt
  (subst_idt_type B S))) (fun _ : ttrans (tr_type tc tdt (subst_idt_type
  A S)) => tinhab (tr_type tc tdt (subst_idt_type B S)))))
  (tr_type_subst_idt1 : @eq type_translation pv (tr_type tc tdt
  (subst_idt_type B S))) (tr_type_subst_idt2 : @eq type_translation
  (tr_type (subst_idt_type_context tc tdt) tdt A) (tr_type tc tdt
  (subst_idt_type A S))), @eq (ttrans pv) (@eq_rect_r type_translation
  (@mkTT (ttrans (tr_type tc tdt (subst_idt_type A S)) -> ttrans
  (tr_type tc tdt (subst_idt_type B S))) (fun _ : ttrans (tr_type tc tdt
  (subst_idt_type A S)) => tinhab (tr_type tc tdt (subst_idt_type B
  S)))) (fun pattern_value : type_translation => (ttrans (tr_type tc tdt
  (subst_idt_type A S)) -> ttrans (tr_type tc tdt (subst_idt_type B S)))
  -> ttrans pattern_value) (fun X : ttrans (tr_type tc tdt
  (subst_idt_type A S)) -> ttrans (tr_type tc tdt (subst_idt_type B S))
  => X) (@mkTT (ttrans (tr_type (subst_idt_type_context tc tdt) tdt A)
  -> ttrans pv) (fun _ : ttrans (tr_type (subst_idt_type_context tc tdt)
  tdt A) => tinhab pv)) tr_type_subst_idt0 ab (@eq_rect_r
  type_translation (tr_type tc tdt (subst_idt_type A S)) (fun
  pattern_value : type_translation => ttrans (tr_type tc tdt
  (subst_idt_type A S)) -> ttrans pattern_value) (fun X : ttrans
  (tr_type tc tdt (subst_idt_type A S)) => X) (tr_type
  (subst_idt_type_context tc tdt) tdt A) tr_type_subst_idt2 a))
  (@eq_rect_r type_translation (tr_type tc tdt (subst_idt_type B S))
  (fun pattern_value : type_translation => ttrans (tr_type tc tdt
  (subst_idt_type B S)) -> ttrans pattern_value) (fun X : ttrans
  (tr_type tc tdt (subst_idt_type B S)) => X) pv tr_type_subst_idt1 (ab
  a)))) (tr_type (subst_idt_type_context tc tdt) tdt B));
        rewrite (tr_type_subst_idt tc tdt B) tr_type_subst_idt => h1 h2 h3.
    replace h1 with (refl_equal (mkTT
      (fun _ : tr_type tc tdt (subst_idt_type A S) =>
        tinhab (tr_type tc tdt (subst_idt_type B S)))));
    last (symmetry; apply axiom_K).
    replace h2 with (refl_equal (tr_type tc tdt (subst_idt_type B S)));
      last (symmetry; apply axiom_K).
    by replace h3 with (refl_equal (tr_type tc tdt (subst_idt_type A S)));
      last (symmetry; apply axiom_K).
  Qed.


  Lemma lem4 : forall tc tdt A g r a,
    tr_cs A (subst_idt_tr_context g r) (conv tc tdt A a) =
    subst_idt_tr_context (A::g) (tr_cs (subst_idt_type A S) r a).
  Proof.
    by move => tc tdt A g r a; extensionality n;
      rewrite /subst_idt_tr_context /tr_cs /=; case n.
  Qed.


  (* Substitution lemma for sem_var *)

  Lemma sem_var_subst_idt : forall (n: nat) (g: context),
    match sem_var g n, sem_var (subst_idt_context g S) n with
      | Some (existT A evA), Some (existT B evB) =>
        match cast B (subst_idt_type A S) with
          | Cast k => forall tc tdt r,
            evA
            (subst_idt_type_context tc tdt)
            tdt
            (subst_idt_tr_context g r)
            = conv tc tdt A (k (tr_type tc tdt) (evB tc tdt r))
          | NoCast => false
        end
      | None, _ => true
      | _, _ => false
    end.
  Proof.
    elim => [|n IHn] [|A g] //=.
    - rewrite cast_same; move => tc tdt r; apply lem1.
    - move: (IHn g) => {IHn}; case: (sem_var g n) => [[B evB]|] //=.
    case: (sem_var (subst_idt_context g S) n) => [[C evC]|] //=.
    case: (cast C (subst_idt_type B S)) => [k|] //=.
    by move => IH tc tdt r; rewrite <- (IH tc tdt (tr_tl r));
      rewrite lem2.
  Qed.


  (* Useful lemma in the epsilon case of the substitution lemma for
     sem_term *)

  Lemma inh_eq : forall (A toto: type_translation),
    inhabited A -> (toto = A) -> inhabited toto.
  Proof. by move => A toto h1 ->. Defined.


  (* Substitution lemma for sem_term *)

  Lemma sem_subst_idt : forall (t: term) (g: context),
    match sem_term g t,
      sem_term (subst_idt_context g S) (subst_idt t S) with
      | Some (existT A evA), Some (existT B evB) =>
        match cast B (subst_idt_type A S) with
          | Cast k => forall tc tdt se sdt r,
            evA
              (subst_idt_type_context tc tdt)
              tdt
              (subst_idt_term_context se)
              (subst_idt_term_definition_tr sdt)
              (subst_idt_tr_context g r)
              = conv tc tdt A (k (tr_type tc tdt) (evB tc tdt se sdt r))
          | NoCast => false
        end
      | None, _ => true
      | _, _ => false
    end.
  Proof.
    elim => [n|x X|[C|C|||||||C|C]|c C|t IHt u IHu|A t IHt] g //=.
    - move: (sem_var_subst_idt n g).
    case: (sem_var g n) => [[A evA]|] //=.
    case: (sem_var (subst_idt_context g S) n) => [[B evB]|] //=.
    case: (cast B (subst_idt_type A S)) => [k|] //=.
    - by rewrite cast_same.
    - rewrite cast_same /conv => tc tdt _ _ _.
      move: (tr_type_subst_idt tc tdt (Arrow C (Arrow C Bool))) => /=.
        rewrite -tr_type_subst_idt => h.
          replace h with (refl_equal (mkTT (fun _ _ : tr_type (subst_idt_type_context tc tdt) tdt C => True))); first done; symmetry; apply axiom_K.
    - rewrite cast_same /conv => tc tdt _ _ _;
      move: (tr_type_subst_idt tc tdt (Arrow (Arrow C Bool) C)) => /=;
        rewrite /eps;
          move: (refl_equal (tr_type tc tdt (subst_idt_type C S))) => h;
            replace (inhabited_tr_type tc tdt (subst_idt_type C S)) with
              (inh_eq (inhabited_tr_type tc tdt (subst_idt_type C S)) h).
    move: h;
    change ((fun pv => forall (h : @eq type_translation pv (tr_type tc
  tdt (subst_idt_type C S))) (tr_type_subst_idt0 : @eq type_translation
  (@mkTT ((ttrans (tr_type (subst_idt_type_context tc tdt) tdt C) ->
  Prop) -> ttrans (tr_type (subst_idt_type_context tc tdt) tdt C)) (fun
  _ : ttrans (tr_type (subst_idt_type_context tc tdt) tdt C) -> Prop =>
  tinhab (tr_type (subst_idt_type_context tc tdt) tdt C))) (@mkTT
  ((ttrans pv -> Prop) -> ttrans pv) (fun _ : ttrans pv -> Prop =>
  tinhab pv))), @eq ((ttrans (tr_type (subst_idt_type_context tc tdt)
  tdt C) -> Prop) -> ttrans (tr_type (subst_idt_type_context tc tdt) tdt
  C)) (@epsilon (ttrans (tr_type (subst_idt_type_context tc tdt) tdt C))
  (inhabited_tr_type (subst_idt_type_context tc tdt) tdt C)) (@eq_rect_r
  type_translation (@mkTT ((ttrans pv -> Prop) -> ttrans pv) (fun _ :
  ttrans pv -> Prop => tinhab pv)) (fun pattern_value : type_translation
  => ((ttrans pv -> Prop) -> ttrans pv) -> ttrans pattern_value) (fun X
  : (ttrans pv -> Prop) -> ttrans pv => X) (@mkTT ((ttrans (tr_type
  (subst_idt_type_context tc tdt) tdt C) -> Prop) -> ttrans (tr_type
  (subst_idt_type_context tc tdt) tdt C)) (fun _ : ttrans (tr_type
  (subst_idt_type_context tc tdt) tdt C) -> Prop => tinhab (tr_type
  (subst_idt_type_context tc tdt) tdt C))) tr_type_subst_idt0 (@epsilon
  (ttrans pv) (@inh_eq (tr_type tc tdt (subst_idt_type C S)) pv
  (inhabited_tr_type tc tdt (subst_idt_type C S)) h)))) (tr_type tc tdt
  (subst_idt_type C S)));
    rewrite -{3}(tr_type_subst_idt tc tdt C) => h h0;
      replace h0 with (refl_equal (mkTT
        (fun _ : tr_type (subst_idt_type_context tc tdt) tdt C -> Prop =>
          tinhab (tr_type (subst_idt_type_context tc tdt) tdt C))));
      last (symmetry; apply axiom_K).
    replace (inh_eq (inhabited_tr_type tc tdt (subst_idt_type C S)) h) with
      (inhabited_tr_type (subst_idt_type_context tc tdt) tdt C); first done.
    apply Classical_Prop.proof_irrelevance.
    by rewrite /inh_eq;
      replace h with (refl_equal (tr_type tc tdt (subst_idt_type C S)));
        last (symmetry; apply axiom_K).
    - rewrite cast_same /conv => tc tdt _ _ _;
      move: (tr_type_subst_idt tc tdt (Arrow (Arrow C Bool) Bool)) => /=;
        rewrite -tr_type_subst_idt => h;
          replace h with (refl_equal (mkTT
        (fun _ : tr_type (subst_idt_type_context tc tdt) tdt C -> Prop =>
         True))); first done; symmetry;
          apply axiom_K.
    - rewrite cast_same /conv => tc tdt _ _ _;
      move: (tr_type_subst_idt tc tdt (Arrow (Arrow C Bool) Bool)) => /=;
        rewrite -tr_type_subst_idt => h;
          replace h with (refl_equal (mkTT
        (fun _ : tr_type (subst_idt_type_context tc tdt) tdt C -> Prop =>
         True))); first done; symmetry;
          apply axiom_K.
    - rewrite cast_same => tc tdt _ sdt _;
      rewrite /subst_idt_term_definition_tr /conv;
        move: (tr_type_subst_idt tc tdt C) => /=;
          rewrite tr_type_subst_idt => h2; replace h2 with (refl_equal
            (tr_type tc tdt (subst_idt_type C S))); first done;
          symmetry; apply axiom_K.
    - move: (IHt g) (IHu g) => {IHt IHu};
      case: (sem_term g t) => [[[|||A B] evA]|];
        case: (sem_term g u) => [[C evC]|] //=.
    case: (sem_term _ _) => [[[|||D E] evD]|];
      case: (sem_term _ _) => [[F evF]|] //=.
    rewrite neq_cast; case e: (_ == _); last done.
    move /eqP: e evD => -> {D} evD; rewrite cast_same neq_cast;
      case e: (_ == _); last done.
    move /eqP: e evD => -> {E} evD; rewrite cast_same => IHt;
      rewrite neq_cast; case e: (_ == _); last done.
    move /eqP: e evF => -> {F} evF; rewrite cast_same => IHu;
      rewrite neq_cast; case e: (_ == _); last done.
    move /eqP: e evC evF IHu => -> {C} evC evF IHu; rewrite !cast_same
      => tc tdt se sdt r; rewrite IHt IHu; apply lem3.
    - move: (IHt (A::g)) => {IHt};
      case: (sem_term (A :: g) t) =>  //= [[B evB]].
    case: (sem_term _ _) => //= [[C evC]].
    rewrite neq_cast; case e: (_ == _); last done.
    move /eqP: e evC => -> {C} evC; rewrite !cast_same =>
      IHt tc tdt se sdt r; extensionality a;
        move: (surj_conv tc tdt A a) => [b ->] {a};
          by rewrite lem3 lem4 IHt.
  Qed.

End Type_substitutions.












(* Term substitutions *)

Definition substitution_idv := list (idV * type * term).


Fixpoint subst_idv_aux x X s :=
  match s with
    | nil => Var x X
    | (y, Y, u)::q =>
      if ((x == y) && (X == Y)) then u else subst_idv_aux x X q
  end.


Fixpoint subst_idv (t: term) (s: substitution_idv) {struct t} : term :=
  match t with
    | Var x X => subst_idv_aux x X s
    | App t1 t2 => App (subst_idv t1 s) (subst_idv t2 s)
    | Abs A t => Abs A (subst_idv t s)
    | u => u
  end.


Fixpoint wf_substitution_idv (s: substitution_idv) :=
  match s with
    | nil => true
    | (_,X,t)::q =>
      match infer nil t with
        | Some Y => if (X == Y) then wf_substitution_idv q else false
        | None => false
      end
  end.


(* Domain of a term substitution *)

Fixpoint dom_substitution_idv (s: substitution_idv) : list idV :=
  match s with
    | nil => nil
    | (x,_,_)::q => x::(dom_substitution_idv q)
  end.


(* Properties *)

Section Term_substitutions.

  Variable s: substitution_idv.


  (* If a variable is not in the domain of a substitution, it is not
     affected by it *)

  Lemma dom_substitution_idv_id : forall x X,
    forallb (fun y : idV => y != x) (dom_substitution_idv s) ->
    subst_idv (Var x X) s = Var x X.
  Proof.
    move => x X; move: s; elim => [|[[y Y] t] q IHq] //=.
    - move /andP => [H1 H2]; move: (negbTE H1);
      rewrite eq_sym => -> /=; by apply IHq.
  Qed.


  Variable hs : wf_substitution_idv s.


  Lemma wf_substitution_idv_close : forall x X,
    wt nil (subst_idv_aux x X s) X.
  Proof.
    move => x X; move: s hs; elim => [|[[y Y] t] q IHq] //=.
    case e1: (infer nil t); last done.
    case e3: (_ == _); last done.
    move => H; case: (_ == _); last by apply IHq.
    case e2: (_ == _); last by apply IHq.
    move /eqP: e2 ->; rewrite /= /wt e1; by move /eqP: e3 ->.
  Qed.


  (* Substitution lemma for wt *)

  Lemma wt_subst_idv : forall t g,
    match infer g t, infer g (subst_idv t s) with
      | Some A, Some B => A == B
      | None, None => true
      | _, _ => false
    end.
  Proof.
    elim => [|x X|[C|C|||||||C|C]|c C|t IHt u IHu|A t IHt] //=.
    - elim => [|n IHn] [|] //=.
    - move => g;
      by rewrite (weakening g (wf_substitution_idv_close x X)).
    - move => g; move: (IHt g) (IHu g); case: (infer _ _) => [[|||A B]|];
      case: (infer _ _) => [[|||C D]|] //=; move /eqP => [<- <-];
        case: (infer _ _) => [E|]; case: (infer _ _) => [F|] //=;
          move /eqP => <-; by case: (_ == _).
    - move => g; move: (IHt (A::g)); case: (infer _ _) => [B|];
      case: (infer _ _) => [C|] //=; by move /eqP => ->.
  Qed.


  Lemma wt_subst_idv_inv2 : forall g t A, wt g t A ->
    wt g (subst_idv t s) A.
  Proof.
    rewrite /wt => g t A H; move: (wt_subst_idv t g);
      rewrite H; case: (infer _ _); last done;
        by move => B; move /eqP <-.
  Qed.


  (* Well-formed term substitutions commute with opening *)

  Lemma subst_idv_lift_term : forall j u i, subst_idv (lift_term u i j) s =
    lift_term (subst_idv u s) i j.
  Proof.
    move => j; elim => [n|x X|[C|C|||||||C|C]|c C|t IHt u IHu|A t IHt] //=
      i.
    - by case: (_ <= _).
    - by rewrite (lift_term_id (A:=X) (wf_substitution_idv_close x X)).
    - by rewrite IHt IHu.
    - by rewrite IHt.
  Qed.


  Lemma subst_idv_subst_db : forall t i u, subst_idv (subst_db t i u) s =
    subst_db (subst_idv t s) i (subst_idv u s).
  Proof.
    elim => [n|x X|[C|C|||||||C|C]|c C|t IHt v IHv|A t IHt] //= i u.
    - by case: (_ < _); [|case: (_ == _)].
    - by rewrite (subst_db_id (A:=X) (wf_substitution_idv_close x X)).
    - by rewrite IHt IHv.
    - by rewrite IHt subst_idv_lift_term.
  Qed.


  Lemma subst_idv_open : forall u t,
    subst_idv (open u t) s = open (subst_idv u s) (subst_idv t s).
  Proof.
    by move => u t; apply subst_idv_subst_db.
  Qed.


  (* Substitution lemma for sem_term *)

  Definition subst_idv_term_context (tc: type_context)
    (tdt: type_definition_tr) (se: term_context tc tdt)
    (sdt: term_definition_tr tc tdt) : term_context tc tdt :=
    fun x X => match wt_sem_type (wf_substitution_idv_close x X) with
                 | exist ev _ => ev tc tdt se sdt (tr_nil tc tdt)
               end.


  Lemma sem_subst : forall (t: term) (g: context),
    match sem_term g t, sem_term g (subst_idv t s) with
      | Some (existT A evA), Some (existT B evB) =>
        match cast A B with
          | Cast k => forall tc tdt se sdt r,
            k (tr_type tc tdt) (evA tc tdt (subst_idv_term_context se sdt) sdt r)
            = evB tc tdt se sdt r
          | NoCast => false
        end
      | None, _ => true
      | _, _ => false
    end.
  Proof.
    elim => [n|x X|[C|C|||||||C|C]|c C|t IHt v IHv|A t IHt] g //=;
      try by rewrite cast_same.
    - case: (sem_var _ _) => [[A evA]|] //=; by rewrite cast_same.
    - rewrite /subst_idv_term_context;
      move: (wt_sem_type (wf_substitution_idv_close x X)) => [ev h];
        move: (sem_weak g (subst_idv_aux x X s) nil);
          rewrite h /=;
            case: (sem_term _ _) => [[B evB]|] //=;
              case: (cast _ _) => [k|] //= => h1 tc tdt se sdt r;
                move: (h1 tc tdt se sdt r);
                  by rewrite tr_nil_weak.
    - move: (IHt g) (IHv g) => {IHt IHv};
      case: (sem_term _ _) => [[[|||A B] evA]|];
        case: (sem_term _ _) => [[[|||C D] evC]|] //=;
          rewrite neq_cast; case e: (_ == _); last done;
            move /eqP: e evC => <- {C} evC; rewrite cast_same;
              rewrite neq_cast; case e: (_ == _); last done;
                move /eqP: e evC => <- {D} evC; rewrite cast_same => IHt;
                  case: (sem_term _ _) => [[E evE]|] //=;
                    case: (sem_term _ _) => [[F evF]|] //=;
                      rewrite neq_cast; case e: (_ == _); last done;
                        move /eqP: e evF => <- {F} evF;
                          rewrite cast_same => IHu;
                            rewrite neq_cast; case e: (_ == _); last done;
                              move /eqP: e evE evF IHu => -> {E} evE evF
                                IHu; rewrite !cast_same => tc tdt se sdt r;
                                  by rewrite IHt IHu.
    - move: (IHt (A::g)) => {IHt}; case: (sem_term _ _) => [[B evB]|] //=;
      case: (sem_term _ _) => [[C evC]|] //=;
        rewrite neq_cast; case e: (_ == _); last done;
          move /eqP: e evC => <- {C} evC; rewrite !cast_same
            => IHt tc tdt se sdt r; extensionality a; by rewrite IHt.
  Qed.

End Term_substitutions.

Unset Implicit Arguments.
