(* ========================================================================= *)
(* Translation of HOL-Light terms from the deep embedding to the shallow one *)
(*                                                                           *)
(*       Chantal Keller, Laboratoire d'Informatique de Polytechnique         *)
(*                                                                           *)
(*    (c) Copyright, Laboratoire d'Informatique de Polytechnique 2010        *)
(*              (c) Copyright, Chantal Keller 2010                           *)
(* ========================================================================= *)


Require Import ssreflect ssrbool eqtype ssrnat.
Require Import axioms hol cast typing.
Require Import List.

Set Implicit Arguments.


(* Type contexts *)

Record type_translation : Type := mkTT
  { ttrans :> Type;
    tinhab : ttrans}.

Definition type_context := idT -> type_translation.
Definition type_definition_tr :=
  defT -> list type_translation -> type_translation.


(* Translation from HOL to Coq: types *)

Fixpoint tr_type (tc: type_context) (tdt: type_definition_tr) (ty: type) :
  type_translation :=
  match ty with
    | TVar a => tc a
    | TDef x l => tdt x (tr_list_type tc tdt l)
    | Bool => mkTT True
    | t1 --> t2 =>
      mkTT (fun _: tr_type tc tdt t1 => (tr_type tc tdt t2).(tinhab))
  end

with tr_list_type (tc: type_context) (tdt: type_definition_tr) (l: list_type) : list type_translation :=
  match l with
    | Tnil => nil
    | Tcons A m => cons (tr_type tc tdt A) (tr_list_type tc tdt m)
  end.


(* How to define type contexts *)

Definition list_tc := list (idT * type_translation)%type.


Fixpoint list_tc2tc (l: list_tc) (X: idT) {struct l} :
  type_translation :=
  match l with
    | nil => mkTT tt
    | (Y, tr)::q => if (Y == X) then tr else list_tc2tc q X
  end.


Definition list_tdt := list (defT * (list type_translation -> type_translation))%type.


Fixpoint list_tdt2tdt (l: list_tdt) (X: defT) (t: list type_translation) {struct l} :
  type_translation :=
  match l with
    | nil => mkTT tt
    | (Y, tr)::q => if (Y == X) then tr t else list_tdt2tdt q X t
  end.


Section Translation.

  Variable tc : type_context.
  Variable tdt : type_definition_tr.


  (* Inhabitants *)

  Definition inhabitant_tr_type (A: type) : tr_type tc tdt A :=
    (tr_type tc tdt A).(tinhab).


  Definition inhabited_tr_type (A: type) : inhabited (tr_type tc tdt A) :=
    inhabits (inhabitant_tr_type A).


  (* Term contexts *)

  Definition term_context := idV -> forall (A: type), tr_type tc tdt A.
  Definition term_definition_tr := defV -> forall (A: type), tr_type tc tdt A.


  (* Translation from HOL to Coq: contexts *)

  Definition tr_context (g: context) : Type :=
    forall (n: nat), match opt_nth n g with
                       | Some ty => (tr_type tc tdt ty).(ttrans)
                       | _ => unit
                     end.


  (* Translation from HOL to Coq: constants *)
  (* Epsilon *)

  Definition eps (A: type) := epsilon (inhabited_tr_type A).


  Definition tr_cst (c: cst) : tr_type tc tdt (type_cst c) :=
    match c as x return tr_type tc tdt (type_cst x) with
      | Heq A => eq (A:=tr_type tc tdt A)
      | Heps A => eps A
      | Hand => and
      | Hor => or
      | Hnot => not
      | Himp => (fun p q => p -> q)
      | Htrue => True
      | Hfalse => False
      | Hforall A => all (A:=tr_type tc tdt A)
      | Hexists A => ex (A:=tr_type tc tdt A)
    end.


  (* Utilities to define semantics *)

  Section Utilities.

    Variable A: type.
    Variable g: context.


    Definition tr_tl (f: tr_context (A::g)) : tr_context g :=
      fun n => f (S n).


    Definition tr_cs (f: tr_context g) (a: tr_type tc tdt A) :
      tr_context (A::g) :=
      fun n => match n as n return (match opt_nth n (A::g) with
                                      | Some B => tr_type tc tdt B
                                      | None => unit
                                    end) with
                 | O => a
                 | S n => f n
               end.


    Lemma tr_tl_cs : forall (f: tr_context g) (a: tr_type tc tdt A) (n: nat),
      tr_tl (tr_cs f a) n = f n.
    Proof. by case 3. Qed.


    Lemma tr_tl_cs_ext : forall (f: tr_context g) (a: tr_type tc tdt A),
      tr_tl (tr_cs f a) = f.
    Proof. move => f a; extensionality n; apply tr_tl_cs. Qed.


    Lemma tr_cs_tl : forall (r: tr_context (A::g)) (n: nat),
      tr_cs (tr_tl r) (r 0) n = r n.
    Proof. by move => r; case. Qed.


    Lemma tr_cs_tl_ext : forall (r: tr_context (A::g)),
      tr_cs (tr_tl r) (r 0) = r.
    Proof. by move => r; extensionality n; apply tr_cs_tl. Qed.

  End Utilities.


  Fixpoint tr_middle (A: type) (g d: context) {struct d} :
    tr_context (d++g) -> tr_type tc tdt A -> tr_context (d++A::g) :=
    match d as d return
      tr_context (d++g) -> tr_type tc tdt A -> tr_context (d++A::g) with
      | nil => tr_cs A (g:=g)
      | B::d => fun f a n =>
        match n return match opt_nth n (B::d++A::g) with
                         | Some C => tr_type tc tdt C
                         | None => unit
                       end with
          | O => f 0
          | S n => tr_middle A g d (tr_tl f) a n
        end
    end.


  Lemma tr_tl_middle : forall (d g: context) (A B: type)
    (r: tr_context (B::d++g)) (a: tr_type tc tdt A),
    tr_middle A g d (tr_tl r) a = tr_tl (tr_middle A g (B::d) r a).
  Proof.
    elim => //=.
  Qed.


  Lemma tr_cs_middle : forall d g A B r a b,
    tr_middle A g (B::d) (tr_cs B r b) a =
    tr_cs B (tr_middle A g d r a) b.
  Proof.
    elim => //=.
  Qed.


  Fixpoint tr_app (d g: context) {struct d} :
    tr_context (d++g) -> tr_context g :=
    match d as d return tr_context (d++g) -> tr_context g with
      | nil => fun f => f
      | A::d => fun f => tr_app d g (tr_tl f)
    end.


  Lemma tr_middle_app : forall (d: context) (U: type) (g: context)
    (r: tr_context (d++g)) (a: tr_type tc tdt U),
    tr_app d (U::g) (tr_middle U g d r a) 0 = a.
  Proof.
    elim => [//=|B d IHd].
    move => U g r a; rewrite [tr_middle]lock /= -lock -tr_tl_middle;
      apply: IHd.
  Qed.


  Fixpoint tr_weak_aux (d g: context) (n: nat) :
    tr_context (d++g) -> match opt_nth n d with
                           | Some ty => (tr_type tc tdt ty).(ttrans)
                           | None => unit
                         end :=
    match n as n, d as d
      return (tr_context (d++g) ->
        match opt_nth n d with
          | Some ty => (tr_type tc tdt ty).(ttrans)
          | None => unit
        end) with
      | _, nil => fun _ => tt
      | O, t::_ => fun r => r 0
      | S p, _::q => fun r => tr_weak_aux q g p (tr_tl r)
    end.


  Definition tr_weak (d g: context) : tr_context (d++g) -> tr_context d :=
    fun r n => tr_weak_aux d g n r.


  Lemma tr_tl_weak_aux : forall A g n d r, tr_tl (fun n => tr_weak_aux
    (A::d) g n r) n = tr_weak_aux d g n (tr_tl r).
  Proof.
    by move => A g; elim => [|n IHn] [|B d] r.
  Qed.


  Lemma tr_tl_weak : forall A d g r,
    tr_tl (tr_weak (A::d) g r) = tr_weak d g (tr_tl r).
  Proof.
    rewrite /tr_weak => A d g r; extensionality n;
      apply tr_tl_weak_aux.
  Qed.


  Lemma tr_cs_weak_aux : forall A a g n d r,
    tr_cs A (fun n0 : nat => tr_weak_aux d g n0 r) a n =
    tr_weak_aux (A :: d) g n (tr_cs A r a).
  Proof.
    move => A a g; case => [|n] //= d r.
    by rewrite tr_tl_cs_ext.
  Qed.


  Lemma tr_cs_weak : forall A d g r a,
    tr_cs A (tr_weak d g r) a = tr_weak (A :: d) g (tr_cs A r a).
  Proof.
    rewrite /tr_weak => A d g r a; extensionality n; apply tr_cs_weak_aux.
  Qed.


  (* Semantics of the empty context *)

  Definition tr_nil : tr_context nil :=
    fun n => match n as n return (
      match opt_nth n nil with
        | Some ty => tr_type tc tdt ty
        | None => unit
      end
    ) with
               | O => tt
               | _.+1 => tt
             end.


  Lemma tr_nil_weak_aux : forall g r n,
    tr_weak_aux nil g n r = tr_nil n.
  Proof.
    by move => g r [|].
  Qed.


  Lemma tr_nil_weak : forall g r,
    tr_weak nil g r = tr_nil.
  Proof.
    rewrite /tr_weak => g r; extensionality n; apply tr_nil_weak_aux.
  Qed.


  (* How to define term contexts *)

  Definition list_se := list (idV * { A : type & tr_type tc tdt A })%type.


  Fixpoint list_se2se (l: list_se) (x: idV) (X: type) {struct l} :
    tr_type tc tdt X :=
    match l with
      | nil => inhabitant_tr_type X
      | (y, existT Y p)::q => if (y == x) then
        match cast Y X with
          | Cast k => k _ p
          | NoCast => list_se2se q x X
        end else list_se2se q x X
    end.


  Definition list_sdt := list (defV * { A : type & tr_type tc tdt A })%type.


  Fixpoint list_sdt2sdt (l: list_sdt) (x: defV) (X: type) {struct l} :
    tr_type tc tdt X :=
    match l with
      | nil => inhabitant_tr_type X
      | (y, existT Y p)::q => if (y == x) then
        match cast Y X with
          | Cast k => k _ p
          | NoCast => list_sdt2sdt q x X
        end else list_sdt2sdt q x X
    end.

End Translation.


(* Translation from HOL to Coq: De Bruijn indices *)

Fixpoint sem_var (g: context) (n: nat) {struct n} :
  option {ty: type & forall (tc: type_context) (tdt: type_definition_tr),
    tr_context tc tdt g -> tr_type tc tdt ty} :=
  match n as n, g as g return
    option {ty: type & forall (tc: type_context) (tdt: type_definition_tr),
      tr_context tc tdt g -> tr_type tc tdt ty}
    with
    | _, nil => None
    | O, A::g =>
      Some (existT (fun ty => forall (tc: type_context)
        (tdt: type_definition_tr), tr_context tc tdt (A::g) ->
        tr_type tc tdt ty) A (fun (tc: type_context)
          (tdt: type_definition_tr) (f: tr_context tc tdt (A::g)) => (f 0)))
    | S n, A::g =>
      if (sem_var g n) is (Some (existT B k))
        then
          (Some (existT (fun ty => forall (tc: type_context)
            (tdt: type_definition_tr), tr_context tc tdt (A::g) ->
            tr_type tc tdt ty) B (fun (tc: type_context)
              (tdt: type_definition_tr) (f:tr_context tc tdt (A::g)) =>
              (k tc tdt (tr_tl f)))))
        else
          None
  end.


Lemma sem_var_correct : forall (n: nat) (g: context) (A: type),
  match (sem_var g n),(sem_var (A::g) (S n)) with
    | Some (existT B _), Some (existT C _) => B==C
    | None, None => true
    | _,_ => false
  end.
Proof.
  elim => [|n IHn] [|B g]//= A.
  by case: (sem_var _ _) => [[]|].
Qed.


Lemma sem_var_sem1 : forall (g: context) (A: type) (n: nat)
  (d: context), n < length d ->
  match sem_var (d++g) n, sem_var (d++A::g) n with
    | Some (existT N evN), Some (existT M evM) =>
      match cast N M with
        | Cast k =>
          forall (tc: type_context) (tdt: type_definition_tr)
            (r: tr_context tc tdt (d++g)) (a: tr_type tc tdt A),
            k (tr_type tc tdt) (evN tc tdt r) = evM tc tdt (tr_middle A g d r a)
        | NoCast => False
      end
    | _, _ => False
  end.
Proof.
  move => g A; elim => [|n IHn] [//=|B d];
    first by rewrite /= cast_same.
  move => hl; move/(_ _ hl): IHn; rewrite [tr_middle]lock /= -lock;
    case: (sem_var _ _) => //= [[N evN]].
  case: (sem_var _ _) => //= [[M evM]].
  by case: (cast _ _) => //= [k] hn tc tdt r a;
    rewrite (hn tc tdt (tr_tl r) a) tr_tl_middle.
Qed.


Lemma sem_var_sem2 : forall (g: context) (A: type) (n: nat)
  (d: context), length d <= n ->
  match sem_var (d++g) n, sem_var (d++A::g) (S n) with
    | Some (existT N evN), Some (existT M evM) =>
      match cast N M with
        | Cast k =>
          forall (tc: type_context) (tdt: type_definition_tr)
            (r: tr_context tc tdt (d++g)) (a: tr_type tc tdt A),
            k (tr_type tc tdt) (evN tc tdt r) = evM tc tdt (tr_middle A g d r a)
        | NoCast => False
      end
    | None, None => True
    | _, _ => False
  end.
Proof.
  move => g A; elim => [|n IHn] [|B d].
  - move => _; case: g => [|B g] //=.
  by rewrite cast_same; move => tc tdt r a; rewrite (tr_tl_cs A r a 0).
  - done.
  - move => _;
    replace (sem_var (nil ++ A :: g) n.+2) with (
      if (sem_var (nil++g) (S n)) is (Some (existT B k))
        then
          (Some (existT (fun ty => forall (tc: type_context)
            (tdt: type_definition_tr), tr_context tc tdt (A::g) ->
            tr_type tc tdt ty) B (fun (tc: type_context)
              (tdt: type_definition_tr) (f: tr_context tc tdt (A::g)) =>
              (k tc tdt (tr_tl f)))))
        else
          None
    ); last by done.
  case: (sem_var _ _) => [[N evN]|//=];
    first by rewrite cast_same /=; move => tc tdt r a; rewrite tr_tl_cs_ext.
  - move => hl;
    replace (sem_var ((B :: d) ++ g) n.+1) with (
      if (sem_var (d++g) n) is (Some (existT C k))
        then
          (Some (existT (fun ty => forall (tc: type_context)
            (tdt: type_definition_tr),tr_context tc tdt (B::d++g) ->
            tr_type tc tdt ty) C (fun (tc: type_context)
              (tdt: type_definition_tr) (f: tr_context tc tdt (B::d++g)) =>
              (k tc tdt (tr_tl f)))))
        else
          None
    ); last by done.
  replace (sem_var ((B :: d) ++ A :: g) n.+2) with (
    if (sem_var (d++A::g) (S n)) is (Some (existT C k))
      then
        (Some (existT (fun ty => forall (tc: type_context)
          (tdt: type_definition_tr), tr_context tc tdt (B::d++A::g) ->
          tr_type tc tdt ty) C (fun (tc: type_context)
            (tdt: type_definition_tr) (f: tr_context tc tdt (B::d++A::g))
            => (k tc tdt (tr_tl f)))))
      else
        None
  ); last by done.
  move/(_ _ hl): IHn;
    case: (sem_var _ _) => [[N evN]|].
  case: (sem_var _ _) => [[M evM]|//=].
  case: (cast _ _) => [k|//=].
  by move => IHn tc tdt r a; rewrite (IHn tc tdt (tr_tl r) a) tr_tl_middle.
  by case: (sem_var _ _).
Qed.


Lemma sem_var_sem3 : forall (d g: context) (U: type),
  sem_var (d++U::g) (length d) = Some (existT
    (fun ty => forall (tc: type_context) (tdt: type_definition_tr),
      tr_context tc tdt (d++U::g) -> tr_type tc tdt ty) U
    (fun _ _ r => (tr_app d (U::g) r) 0)).
Proof.
elim => [|A d] //=.
by move => IHd g U; rewrite (IHd g U).
Qed.


(* Translation from HOL to Coq: terms *)

Fixpoint sem_term (g: context) (t: term) {struct t} :
  option {ty: type & forall (tc: type_context) (tdt: type_definition_tr),
    term_context tc tdt -> term_definition_tr tc tdt ->
    tr_context tc tdt g -> tr_type tc tdt ty} :=
  match t with
    | Dbr n =>
      match sem_var g n with
        | Some (existT A ev) =>
          Some (existT (fun ty => forall (tc: type_context)
            (tdt: type_definition_tr), term_context tc tdt ->
            term_definition_tr tc tdt -> tr_context tc tdt g ->
            tr_type tc tdt ty) A (fun tc tdt _ _ r => ev tc tdt r))
        | None => None
      end
    | Var x A => Some (existT (fun ty => forall (tc: type_context)
      (tdt: type_definition_tr), term_context tc tdt ->
      term_definition_tr tc tdt -> tr_context tc tdt g -> tr_type tc tdt ty)
      A (fun tc tdt se _ _ => se x A))
    | Cst c => Some (existT (fun ty => forall (tc: type_context)
      (tdt: type_definition_tr), term_context tc tdt ->
      term_definition_tr tc tdt -> tr_context tc tdt g -> tr_type tc tdt ty)
    (type_cst c) (fun tc tdt _ _ _ => tr_cst tc tdt c))
    | Def c A => Some (existT (fun ty => forall (tc: type_context)
      (tdt: type_definition_tr), term_context tc tdt ->
      term_definition_tr tc tdt -> tr_context tc tdt g -> tr_type tc tdt ty)
    A (fun tc tdt _ sdt _ => sdt c A))
    | App t1 t2 =>
      match sem_term g t1, sem_term g t2 with
        | Some (existT (A1 --> B1) ev1), Some (existT A2 ev2) =>
          match (cast A2 A1) with
            | Cast k => Some (existT (fun ty => forall (tc: type_context)
              (tdt: type_definition_tr), term_context tc tdt ->
              term_definition_tr tc tdt -> tr_context tc tdt g ->
              tr_type tc tdt ty) B1 (fun tc tdt se sdt interpg =>
                ev1 tc tdt se sdt interpg (k (tr_type tc tdt) (ev2 tc tdt se sdt interpg))))
            | NoCast => None
          end
        | _, _ => None
      end
    | Abs A t => if sem_term (A::g) t is Some (existT B ev)
      then
        Some (existT (fun ty => forall (tc: type_context)
          (tdt: type_definition_tr), term_context tc tdt ->
          term_definition_tr tc tdt -> tr_context tc tdt g ->
          tr_type tc tdt ty) (A --> B) (fun tc tdt se sdt interpg a =>
            ev tc tdt se sdt (tr_cs _ interpg a)))
      else
        None
  end.


Lemma infer_sem : forall (t: term) (g: context),
  match infer g t, sem_term g t with
    | None, None => true
    | Some A, Some (existT B _) => A == B
    | _, _ => false
  end.
Proof.
  elim => [|x A g|c g|c A g|t IHt u IHu g|A t IHt g]; try by simpl.
  - elim => [|n IHn] [|A g]; try (simpl; done).
  replace (infer (A :: g) (Dbr n.+1)) with (infer g (Dbr n)); last trivial.
  move: (sem_var_correct n g A) (IHn g) => {IHn}; rewrite /sem_term;
    case: (sem_var g n) => [[A1 ev1]|]; case: (infer _ _);
      case: (sem_var _ _) => //= [[C _] D]; by move/eqP => ->.
  - move: (IHt g) (IHu g)  => {IHt IHu}; simpl;
    case: (infer _ u) => [U|]; case: (infer _ _)=>[[|||U1 V]|];
      case (sem_term g t)=>[[[|||U2 V'] st]|];
        case: (sem_term g u) => [[U' su]|] //=;
        move/eqP => [-> ->]; move/eqP => ->;
          rewrite /pred1 /= /eq_op /= /eqst;
            by case: (cast _ _) => //=; rewrite cast_same.
  - move: (IHt (A::g)) => {IHt} /=;
    case: (infer _ _); case: (sem_term _ _) => //=[[B st]] C;
      by move/eqP ->.
Qed.


Lemma wt_sem : forall (g: context) (t: term) (A: type),
  wt g t A <-> exists f, sem_term g t = Some (existT _ A f).
Proof.
  move => g t A; rewrite /wt; move: (infer_sem t g);
    case: (infer g t) => [B|] //=; case: (sem_term g t) => [[C f]|] //=.
  move /eqP => ->; split.
  by move => [<-]; exists f.
  by move => [h [h0 _]]; rewrite h0.
  by move => _; split => //=; move => [f h].
Qed.


Lemma wt_sem_inv2 : forall (g: context) (t: term) (A: type),
  wt g t A -> exists f, sem_term g t = Some (existT _ A f).
Proof.
  by move => g t A h; move: (wt_sem g t A) => [h1 _]; apply h1.
Qed.


Lemma wt_sem_type : forall (g: context) (t: term) (A: type),
  wt g t A -> {f | sem_term g t = Some (existT _ A f)}.
Proof.
  move => g t A; rewrite /wt; move: (infer_sem t g);
    case: (infer g t) => [B|] //=; case: (sem_term g t) => [[C f]|] //=.
  move /eqP => ->.
  by move => [<-]; exists f.
Defined.



(* Semantics and close *)

Lemma sem_var_close_aux : forall (g: context) (A: type) (n: nat)
  (d: context),
  match sem_var (d++g) n, sem_var (d++A::g)
    (if n < length d then n else S n)
    with
    | Some (existT B evB), Some (existT C evC) =>
      match cast B C with
        | Cast k => forall tc tdt r a,
          k (tr_type tc tdt) (evB tc tdt r) = evC tc tdt (tr_middle A g d r a)
        | NoCast => False
      end
    | None, None => True
    | _, _ => False
  end.
Proof.
  move => g A n d; case e: (n < length d).
  - move: (sem_var_sem1 g A n d e);
    case: (sem_var (d ++ g) n) => [[N evN]|] //=.
  have h: (length d <= n); first by rewrite leqNgt; apply negbT.
  move: (sem_var_sem2 g A n d h);
    case: (sem_var (d ++ g) n) => [[N evN]|] //=.
Qed.


Lemma sem_close_aux : forall (g: context) (x: idV) (A: type) (t: term)
  (d: context),
  match sem_term (d++g) t, sem_term (d++A::g) (close_aux t x A (length d))
    with
    | Some (existT B evB), Some (existT C evC) =>
      match cast B C with
        | Cast k => forall tc tdt se sdt r, k (tr_type tc tdt) (evB tc tdt se sdt r) =
          evC tc tdt se sdt (tr_middle A g d r (se x A))
        | NoCast => False
      end
    | None, None => True
    | _, _ => False
  end.
Proof.
  move => g x A; elim => [n|y Y|c|c C|t IHt u IHu|B t IHt] d //=.
  - move: (sem_var_close_aux g A n d);
    case: (sem_var (d ++ g) n) => [[B evB]|];
      case: (sem_var (d ++ A :: g) (if n < length d then n else n.+1)) =>
        [[C evC]|] //=.
  case: (cast B C) => [k|] //=.
  - case f: (x == y) => /=; last by rewrite cast_same.
  case e: (A == Y) => /=; last by rewrite cast_same.
  rewrite (sem_var_sem3 d g A); move /eqP: e ->;
    rewrite cast_same => tc tdt se _ r; rewrite tr_middle_app;
      by move /eqP: f ->.
  - by rewrite cast_same.
  - by rewrite cast_same.
  - move: (IHt d) (IHu d) => {IHt IHu};
    case: (sem_term (d ++ g) t) => [[[X|C||B C] evB]|];
      case: (sem_term (d ++ A :: g) (close_aux t x A (length d))) =>
        [[[Y|E||D E] evC]|];
        case: (sem_term (d ++ g) u) => [[F evF]|];
          case: (sem_term (d ++ A :: g) (close_aux u x A (length d))) =>
            [[G evG]|] //=.
  rewrite neq_cast; case eBD: (B == D); last done.
  move /eqP: eBD evC => <- evC {D}; rewrite cast_same;
    rewrite neq_cast; case eCE: (C == E); last done.
  move /eqP: eCE evC => <- evC {E}; rewrite cast_same => IHt;
    rewrite neq_cast; case eFG: (F == G); last done.
  move /eqP: eFG evG => <- evG {G}; rewrite cast_same => IHu;
    rewrite neq_cast; case eFB: (F == B); last done.
  move /eqP: eFB evF evG IHu => -> evF evG IHu {F}; rewrite !cast_same
    => tc tdt se sdt r; by rewrite IHt IHu.
  - move: (IHt (B::d)) => {IHt};
    case: (sem_term ((B :: d) ++ g) t) => [[C evC]|];
      case: (sem_term (B :: d ++ A :: g) (close_aux t x A (length d).+1)) =>
        [[D evD]|] //=.
  rewrite neq_cast; case eCD: (C == D); last done.
  move /eqP: eCD evD => <- evD {D}; rewrite !cast_same =>
    IHt tc tdt se sdt r; extensionality b; by rewrite IHt -tr_cs_middle.
Qed.


Lemma sem_close : forall (g: context) (x: idV) (A: type) (t: term),
  match sem_term g t, sem_term (A::g) (close t x A)
    with
    | Some (existT B evB), Some (existT C evC) =>
      match cast B C with
        | Cast k => forall tc tdt se sdt r, k (tr_type tc tdt) (evB tc tdt se sdt r) =
          evC tc tdt se sdt (tr_cs A r (se x A))
        | NoCast => False
      end
    | None, None => True
    | _, _ => False
  end.
Proof.
  rewrite /close => g x A t /=; exact (sem_close_aux g x A t nil).
Qed.


(* Semantics and weakening *)

Lemma sem_var_weak : forall g n d,
  match sem_var d n, sem_var (d++g) n with
    | None, _ => true
    | Some _, None => false
    | Some (existT A evA), Some (existT B evB) =>
      match cast A B with
        | Cast k => forall tc tdt (r: tr_context tc tdt (d++g)),
          k (tr_type tc tdt) (evA tc tdt (tr_weak d g r)) = evB tc tdt r
        | NoCast => false
      end
  end.
Proof.
  move => g; elim => [|n IHn] [|A d] //=.
  - by rewrite cast_same.
  - move: (IHn d); case: (sem_var d n) => //= [[B evB]].
  case: (sem_var (d++g) n) => //= [[C evC]].
  rewrite neq_cast; case eBC: (B == C); last done.
  move /eqP: eBC evC => <- {C} evC; rewrite cast_same => {IHn} IH tc tdt r;
    by rewrite -IH tr_tl_weak.
Qed.


Lemma sem_weak : forall g t d,
  match sem_term d t, sem_term (d++g) t with
    | None, _ => true
    | Some _, None => false
    | Some (existT A evA), Some (existT B evB) =>
      match cast A B with
        | Cast k => forall tc tdt se sdt (r: tr_context tc tdt (d++g)),
          k (tr_type tc tdt) (evA tc tdt se sdt (tr_weak d g r)) = evB tc tdt se sdt r
        | NoCast => false
      end
  end.
Proof.
  move => g; elim => [n|x X|c|c C|t IHt u IHu|A t IHt] d /=;
    try by rewrite cast_same.
  - move: (sem_var_weak g n d);
    case: (sem_var d n) => //= [[A evA]].
  case: (sem_var (d++g) n) => //= [[B evB]].
  rewrite neq_cast; case eAB: (A == B); last done.
  by move /eqP: eAB evB => <- {B} evB; rewrite cast_same.
  - move: (IHt d) (IHu d) => {IHt IHu}.
  case: (sem_term d t) => [[[|||A E] evA]|];
    case: (sem_term (d ++ g) t) => //= [[[|||B F] evB]];
      case: (sem_term d u) => [[C evC]|];
        case: (sem_term (d ++ g) u) => [[D evD]|] //=.
  rewrite neq_cast; case eAB: (A == B); last done.
  move /eqP: eAB evB => <- {B} evB; rewrite cast_same;
    rewrite neq_cast; case eEF: (E == F); last done.
  move /eqP: eEF evB => <- {F} evB; rewrite cast_same => IHt;
    rewrite neq_cast; case eCD: (C == D); last done.
  move /eqP: eCD evD => <- {D} evD; rewrite cast_same => IHu;
    rewrite neq_cast; case eCA: (C == A); last done.
  move /eqP: eCA evC evD IHu => -> {C} evC evD IHu;
    rewrite !cast_same => tc tdt se sdt r; by rewrite IHt IHu.
  - move: (IHt (A::d)) => {IHt};
    case: (sem_term (A :: d) t) => //= [[B evB]].
  case: (sem_term (A :: d ++ g) t) => //= [[C evC]].
  rewrite neq_cast; case eBC: (B == C); last done.
  move /eqP: eBC evC => <- {C} evC;
    rewrite !cast_same => IH tc tdt se sdt r; extensionality a;
      by rewrite -IH tr_cs_weak.
Qed.


(* Changing the semantic environment *)

Definition change_se tc tdt (se: term_context tc tdt) (x: idV) (X: type)
  (a: tr_type tc tdt X) : term_context tc tdt :=
  fun y Y => match cast X Y with
               | Cast k => if x == y then k (tr_type tc tdt) a else se y Y
               | NoCast => se y Y
             end.


Lemma change_se_not_free : forall x X u g, is_not_free x X u ->
  match sem_term g u with
    | Some (existT U evU) => forall tc tdt se sdt r a,
      evU tc tdt (change_se se x X a) sdt r = evU tc tdt se sdt r
    | None => True
  end.
Proof.
  move => x X; elim => [n|y Y|c|c C|t IHt u IHu|A t IHt] //= g H1.
  - by case: (sem_var _ _) => //= [[A evA]].
  - rewrite /change_se => tc tdt se _ _ a; move /orP: H1 => [H1|H1].
  case (cast X Y); last done.
  by move: (negbTE H1) ->.
  rewrite neq_cast; case eXY: (_ == _); last done.
  by move: (negbTE H1) eXY ->.
  - move /andP: H1 => [H1 H2]; move: (IHt g H1) (IHu g H2);
    case: (sem_term _ _) => [[[Y|C l||A B] evT]|] //= => {IHt} IHt.
  case: (sem_term _ _) => [[U evU]|] //= => {IHu} IHu.
  rewrite neq_cast; case eUA: (_ == _); last done.
  move /eqP: eUA evU IHu => -> {U} evU IHu;
    rewrite cast_same => tc tdt se sdt r a; by rewrite IHt IHu.
  - move: (IHt (A::g) H1);
    case: (sem_term _ _) => [[T evT]|] //= => {IHt} IHt tc tdt se sdt r a.
  extensionality b; by rewrite IHt.
Qed.


Lemma change_se_id : forall tc tdt (se: term_context tc tdt) x X a,
  change_se se x X a x X = a.
Proof.
  rewrite /change_se => tc tdt se x X a /=; by rewrite cast_same eq_refl.
Qed.

Unset Implicit Arguments.
