(* ========================================================================= *)
(* Typing of the HOL-Light terms                                             *)
(*                                                                           *)
(*       Chantal Keller, Laboratoire d'Informatique de Polytechnique         *)
(*                                                                           *)
(*    (c) Copyright, Laboratoire d'Informatique de Polytechnique 2010        *)
(*              (c) Copyright, Chantal Keller 2010                           *)
(* ========================================================================= *)


Require Import ssreflect ssrbool eqtype ssrnat.
Require Import list_ext axioms hol cast.

Require Import List.

Set Implicit Arguments.


(* Typing: De Bruijn indices *)

Definition context := list type.


(* How to deal with such contexts *)

Fixpoint opt_nth (A: Type) (n: nat) {struct n} : list A -> option A :=
  fun l => match n,l with
             | O, (x::_) => Some x
             | S m, (_::l) => opt_nth m l
             | _,_ => None
           end.


Lemma opt_nth_app : forall A g1 g2 (x: A),
  opt_nth (length g1)(g1++x::g2) = Some x.
Proof.
  move=>A; elim=>//=.
Qed.


Lemma opt_nth_le : forall A (l1: list A) l2 n,
  n < (length l1) ->
  opt_nth n (l1++l2) = opt_nth n l1.
Proof.
  move=> A; elim =>//= a l1 hl1 l2 [|n]//= h; auto.
Qed.


Lemma opt_nth_gt : forall A (l1 : list A) l2 n,
  n >= length l1 ->
  opt_nth n (l1++l2) = opt_nth (n - length l1) l2.
Proof.
  move=> A; elim=>/=[| a l1 hl1] l2 n; first by rewrite subn0.
  case:n=>//=n; rewrite subSS ltnS; auto.
Qed.


Lemma gt_opt_nth : forall A a (l:list A) n,
  opt_nth n l = Some a -> n < length l.
Proof.
  by move=> A a; elim =>[|b l hl]; move=> [|n]//=; rewrite ltnS; auto.
Qed.


Lemma opt_nth_nil : forall n A,
  opt_nth (A:=A) n nil = None.
Proof.
  by case.
Qed.


Lemma opt_nth_map : forall (A B: Type) (f: A -> B) (n: nat) (l: list A),
  match opt_nth n l with
    | Some a => Some (f a)
    | None => None
  end = opt_nth n (map f l).
Proof.
  move => A B f; elim => [|n IHn] [|a l] //=.
Defined.


Fixpoint opt_nth_map2 (A B: Type) (f: A -> B) (n: nat) (l: list A)
  {struct n} : opt_nth n (map f l) -> match opt_nth n l with
                                        | Some a => Some (f a)
                                        | None => None
                                      end :=
  match n as n, l as l return (opt_nth n (map f l) ->
    match opt_nth n l with
      | Some a => Some (f a)
      | None => None
    end) with
    | O, (x::_) => fun y => y
    | S m, _::q => opt_nth_map2 f m q
    | _, _ => fun y => y
  end.


(* Typing: constants *)

Definition type_cst (c: cst) : type :=
  match c with
    | Heq A     => A --> A --> Bool
    | Heps A    => (A --> Bool) --> A
    | Hand      => Bool --> Bool --> Bool
    | Hor       => Bool --> Bool --> Bool
    | Hnot      => Bool --> Bool
    | Himp      => Bool --> Bool --> Bool
    | Htrue     => Bool
    | Hfalse    => Bool
    | Hforall A => (A --> Bool) --> Bool
    | Hexists A => (A --> Bool) --> Bool
  end.


(* Typing: terms *)

Fixpoint infer (g: context) (t: term) : option type :=
  match t with
    | Dbr n => opt_nth n g
    | Var _ A => Some A
    | Cst c => Some (type_cst c)
    | Def _ ty => Some ty
    | App t1 t2 =>
      match infer g t1, infer g t2 with
        | Some (u1 --> u2), Some v => if v == u1 then Some u2 else None
        | _, _ => None
      end
    | Abs ty te =>
      match infer (ty::g) te with
        | Some ty2 => Some (ty --> ty2)
        | _ => None
      end
  end.


Definition wt (g: context) (te: term) (ty: type) : Prop :=
  infer g te = Some ty.


(* Inversion lemmas *)

Lemma wt_app_inv1 : forall (g: context) (t1 t2: term) (A: type),
  wt g (App t1 t2) A -> exists B, wt g t1 (B --> A) /\ wt g t2 B.
Proof.
  by move => g t1 t2 A; rewrite /wt => /=;
    case: (infer g t1) => [[|||B C]|] //=; case: (infer g t2) => [D|] //=;
      case e: (D == B) => //=; move: e; move /eqP => ->; move => [->];
        exists B.
Qed.


Lemma wt_app_inv2 : forall (g: context) (t1 t2: term) (A B: type),
  (wt g t1 (B --> A) /\ wt g t2 B) -> wt g (App t1 t2) A.
Proof.
  by move => g t1 t2 A B [h1 h2]; move: h1 h2; rewrite /wt => /=;
    case: (infer g t1) => [[|||C D]|] //=;
      move => [-> ->];
        case: (infer g t2) => [E|] //=;
          move => [->];
            move: (eq_refl B);
              case: (B == B).
Qed.


Lemma wt_abs_inv1 : forall (g: context) (t: term) (A B: type),
  wt g (Abs A t) B -> exists C, B == (A --> C) /\ wt (A::g) t C.
Proof.
by move => g t A B; rewrite /wt => /=;
  case: (infer (A :: g) t) => [C|] //=;
    case: B => [|||D E] //=;
      move => [-> ->];
        exists E.
Qed.


Lemma wt_abs_inv2 : forall (g: context) (t: term) (A B: type),
  wt (A::g) t B -> wt g (Abs A t) (A --> B).
Proof.
  by move => g t A B; rewrite /wt => /=;
    case (infer (A :: g) t) => [C|] //=;
      move => [->].
Qed.


Lemma wt_var_inv1 : forall (g: context) (y: idV) (A B: type),
  wt g (Var y A) B -> A == B.
Proof.
  by rewrite /wt => _ _ A B /= [->].
Qed.


Lemma wt_eq_inv1 : forall (g: context) (u v: term) (A: type),
  wt g (App (App (Cst (Heq A)) u) v) Bool -> wt g u A /\ wt g v A.
Proof.
  by move => g u v A h; move: (wt_app_inv1 h) => {h} [B [h1 h2]];
    move: (wt_app_inv1 h1) => {h1} [C [h0 h1]];
      rewrite /wt in h0; simpl in h0; move: h0 h1 h2 => [<- <-].
Qed.


Lemma wt_eq_inv2 : forall (g: context) (u v: term) (A: type),
  wt g u A -> wt g v A -> wt g (App (App (Cst (Heq A)) u) v) Bool.
Proof.
by move => g u v A h1 h2;
  do 2 (apply: wt_app_inv2; split; last by eassumption); rewrite /wt.
Qed.


(* Checks that a variable is not free in a term *)

Fixpoint is_not_free (x: idV) (X: type) (t: term) : bool :=
  match t with
    | Var y Y => (x != y) || (X != Y)
    | App t1 t2 => (is_not_free x X t1) && (is_not_free x X t2)
    | Abs _ u => is_not_free x X u
    | _ => true
  end.


(* Free variables of a term (we do not care about repetitions) *)

Fixpoint frees (t: term) : list (idV * type) :=
  match t with
    | Var y Y => (y,Y)::nil
    | App t1 t2 => (frees t1) ++ (frees t2)
    | Abs _ u => frees u
    | _ => nil
  end.


Definition frees_idv (t: term) : list idV :=
  List.map (fun (r: idV * type) => let (x,_) := r in x) (frees t).


(* If a variable is not in the free variables of a term, then it is not
   free in the term *)


Lemma frees_is_not_free : forall x X t, forallb (fun (r: idV * type) =>
  let (y,Y) := r in (x != y) || (X != Y)) (frees t) ->
  is_not_free x X t.
Proof.
  move => x X; elim => [|y Y|||t IHt u IHu|A t IHt] //=.
  - by rewrite andb_true_r.
  - move => H; move /andP: (forallb_app1 _ _ _ H) => [H1 H2];
    move: (andb_true_iff (is_not_free x X t) (is_not_free x X u)) => [_ H3];
      apply H3; split; [by apply IHt|by apply IHu].
Qed.


Lemma frees_idv_is_not_free : forall x t,
  forallb (fun y => x != y) (frees_idv t) ->
  forall X, is_not_free x X t.
Proof.
  rewrite /frees_idv => x t h X; apply frees_is_not_free; move: t h;
    elim => [n|y Y|c|c C|t IHt u IHu|A t IHt] //=.
  - rewrite !andb_true_r => h; by apply orb_true_intro; left.
  - move => h; apply forallb_app2; apply andb_true_intro;
    rewrite map_app in h; move /andP: (forallb_app1 _ _ _ h) => [h1 h2];
      split; [by apply IHt|by apply IHu].
Qed.


Lemma not_eq_sym : forall (A: eqType) (x y: A), (x != y) = (y != x).
Proof. move => A x y; apply negbLR; by rewrite negb_involutive. Qed.


Lemma frees_idv_is_not_free2 : forall x t,
  forallb (fun y => y != x) (frees_idv t) ->
  forall X, is_not_free x X t.
Proof.
  move => x t H1 X; apply frees_idv_is_not_free;
    have h: ((fun y : idV => x != y) = (fun y : idV => y != x)).
  extensionality y; apply not_eq_sym.
  by rewrite h.
Qed.


(* Locally-nameless utility *)

Fixpoint close_aux (t: term) (x: idV) (A: type) (i: nat) {struct t} :
  term :=
  match t with
    | Dbr n => Dbr (if n < i then n else (S n))
    | Var y B => if ((x == y) && (A == B)) then Dbr i else Var y B
    | App t1 t2 => App (close_aux t1 x A i) (close_aux t2 x A i)
    | Abs B u => Abs B (close_aux u x A (S i))
    | u => u
  end.

Definition close (t: term) (x: idV) (A: type) := close_aux t x A O.


Lemma wt_close_aux : forall (g: context) (x: idV) (A: type) (t: term)
  (d: context),
  infer (d++g) t = infer (d++A::g) (close_aux t x A (length d)).
Proof.
  move => g x A; elim => [n|y Y|c|c ty|t1 IHt1 t2 IHt2|B t IHt] //=.
  - move => d; case e: (_ < _); rewrite /infer;
    first by rewrite !opt_nth_le.
  rewrite opt_nth_gt; last by rewrite leqNgt; apply negbT.
  rewrite opt_nth_gt; last by apply leqW; rewrite leqNgt; apply negbT.
  by rewrite leq_subS; last by rewrite leqNgt; apply negbT.
  - move => d; case: (_ == _) => /=; last done.
  case e: (_ == _) => /=; last done.
  by rewrite opt_nth_app; move /eqP: e ->.
  - by move => d; rewrite (IHt1 d) (IHt2 d).
  - by move => d; rewrite (IHt (B::d)).
Qed.


Lemma wt_close_aux_inv1 : forall (g: context) (x: idV) (A: type) (t: term)
  (d: context) (T: type),
  wt (d++A::g) (close_aux t x A (length d)) T -> wt (d++g) t T.
Proof.
  by rewrite /wt => g x A t d T; rewrite (wt_close_aux g x A t d).
Qed.


Lemma wt_close_aux_inv2 : forall (g: context) (x: idV) (A: type) (t: term)
  (d: context) (T: type),
  wt (d++g) t T -> wt (d++A::g) (close_aux t x A (length d)) T.
Proof.
  by rewrite /wt => g x A t d T; rewrite (wt_close_aux g x A t d).
Qed.


Lemma wt_close_inv1 : forall (g: context) (t: term) (x: idV) (A T: type),
  wt (A::g) (close t x A) T -> wt g t T.
Proof.
  move => g t x A T; apply: (wt_close_aux_inv1 g x A t nil).
Qed.


Lemma wt_close_inv2 : forall (g: context) (t: term) (x: idV) (A T: type),
  wt g t T -> wt (A::g) (close t x A) T.
Proof.
  move => g t x A T; apply: (wt_close_aux_inv2 g x A nil).
Qed.


(* Closing and free variables *)

Lemma close_aux_not_free : forall x X u i,
  is_not_free x X (close_aux u x X i).
Proof.
  move => x X; elim => [n|y Y|c|c C|t IHt u IHu|A t IHt] i //=.
  - case e: (_ == _) => /=; last (apply orb_true_intro; left; by rewrite e).
  case f: (_ == _) => //=; apply orb_true_intro; right; by rewrite f.
  - apply andb_true_intro; split; [apply IHt|apply IHu].
Qed.


Lemma close_not_free : forall x X u,
  is_not_free x X (close u x X).
Proof.
  move => x X u; apply close_aux_not_free.
Qed.


(* Weakening *)

Lemma weakening : forall (u: term) (d g: context) (A: type),
  wt d u A -> wt (d++g) u A.
Proof.
  rewrite /wt; elim => //=.
  - move => n d g A h; transitivity (opt_nth n d); trivial;
    apply opt_nth_le; eapply gt_opt_nth; eexact h.
  - move => t IHt u IHu d g A; move: (IHt d g) => {IHt};
    move: (IHu d g) => {IHu}; case: (infer d t) => [[|||B C]|] //=;
      case: (infer d u) => [D|] //=; case e: (D == B) => //=;
        move => h; move: (h D (refl_equal _)) => {h} ->;
          move => h; move: (h (B --> C) (refl_equal _)) => {h} ->;
            by rewrite e.
  - move => A t IHt d g B; move: (IHt (A::d) g) => {IHt};
    case: (infer (A :: d) t) => [C|] //=; by move => h;
      move: (h C (refl_equal _)) => {h} ->.
Qed.

Unset Implicit Arguments.
